#ifndef NUS_ORDER_H
#define NUS_ORDER_H

// $Id: NUS_Order.h,v 1.1 2006/08/10 17:16:02 nicholas Exp $

#include "Global.H"
#include "Vector.H"

#include "NUS_Vector3D.h"
#include "NUS_Unit.h"
#include "NUS_Base.h"
#include <string>


namespace NUS_Simul{

	typedef Vector<Vector3D> WayPointVector;
	class Base;
	class Unit;
	class Group;
	
	class Order
	{
	public:
		Order(const WayPointVector& wayPoints,real8 speed);

		Order(const Order& b);

		Order* clone() const;

		WayPointVector getWayPoints() const { return wayPoints;}
		void setWayPoints(WayPointVector w) { wayPoints = w; }
		real8 getSpeed() const { return speed;}
		real8 getOrderTimeElapsed() const {return orderTimeElapsed;}

		void setActionType(const std::string& actionType) { this->actionType = actionType;}
		void setOrderTimeElapsed(real8 time) {this->orderTimeElapsed = time;}
		void setTarget(Base* base){this->targetBase=base;}
		void setTarget(Unit* unit){this->targetUnit=unit;}
		void setTarget(Group* group){this->targetGroup=group;}
		

		std::string getActionType() const {return actionType;}

		void setCompleted(bool completed) {this->completed = completed;}
		bool isCompleted() const {return completed;}

		Vector3D getFinalDestination(){
			if(wayPoints.empty()) ERR("trying to get destination of finished order!");

			return wayPoints.back();
		}

		Vector3D getDestination(){
			if(wayPoints.empty()) ERR("trying to get destination of finished order!");

			return wayPoints.front();
		}

		GameObj *getTargetObject() {
			if (targetUnit)
				return targetUnit->getObject();
			else {
				if (targetBase)
					return targetBase->getObject();
			}
			
			return NULL;
		}
		
		void goToNextWayPoint();
	
		bool executed;
		Base *targetBase;
		Unit *targetUnit;
		Group *targetGroup;

	protected:
		WayPointVector wayPoints;

		real8 speed; // the speed at which this unit travels

		std::string actionType; // "ATTACKGROUP", "ATTACKBASE", "ATTACKUNIT", "MOVE"

		real8 orderTimeElapsed;
		bool completed;

	private:
		Order& operator=(const Order& b);
	};
}

#endif
