#ifndef NUS_PLAYER_H
#define NUS_PLAYER_H

// $Id: NUS_Player.h,v 1.1 2006/08/10 17:16:02 nicholas Exp $

#include "Global.H"
#include "Vector.H"

#include <iostream>
#include <string>
#include <sstream>
#include <vector>
#include <set>
#include <map>
#include <cmath>
#include <algorithm>

#include "NUS_Base.h"
#include "NUS_Group.h"
#include "NUS_SimulationAI.H"
#include "InfluenceMap.H"

namespace NUS_Simul{

	class MapObject;
	class Base;
	class SimulationState;
	class ExtraObjectData;
	
	struct Task;

	typedef Vector<Group*> GroupVector;
	typedef Vector<Base*> BaseVector;
	typedef std::map<Task*,Task*> TaskCopiesMap;

	//typedef std::map<GameObj*, ExtraObjectData> ObjectDataMap;
	typedef std::map<Base*, ExtraObjectData> BaseDataMap;
	typedef std::map<Unit*, ExtraObjectData> UnitDataMap;

	typedef std::map<GameObj*, Base*> ObjectBaseMap;
	typedef std::map<GameObj*, Unit*> ObjectUnitMap;

	typedef std::map<Unit*, Group*> UnitGroupMap;
	
	class Player
	{
	public:
		Player(uint4 playerID, sint4 width, sint4 height);
		Player(uint4 playerID);
		~Player();
		Player(const Player& p);

		Player* clone();

		uint4 getID() const {return playerID;}

		void addBase(Base* newBase);

	//void drawObjects();
	//void drawObjectsInSelectionMode();

		void executeOrders(real8 timeElapsed);
		void updateOrders(SimulationState& state);
		void doOrders(SimulationAI* simAI);	//"do" all the actions out really

		void addOrtsBase(GameObj *obj);

		void addOrtsUnit(GameObj *obj);

		void removeObject(GameObj *obj);
		
		void removeOrtsBase(GameObj *obj);
		
		void removeOrtsBase(Base* obj);
		
		void removeOrtsUnit(Unit* obj);
		void removeOrtsUnit(GameObj *obj) ;
		
		UnitDataMap& getOrtsUnits(){return ortsUnits;}
		BaseDataMap& getOrtsBases(){return ortsBases;}
		
		ObjectUnitMap& getObjectUnitMap(){ return objectUnitMap; }
		ObjectBaseMap& getObjectBaseMap(){ return objectBaseMap; }
		UnitGroupMap& getUnitGroupMap(){ return unitGroupMap; }
		real4 getInfluence(NUS_TerrainBase::Loc loc);
		
	// OPTION 1: Player keeps track of Groups in addition to tasks
	// ISSUES: 1. Have to delete them once a task (can be sub-task) is completed 
	// Ans: Player owns the groups, and deletes them
	// 2. How to add groups to Player when it's NUS_Strategies.cpp generating 
	// them?! 
	// Ans: Task keeps a copy of SimulationState which keeps a copy of player
	
		void addGroup(Group* newGroup); // called by Tasks
		GroupVector& getGroups() {return groups;}
		void removeGroup(Group *gp);

		BaseVector* getBases() {return &bases;}

		void setVictoryType(sint4 type) { victoryType = type; }
		real8 getVictoryValue(); 

		void setStrategy(Task* strategy);
		Task* getStrategy() {return currentStrategy;}
		
		void updateObject(GameObj *obj);
		
		// NOTE: this should be called from Task::merge(Group *, Group*)
		void merge(Group *original, Group *other);

	private:
		void removeInternalUnit(Unit *obj);
		void removeInternalGroup(Group *group);
		
		Player &operator=(const Player &);

		BaseVector bases;     // owns objects pointed to
		GroupVector groups;
		
		Task* currentStrategy;   // owner
		uint4 playerID;
		
		std::map<Unit*,NUS_TerrainBase::Loc> lastInfluencePos; 	//map of unit to the distance it move since last update 

		UnitDataMap ortsUnits;
		BaseDataMap ortsBases;
		
		ObjectBaseMap objectBaseMap;
		ObjectUnitMap objectUnitMap;
		
		UnitGroupMap unitGroupMap;
		
		uint4 victoryType;
		
		InfluenceMap attack_influence;	
		bool useInfluence;
	};
}

REGISTER_TYPEOF(2841, NUS_Simul::ObjectDataMap::iterator);
REGISTER_TYPEOF(2842, NUS_Simul::ObjectDataMap::const_iterator);
REGISTER_TYPEOF(2843, NUS_Simul::BaseVector::iterator);
REGISTER_TYPEOF(2844, NUS_Simul::BaseVector::const_iterator);
REGISTER_TYPEOF(2845, NUS_Simul::GroupVector::iterator);
REGISTER_TYPEOF(2846, NUS_Simul::GroupVector::const_iterator);

#endif

