
#include "NUS_SimulationAI.H"

#include "NUS_SimulationState.h"
#include "NUS_Player.h"

#include "GameStateModule.H"
#include "Game.H"
#include "GameChanges.H"
//#include "GfxModule.H"
#include "NUS_TerrainModule.H"
#include "NUS_SquadCombatAI.H"

#include "NUS_Player.h"

#include "Options.H"

using namespace std;

static const bool debug = false;

namespace NUS_Simul {

	const sint4 SimulationAI::FROM = EventFactory::new_who();

	void SimulationAI::add_options()
	{
		::Options o("SimulationAI");
		o.put("-hunter", "Use Hunter, with targets being closest enemy unit to any unit in group");
		o.put("-hunter2","Use Hunter, with target being closest to ALL our units in group");
		o.put("-spread", "every unit continuously attacks nearest base to it");
		o.put("-halfdef", "Use task that send half units to mass and attack, while the other half defends bases spread out evenly");
		o.put("-joinspread","Join into small groups, which then attack nearest bases");
		o.put("-groupsize",sint4(50),"Specify how large groups are(used by hunter currently(default is use all 50 units)");
		o.put("-smartjoinsize",sint4(5),"specify how large local groups are formed before doing a larger join");
		o.put("-smartjoin", "Specify whether to form local groups when joining, or whether to just go and join");
		o.put("-moveforattack","When told to attack a unit, instead of using an attack task, use a move to location task(means shooterAI is not called, and there is less replanning(used by hunter1/2 right now)");
		o.put("-findspot","finds a smarter spot to do joins than just the average location of the units");
		o.put("-findbeststrategy","finds best strategy");
		o.put("-searchStrategy", "search for best strategy");
		o.put("-waypoint", "stop and wait for group member at waypoint");
		o.put("-victoryType", sint4(0), "Victory type - 0 for all bases destroyed, 1 for all units destroyed");
		o.put("-influence", "use influence map");
		o.put("-hunter3", "Use Hunter, with target being closest enemy unit to any unit in group and has weaker influence");
		o.put("-timeoutJoin",real8(0.0), "timeout for joins");
		o.put("-TravoAttack", "Use travoattack, win entire game instantly");
		o.put("-game4", "hack to improve performance of game 4");
		o.put("-moveVicinty", sint4(-1), "distance for move vicinty");
		o.put("-alwaysCheckStrategy", "always check strategy");
	}

	SimulationAI::SimulationAI(GameStateModule &gsm_, Terrain &timp_) : gsm(gsm_), timp(timp_){
		simulationState = new SimulationState(this);
	// crashes here for some reason, need to put it later
	//simulationState->setOurID(gsm.get_game().get_client_player());
		bestStrategy = NULL;
		lastSearchTime = 0.0;
		
		simulationResult = NULL;
	}
	SimulationAI::~SimulationAI(){
		if (simulationResult) delete simulationResult;
	}

//===================================================================

	void SimulationAI::act(GameStateModule &/*gsm*/){
		fprintf(stderr,"u...");
		simulationState->updateStrategies();
		getMe()->doOrders(this);
	}

//===================================================================

/**
	* Returns true if event was handled.
*/
		bool SimulationAI::handle_event(const Event &e)
		{
			//if (debug) std::cout <<  "Handling of event " << e.get_what() << std::endl;
			bool needStrategyUpdate = false;

			if (e.get_what() == NUS_TerrainModule::PATH_SUCCESS) {
				//if (debug) std::cout <<  "Handling PATH_SUCCESS" << std::endl;

				const NUS_TerrainModule::PathStatusEvent *p =
				dynamic_cast<const NUS_TerrainModule::PathStatusEvent*>(&e);
				assert(p && "SUCCESS: not a PathStatusEvent");

				FORALL(simulationState->getPlayers(),it){
					
					std::map<GameObj*, Unit*>::iterator it2 = (it->second)->getObjectUnitMap().find(p->get_obj());
					if (it2 == (it->second)->getObjectUnitMap().end()) continue;

						// if we were just moving... then we are done!
						/*if(it->second.state == ExtraObjectData::MOVE){
					it2->second.isDone = true;
						//needStrategyUpdate = true;
					}*/

						//we want to do update path here
						//find the group that this unit belongs to
					Unit *unit = (it->second)->getObjectUnitMap()[p->get_obj()];
					(it->second)->getOrtsUnits()[unit].isDone = true;
					(it->second)->getOrtsUnits()[unit].isExecuting = false;

					Group* objInGroup = (it->second)->getUnitGroupMap()[unit];
					needStrategyUpdate = updatePath(objInGroup);
					break;
				}
				
				//if (debug) std::cout <<  "End Handling PATH_SUCCESS" << std::endl;
			}

				if (e.get_what() == NUS_TerrainModule::PATH_FAILURE) 
				{
					if (debug) std::cout <<  "Handling PATH_FAILURE" << std::endl;

					const NUS_TerrainModule::PathStatusEvent *p =
						dynamic_cast<const NUS_TerrainModule::PathStatusEvent*>(&e);
					assert(p && "FAILURE: not a PathStatusEvent");

					//if (p->get_details() == NUS_TerrainModule::PATH_PLAN_FAILURE){
					if (debug) std::cout <<  "Path failure! This is bad, and shouldn't happen often" << endl;

					FORALL(simulationState->getPlayers(),it){

						std::map<GameObj*, Unit*>::iterator it2 = (it->second)->getObjectUnitMap().find(p->get_obj());
						if (it2 == (it->second)->getObjectUnitMap().end()) continue;

						/*	Unit *unit = (it->second)->getObjectUnitMap()[p->get_obj()];
							
							(it->second)->getStrategy()->removeUnit(unit);*/
						
						// TODO: right now, just say we are done if path fails.. not sure what to do here

						/*
						if(it2->second.hasReplanned){
							it2->second.isDone = true;
							needStrategyUpdate=true;
							continue;
						}

						if(it2->second.target == NULL){
							if(it2->second.location.x != -1){
								NUS_TerrainBase::Task task(NUS_TerrainBase::Goal::LOCATION, NUS_TerrainBase::Goal::TOUCH,
									NUS_TerrainBase::Task::INDIVIDUAL);



								task.goal.loc = it2->second.location;
								task.objs.insert(p->get_obj());
								doPath(task);
								it2->second.hasReplanned = true;

							}else{
								ERR("somethign is wrong, we should have either a target, or location");
							}
						}else{
							NUS_TerrainBase::Task task(NUS_TerrainBase::Goal::OBJ, NUS_TerrainBase::Goal::TOUCH,
								NUS_TerrainBase::Task::INDIVIDUAL);

							task.goal.obj = it2->second.target;
							task.objs.insert(p->get_obj());
							doPath(task);
							it2->second.hasReplanned = true;
							}*/
								
							Unit *unit = (it->second)->getObjectUnitMap()[p->get_obj()];
							//(it->second)->getOrtsUnits()[unit].isDone = true;

							/*
							Group *group = (it->second)->getUnitGroupMap()[unit];

							if ((it->second)->getOrtsUnits()[unit].hasReplanned) 
							{
								UnitVector unitVector(group->getUnits());
								
								FORALL(unitVector,uit) (it->second)->getOrtsUnits()[*uit].isDone = true;
												
								//Order *order = group->getOrder();
								
								//if (order) order->setCompleted(true);

								(it->second)->getOrtsUnits()[unit].isExecuting = false;
							}
							else */
							{
								UnitVector unitVector;				
								//unitVector.clear();
							
								unitVector.push_back(unit);
							
								replan(unitVector);
								
								(it->second)->getOrtsUnits()[unit].isDone = true;
							}
							
							needStrategyUpdate=true;

							(it->second)->getOrtsUnits()[unit].hasReplanned = true;
							
							//Group* objInGroup = (it->second)->getUnitGroupMap()[unit];
							//updatePath(objInGroup);							
						}
						
						if (debug) std::cout <<  "End Handling PATH_FAILURE" << std::endl;
					}
				

					if (e.get_what() == NUS_SquadCombatAI::TARGET_DESTROYED) {
						if (debug) std::cout <<  "Handling NUS_SquadCombatAI::TARGET_DESTROYED" << std::endl;

						const NUS_SquadCombatEvent *s = dynamic_cast<const NUS_SquadCombatEvent*>(&e);
						assert(s && "FAILURE: not a NUS_SquadCombatEvent");

						FORALL (s->units, u) {
							Player *player = simulationState->getPlayer(*(*u)->sod.owner);

							std::map<GameObj*, Unit*>::iterator it = player->getObjectUnitMap().find(*u);
							if (it == player->getObjectUnitMap().end()) continue;

							Unit *unit = player->getObjectUnitMap()[*u];
							player->getOrtsUnits()[unit].isDone = true;
							needStrategyUpdate = true;
						}

						stopCombat(s->units);
						if (debug) std::cout <<  "End Handling NUS_SquadCombatAI::TARGET_DESTROYED" << std::endl;
					}						

					bool mode;
					if (e.get_who() == GameStateModule::FROM) {

						if (e.get_what() == GameStateModule::VIEW_MSG) {
							//if (debug) std::cout <<  "Handling VIEW_MSG" << std::endl;

							const sint4 me = gsm.get_game().get_client_player();

						// this is a dumb place to put it, but it crashes if it is put into the constuctor
							simulationState->setOurID(gsm.get_game().get_client_player());

						// add new objects
						// NOTE: we assume this is only called once right now, since we see the whole map
						// this WILL NOT work for fog of war!

							bool addedUnit = false;

							FORALL (gsm.get_changes().new_objs, i) {
								GameObj *gob = (*i)->get_GameObj();
								if (gob == 0) continue;

								Player* player = simulationState->getPlayer(*gob->sod.owner);
								if(player == NULL){ // player does not exist yet.. create one
									
									bool useInfluence;
									::Options::get("-influence", useInfluence);
									
									if(useInfluence){
										sint4 tile_points = gsm.get_game().get_tile_points();

									  	sint4 width = gsm.get_game().get_map().get_width() * tile_points;
									  	sint4 height = gsm.get_game().get_map().get_height() * tile_points;
										player = new Player(*gob->sod.owner, width, height);
									}
									else{
										player = new Player(*gob->sod.owner);
									}
										
									simulationState->addPlayer(player);

									sint4 victoryType;
									::Options::get("-victoryType", victoryType);
								
									player->setVictoryType(victoryType);
								}
	
								/*
								if(isControllable(gob))
								{
									player->addOrtsUnit(gob);
									addedUnit = true;
								}else if(gob->bp_name() == "controlCenter"){
									player->addOrtsBase(gob);
								}
								*/
								if(gob->bp_name() == "controlCenter"){
									player->addOrtsBase(gob);
								}
								else
								{
									player->addOrtsUnit(gob);
									addedUnit = true;
								}
							}
							
							Player* ourPlayer = simulationState->getPlayer(simulationState->getOurID());

							if(addedUnit){				
								if(ourPlayer == NULL) ERR("umm, weirdness, our player should exist always");

								if(ourPlayer->getStrategy() != NULL){
									ERR("GOT NEW OBJECTS IN MULTIPLE FRAMES! CAN'T HANDLE THIS RIGHT NOW!");
								}

								GroupVector ourControllableObjects = ourPlayer->getGroups();

							/*FORALL(ourPlayer->getOrtsUnits(),it){
								ourControllableObjects.push_back(it->first);
								}*/

								if(!ourControllableObjects.empty()){
									sint4 gSize,smartSize;
									bool useSmart = false;
									bool useMove = false;
									bool useFindSpot = false;
									::Options::get("-groupsize", gSize);
									::Options::get("-smartjoinsize", smartSize);
									::Options::get("-smartjoin", useSmart);
									::Options::get("-moveforattack", useMove);
									::Options::get("-findspot", useFindSpot);

									simulationState->groupSize = (uint4)gSize;
									simulationState->smartJoinSize = (uint4)smartSize;
									simulationState->smartJoin = useSmart;
									simulationState->useMoveForAttack = useMove;
									simulationState->findspot = useFindSpot;

									::Options::get("-searchStrategy", mode);
									if (mode)
									{
									// FIXME: Change strategy sometimes?
										if (bestStrategy) {
											ourPlayer->setStrategy(bestStrategy(me, simulationState, ourControllableObjects, 0.0));
										} 
										else
										{
											cout << "starting search for best strategy" << endl;

											SimulationResult simResult;

											simResult.isDone = false;
											simResult.ourStrategyIndex = 1;
											simResult.theirStrategyIndex = 1;
											simResult.bestValue = -2e30;
											simResult.currentValue = 0.0;
											simResult.bestStrategyIndex = 0;

											lastSearchTime = getCurrentTime();

											bestStrategy = simulationState->searchStrategy(me, getCurrentTime(), 10000.0, 1000.0, simResult);

											if (bestStrategy)
											{
												cout << "found a best strategy" << endl;
												ourPlayer->setStrategy(bestStrategy(me, simulationState, ourControllableObjects, 0.0));
											}
											else
											{
												// set a default?
												cout << "failed to find a best strategy" << endl;
											}
										}						
									}
									else
									{
										::Options::get("-findbeststrategy", mode);
										if (mode)
										{
											if (simulationResult == NULL) {
												simulationResult = new SimulationResult;
												
												simulationResult->isDone = false;
												simulationResult->ourStrategyIndex = 1;
												simulationResult->theirStrategyIndex = 1;
												simulationResult->bestValue = -2e30;
												simulationResult->currentValue = 0.0;
												simulationResult->bestStrategyIndex = 0;
											}
											
											cout << "starting search for best strategy" << endl;
											
											// lazy! :
											SimulationResult &simResult(*simulationResult);

											bestStrategy = simulationState->findBestStrategy(me, getCurrentTime(), 0.5, 10.0, simResult);

											if (simResult.isDone)
											{
												cout << "found a best strategy" << endl;
												ourPlayer->setStrategy(bestStrategy(me, simulationState, ourControllableObjects, 0.0));
												needStrategyUpdate = true;
												delete simulationResult;
												simulationResult = NULL;
											}
											else
											{
												//needStrategyUpdate = true;
												cout << "failed to find a best strategy" << endl;
											}
											
										}
										else
										{
											::Options::get("-hunter", mode);
											if (mode)
												ourPlayer->setStrategy(createHunter(me, simulationState, ourControllableObjects, 0.0));
											else{
												::Options::get("-hunter2", mode);
												if (mode)
													ourPlayer->setStrategy(createHunter2(me, simulationState, ourControllableObjects, 0.0));
												else{
													::Options::get("-spread", mode);
													if (mode)
														ourPlayer->setStrategy(createSpreadAttack(me, simulationState, ourControllableObjects, 0.0));
													else{
														::Options::get("-halfdef", mode);
														if (mode)
															ourPlayer->setStrategy(createHalfUnitsBaseDefence_MassAttack(me, simulationState, ourControllableObjects, 0.0));
														else{
															::Options::get("-joinspread", mode);
															if (mode)
																ourPlayer->setStrategy(createJoinSpreadAttack(me, simulationState, ourControllableObjects, 0.0));
															else{
																::Options::get("-hunter3", mode);
																if (mode)
																	ourPlayer->setStrategy(createHunter3(me, simulationState, ourControllableObjects, 0.0));
																else{
																	::Options::get("-TravoAttack", mode);
																	if (mode)
																	ourPlayer->setStrategy(createTravoAttack(me, simulationState, ourControllableObjects, 0.0));
																	else {
																		ourPlayer->setStrategy(createMassAttack(me, simulationState, ourControllableObjects, 0.0));
																	}
																}
															}
														}
													}
												}
											}
										}
									}
								}


								needStrategyUpdate = true;
							}

							// remove dead objects
							//if (debug) std::cout <<  "Number of dead objs:" << gsm.get_changes().dead_objs.size() << std::endl;
							
							FORALL (gsm.get_changes().dead_objs, i) {
								GameObj *gob = (*i)->get_GameObj();
								if (gob == 0) continue;
								
								if (debug) std::cout <<  "Something just died!" << std::endl;
								
								Player* player = simulationState->getPlayer(*gob->sod.owner);
								if(player == NULL) ERR("problem, we have an object that does not belong to a player we know about");

								// if it's our unit, we have to make sure our current task also loses it
								needStrategyUpdate = informObjectsOfDeath(gob);
								if (debug) std::cout <<  "informed objects of death" << std::endl;

								player->removeObject(gob);
								if (debug) std::cout <<  "player has handled death" << std::endl;								
								needStrategyUpdate=true;				
							}

							Player *playerme = getMe();
							
							FORALL (playerme->getOrtsUnits(), it) {
								Unit *unit = it->first;
								
								GameObj *gob = unit->getObject();
								
								if (gob->is_dead()) continue;
								
								unit->addNewLoc(NUS_TerrainBase::Loc(*gob->sod.x, *gob->sod.y));
							}
							
							FORALL (gsm.get_changes().changed_objs, it) {
								GameObj *gob = (*it)->get_GameObj();
								if (gob == NULL) continue;
								
								simulationState->getPlayer(*gob->sod.owner)->updateObject(gob);
								/*
								if (simulationState->getPlayer(*gob->sod.owner) != ourPlayer) {
									FORALL(ourPlayer->getGroups(), git) {
										Order *order = (*git)->getOrder();
										
										if (order && order->getTargetObject() == *it) {
											// has it moved far??
											if (!order->getWayPoints().empty()) {
												Vector3D dest = order->getFinalDestination();
											
												if (sqrt( (dest.x-(*it)->get_int("x"))*(dest.x-(*it)->get_int("x")) \
												+ (dest.y-(*it)->get_int("y"))*(dest.y-(*it)->get_int("y"))) > 50.0 ) 
												{
													UnitVector &unitVector((*git)->getUnits());
													order->setCompleted(true);
													FORALL(unitVector,uit) ourPlayer->getOrtsUnits()[*uit].isDone = true;
													needStrategyUpdate=true;
													//replan(unitVector);
												}
											}
										}
									}
								}
								*/
								
								// REPLANS IF TARGET HAS MOVED -- DOES NOT WORK WELL!
								// check if it's a target of one of our groups
								if (simulationState->getPlayer(*gob->sod.owner) != ourPlayer) {
									FORALL(ourPlayer->getGroups(), git) {
										Order *order = (*git)->getOrder();
										
										if (order && order->getTargetObject() == *it) {
											// has it moved far??
											if (!order->getWayPoints().empty()) {
												Vector3D dest = order->getFinalDestination();
											
												if (sqrt( (dest.x-(*it)->get_int("x"))*(dest.x-(*it)->get_int("x")) \
												+ (dest.y-(*it)->get_int("y"))*(dest.y-(*it)->get_int("y"))) >= (*git)->getMaxAttackRange()*1.0 ) 
												{
													UnitVector &unitVector((*git)->getUnits());
														
													FORALL(unitVector,uit) ourPlayer->getOrtsUnits()[*uit].isDone = true;
													//FORALL(unitVector,uit) timp.update_obj((*uit)->getObject());	
													needStrategyUpdate=true;							
													order->setCompleted(true);
													//replan(unitVector);
												}
											}
										}
									}
								}
								
							}

							::Options::get("-alwaysCheckStrategy", mode);
							if (mode) needStrategyUpdate = true;
							
							/* Do we need to replan? */
							::Options::get("-searchStrategy", mode);
							if ( mode && getCurrentTime() - lastSearchTime > 10.0 )
							{
								cout << "starting search for best strategy" << endl;

								const sint4 me = simulationState->getOurID();

								SimulationResult simResult;

								simResult.isDone = false;
								simResult.ourStrategyIndex = 1;
								simResult.theirStrategyIndex = 1;
								simResult.bestValue = -2e30;
								simResult.currentValue = 0.0;
								simResult.bestStrategyIndex = 0;

								lastSearchTime = getCurrentTime();

								bestStrategy = simulationState->searchStrategy(me, getCurrentTime(), 10000.0, 1000.0, simResult);

								if (bestStrategy)
								{
									cout << "found a best strategy" << endl;

									Player* ourPlayer = simulationState->getPlayer(me);
									if(ourPlayer == NULL) ERR("umm, weirdness, our player should exist always");

									GroupVector ourControllableObjects = ourPlayer->getGroups();

									ourPlayer->setStrategy(bestStrategy(me, simulationState, ourControllableObjects, 0.0));
								}
								else
								{
									needStrategyUpdate = true;
									cout << "failed to find a best strategy" << endl;
								}
							}	

							  Game &game = gsm.get_game();	
						      sint4 vf = game.get_view_frame();
						      sint4 af = game.get_action_frame();
						      //sint4 sa = game.get_skipped_actions();
						      //sint4 ma = game.get_merged_actions();

							::Options::get("-findbeststrategy", mode);
							if (mode)
							{								
								if (vf !=0 && simulationResult && !simulationResult->isDone) {
									cout << "continuing search for best strategy" << endl;

									const sint4 me = gsm.get_game().get_client_player();

									//Player* ourPlayer = simulationState->getPlayer(simulationState->getOurID());
									//GroupVector ourControllableObjects = ourPlayer->getGroups();

									// lazy! :
									SimulationResult &simResult(*simulationResult);

									bestStrategy = simulationState->findBestStrategy(me, getCurrentTime(), 0.5, 10.0, simResult);

									if (simResult.isDone)
									{
										cout << "found a best strategy" << endl;
										//ourPlayer->setStrategy(bestStrategy(me, simulationState, ourControllableObjects, 0.0));
										//delete simulationResult;
										//simulationResult = NULL;
										//needStrategyUpdate = true;
									}
									else
									{
										cout << "failed to find a best strategy" << endl;
									}
								}
							}							
							
							if (mode && simulationResult && simulationResult->isDone)
							{
								if (abs(vf-af) < 5) {
									GroupVector ourControllableObjects = ourPlayer->getGroups();
									bestStrategy = strategyList[simulationResult->bestStrategyIndex];
									ourPlayer->setStrategy(bestStrategy(me, simulationState, ourControllableObjects, 0.0));
									delete simulationResult;
									simulationResult = NULL;
									needStrategyUpdate = true;
								}
							}
														
							/// hack for game 4
							bool mode;
							::Options::get("-game4", mode);
							if(mode){
								FORALL(getMe()->getGroups(), it){
									if((*it)->getOrder() && (*it)->getOrder()->executed && (*it)->getOrder()->getActionType()=="ATTACK" && !(*it)->getOrder()->isCompleted()){
										//check if closet unit and farthest unit to target unit is not too far..
										real8 maxDistance = 0;
										real8 minDistance = 2e8;
										Unit *minUnit, *maxUnit;
										GameObj* target =  (*it)->getOrder()->getTargetObject();
										FORALL((*it)->getUnits(), it2){
											if((*it2)->distance_to(target)>maxDistance){
												maxUnit = *it2;
												maxDistance = (*it2)->distance_to(target);
											}
											/*
											if((*it2)->distance_to(target)<minDistance){
												minUnit = *it2;
												minDistance = (*it2)->distance_to(target);
											}*/
										}
										//std::cout << "distance is " << maxDistance - minDistance << std::endl;
										UnitVector stopUnits;
										UnitVector goUnits;
										FORALL((*it)->getUnits(), it2){
											//this unit is too close, stop it
											if( maxDistance - (*it2)->distance_to(target) > 60.0){
												if (getMe()->getOrtsUnits()[*it2].isExecuting) stopUnits.push_back(*it2);
											}
											if( maxDistance - (*it2)->distance_to(target) <= 60.0){
												if (!getMe()->getOrtsUnits()[*it2].isExecuting) goUnits.push_back(*it2);
											}
										}
										stopPath(getMe(), stopUnits);
										replan(goUnits);
										
									}
								}
							}
							
							//if (debug) std::cout <<  "End Handling VIEW_MSG" << std::endl;
						}
					}


	// TODO for now, leave this check out.  it just results in slightly more work.  if i leave the check in
	// it seems to fail sometimes... not sure why.
	if(needStrategyUpdate){
		//if (debug) std::cout <<  "start needStrategyUpdate" << std::endl;
		act(gsm);
		//if (debug) std::cout <<  "Done needStrategyUpdate" << std::endl;
		//if (debug) std::cout <<  "Done handling event of type " << e.get_what() << std::endl;
		return true;
	}
	//if (debug) std::cout <<  "Done handling event of type " << e.get_what() << std::endl;
	return false;
	}

bool SimulationAI::informObjectsOfDeath(GameObj* obj){
	// FIXME: THIS FUNCTION SEEMS BUGGY!!
	
	bool needStrategyUpdate=false;
	//Player* player = simulationState->getPlayer(*obj->sod.owner);
	//if(player == NULL) ERR("bad stuff happening, our player should always exist!");

	Player* owner = simulationState->getPlayer(*obj->sod.owner);
	
	FORALL(simulationState->getPlayers(), it){
		
		Vector<GameObj*> stopAttacking;
		if (it->second == owner) 
		{
			//if (debug) std::cout <<  "SimulationAI::informObjectsOfDeath: handled own object death" << std::endl;
			
			if ((it->second)->getObjectUnitMap().find(obj) != (it->second)->getObjectUnitMap().end()) {
				Unit *unit = (it->second)->getObjectUnitMap()[obj];
				if ( (it->second)->getOrtsUnits()[unit].state == ExtraObjectData::ATTACK \
					|| (it->second)->getOrtsUnits()[unit].state == ExtraObjectData::MOVE )
				{
					//if (debug) std::cout <<  "stopping my own attack" << std::endl;
					(it->second)->getOrtsUnits()[unit].isDone = true;
					stopAttacking.push_back(obj);	
				}
			}
			needStrategyUpdate = true;		
		}
		else 
		{
			//if (debug) std::cout <<  "SimulationAI::informObjectsOfDeath: handled other player object death" << std::endl;
			
			Player *currPlayer = it->second;
			FORALL((it->second)->getOrtsUnits(),it2){
				Unit *unit = it2->first;
				if (unit == NULL) ERR("SimulationAI::informObjectsOfDeath: NULL object encountered");
				
				ExtraObjectData data( it2->second );
				// if the objects we had a target to was killed... this unit is done its task
				if(data.state == ExtraObjectData::ATTACK || data.state == ExtraObjectData::DEFEND){
				if((data.targetUnit && data.targetUnit->getObject() == obj) \
				|| (data.targetBase && data.targetBase->getObject() == obj) ){
					// stop this attack!
					//if (debug) std::cout <<  "stopping other attack!" << std::endl;
					stopAttacking.push_back(it2->first->getObject());
					data.isDone = true;
					needStrategyUpdate = true;
					
					Order *order = currPlayer->getUnitGroupMap()[unit]->getOrder();
					//if (order == NULL) ERR("SimulationAI::informObjectsOfDeath: order is NULL"); 
					if (order) order->setCompleted(true);
				}
				}
			}
			
		}

		if (!stopAttacking.empty())
			call_handlers(NUS_TerrainModule::PathStopEvent(FROM, NUS_TerrainModule::FIND_PATH_STOP, stopAttacking));

	}
	return needStrategyUpdate;
// TODO: maybe add updates for other objects later.  not needed right now, since we only control friendly units
}

bool SimulationAI::isControllable(GameObj* obj){
	// first, check if we even have a weapon...and can move...
	if ((obj->get_max_speed() > 0) && (obj->component("weapon") != NULL)){
		// if we can, make sure it can be used....this will likely always pass
		if (obj->component("weapon")->has_attr("max_damage")){
			// make sure we can actually do dmg
			if(obj->component("weapon")->get_int("max_damage") > 0){ // if we can attack
				return true;
		}
	}
}
return false;
}

//might want to combine this with the following function, but keep it first...
sint4 SimulationAI::doPath(NUS_TerrainBase::Task& task){
	//if (debug) std::cout <<  "Send Pathfinding Message from GAME 2\n";
	Player *player = getMe();
	
	FORALL (task.objs, it)
	{
		GameObj *gameObj = (GameObj*)*it;
		Unit *unit = player->getObjectUnitMap()[gameObj];
		
		//if (player->getOrtsUnits()[unit].isExecuting) stopPath(task);
		
		player->getOrtsUnits()[unit].isExecuting = true;
	}
	
	return call_handlers(NUS_TerrainModule::PathEvent(FROM, NUS_TerrainModule::FIND_PATH_MSG, task));
}

//send move command for an unit
//precondition: currObj is GameObj of unit to move
//							radius is the radius of circular area
// 							dest is centre in Vector3D
//postcondition: command sent, success or failure got to check from events
sint4 SimulationAI::doPath(Player* currPlayer, Vector<Unit*> &unitVector, Vector3D dest, sint4 radius, bool ignore_ALL){

	FORALL (unitVector, it) {
		//if (currPlayer->getOrtsUnits()[*it].isExecuting) stopPath(currPlayer, unitVector);
		currPlayer->getOrtsUnits()[*it].isExecuting = true;
	}
	
	NUS_TerrainBase::Loc destinationLoc(sint4(dest.x),sint4(dest.y));

	NUS_TerrainBase::Task task(NUS_TerrainBase::Goal::LOCATION, NUS_TerrainBase::Goal::VICINITY,
		NUS_TerrainBase::Task::INDIVIDUAL);

	task.consider = NUS_TerrainBase::Task::IGNORE_ALL;
	//if (ignore_ALL) task.consider = NUS_TerrainBase::Task::IGNORE_ALL;
	//if (!ignore_ALL) task.consider = NUS_TerrainBase::Task::IGNORE_FOES;
	
	ExtraObjectData newData;
	newData.location = destinationLoc;
	newData.state = ExtraObjectData::MOVE;
	newData.isDone = false;

	FORALL (unitVector,it) {
		currPlayer->getOrtsUnits()[*it] = newData;
		task.objs.insert((*it)->getObject());
	}
	
//note: we move unit to be within a circular target area by using NUS_TerrainBase::Goal::VICINITY option
	task.goal.loc = destinationLoc;
	
	sint4 groupsize;
	::Options::get("-groupsize", groupsize);

	task.goal.distance = radius; //groupsize * unitVector[0]->getObject()->get_int("radius") / 2;

	return doPath(task);
}

sint4 SimulationAI::stopPath(Player* currPlayer, Vector<Unit*> &unitVector){
	Vector<GameObj*> gameObjects;
	
	FORALL (unitVector,it) {
		gameObjects.push_back((*it)->getObject());
		
		currPlayer->getOrtsUnits()[*it].isExecuting = false;
	}
	
	//return 1;
	return call_handlers(NUS_TerrainModule::PathStopEvent(FROM, NUS_TerrainModule::FIND_PATH_STOP, gameObjects));
}

//plan path for a group
//preconditon: start and end point in Vector3D, empty WayPointVector
//post condition: WayPointVector contain the path in vector
real8 SimulationAI::planPath(const NUS_TerrainBase::Loc startPt, const NUS_TerrainBase::Loc endPt, sint4 radius, WayPointVector& wayPoints){

	Path path;
	// options are NUS_TerrainBase::IGNORE_ALL_MOBILE_OBJS and 
	/*
	enum ConsiderObjects {
		CONSIDER_ALL                = 0,
		/// Ignore all friendly (owned) mobile objects.
			IGNORE_MOBILE_FRIENDS   = 0x001, 
		/// Ignore all enemy mobile objects.
			IGNORE_MOBILE_FOES      = 0x010,
		/// Ignore all neutral ("sheep") mobile objects.
			IGNORE_MOBILE_NEUTRALS  = 0x100,
		/// Ignore all mobile objects, considering only boundaries and non-mobile objects.
			IGNORE_ALL_MOBILE_OBJS      = IGNORE_MOBILE_FRIENDS | IGNORE_MOBILE_FOES | IGNORE_MOBILE_NEUTRALS,
		};
	*/	
	real8 length;

	//if (debug) std::cout <<  "planPath: start" << std::endl;
    	
	length = timp.find_path(startPt, endPt, radius, &path, NUS_TerrainBase::IGNORE_ALL_MOBILE_OBJS );
	//length = timp.find_path(startPt, endPt, radius, &path, NUS_TerrainBase::IGNORE_MOBILE_FOES );
	//length = timp.find_path(startPt, endPt, radius, &path, NUS_TerrainBase::CONSIDER_ALL);
	
	FORALL(path.locs, it){
		wayPoints.push_back(Vector3D(it->x,it->y,0.0));
		//if (debug) std::cout <<  Vector3D(it->x,it->y,0.0).toString() << "\t";
	}
	
	return length;
}


//plan path for a group
//preconditon: start and end point in Vector3D, empty WayPointVector
//post condition: WayPointVector contain the path in vector
real8 SimulationAI::planPath(const NUS_TerrainBase::Loc startPt, const GameObj *target, sint4 radius, WayPointVector& wayPoints){

	Path path;
	// options are NUS_TerrainBase::IGNORE_ALL_MOBILE_OBJS and 
	/*
	enum ConsiderObjects {
		CONSIDER_ALL                = 0,
		/// Ignore all friendly (owned) mobile objects.
			IGNORE_MOBILE_FRIENDS   = 0x001, 
		/// Ignore all enemy mobile objects.
			IGNORE_MOBILE_FOES      = 0x010,
		/// Ignore all neutral ("sheep") mobile objects.
			IGNORE_MOBILE_NEUTRALS  = 0x100,
		/// Ignore all mobile objects, considering only boundaries and non-mobile objects.
			IGNORE_ALL_MOBILE_OBJS      = IGNORE_MOBILE_FRIENDS | IGNORE_MOBILE_FOES | IGNORE_MOBILE_NEUTRALS,
		};
	*/	
	real8 length;

	//if (debug) std::cout <<  "planPath: start" << std::endl;
	NUS_TerrainBase::Goal goal(NUS_TerrainBase::Goal::OBJ, NUS_TerrainBase::Goal::VICINITY);
	goal.obj = target;
	goal.distance = target->get_int("radius") + radius;
	
	// NOTE: This is based on modified version of pathfinding routines in NUS_ST_ForceField. UNTESTED!
	length = timp.find_path(startPt, goal, radius, &path, NUS_TerrainBase::IGNORE_ALL_MOBILE_OBJS );
    	
	//length = timp.find_path(startPt, endPt, radius, &path, NUS_TerrainBase::IGNORE_ALL_MOBILE_OBJS );
	//length = timp.find_path(startPt, endPt, radius, &path, NUS_TerrainBase::IGNORE_MOBILE_FOES );
	//length = timp.find_path(startPt, endPt, radius, &path, NUS_TerrainBase::CONSIDER_ALL);
	
	FORALL(path.locs, it){
		wayPoints.push_back(Vector3D(it->x,it->y,0.0));
		//if (debug) std::cout <<  Vector3D(it->x,it->y,0.0).toString() << "\t";
	}
	
	return length;
}

//execute the path of order of a group, called once when we want to execute an order for plan
//precondition: currGroup is group to move
//postcondition: move command sent for each unit

void SimulationAI::executePath(Group* currGroup){
	if (debug) std::cout <<  "executing Path" << std::endl;
	if(!currGroup->getOrder()->isCompleted()){
		bool mode;
		::Options::get("-waypoint", mode);
		Vector3D dest;
		
		if(mode){	//go waypoint by waypoint
			dest = currGroup->getOrder()->getDestination();
		}else{		//go straight to the final destination
			dest = currGroup->getOrder()->getFinalDestination();
		}
		
		Player* ourPlayer = getMe();

		if(ourPlayer == NULL) ERR("umm, weirdness, our player should exist always");

		/*FORALL(currGroup->getUnits(), it) {
			doPath(ourPlayer, *it, dest, sint4(currGroup->getExtents()));
		}*/
		
			sint4 groupsize;
			::Options::get("-groupsize", groupsize);
		sint4 radius = ((currGroup->getNumUnits() > groupsize ) ? currGroup->getNumUnits() : groupsize ) * currGroup->getUnits()[0]->getObject()->get_int("radius") / 3; 

		sint4 moveVicinty;
		::Options::get("-moveVicinty", moveVicinty);
		
		if (moveVicinty != -1) radius = moveVicinty;

		doPath(ourPlayer, currGroup->getUnits(), dest, radius); 
		//sint4(currGroup->getExtents()));
	}
}

	// WARNING! 
	// Assumes units are from the same group
	void SimulationAI::replan(UnitVector &unitVector) {
		
			// REPLAN MOVES WITH ONLY IGNORE_MOBILE_FOES (IF IT ACTUALLY WORKS)
			
		if (unitVector.empty()) return;
		
		Player* ourPlayer = getMe();
		
		Group* group = ourPlayer->getUnitGroupMap()[unitVector[0]];
		
		Order *currentOrder = group->getOrder();
		
		if (currentOrder && !currentOrder->isCompleted()) {
			if (currentOrder->getActionType() == "MOVE") {
				bool mode;
				::Options::get("-waypoint", mode);
				Vector3D dest;

				if(mode){	//go waypoint by waypoint
					dest = group->getOrder()->getDestination();
				}else{		//go straight to the final destination
					dest = group->getOrder()->getFinalDestination();
				}
				
				sint4 groupsize;
				::Options::get("-groupsize", groupsize);
				sint4 radius = ((group->getNumUnits() > groupsize ) ? group->getNumUnits() : groupsize ) * group->getUnits()[0]->getObject()->get_int("radius") / 3; 
				
				//move to the next way point
				FORALL (unitVector,it) {
					UnitVector uV;
					uV.push_back(*it);
					doPath(ourPlayer, uV, dest, radius);
					//group->getExtents());							
				}
			}
			else if (currentOrder->getActionType() == "ATTACK") {
				GameObj* currObj = unitVector[0]->getObject();
				GameObj* target = currentOrder->getTargetObject();
				
				if (target->is_dead()) return;

				uint4 distance = 0;

				uint4 weaponRange = currObj->component("weapon")->get_int("max_ground_range")/2;
				uint4 objRadius = currObj->get_int("radius")*4;
				distance = objRadius;

				if (distance > weaponRange) distance = weaponRange;

				NUS_TerrainBase::Task task(NUS_TerrainBase::Goal::OBJ, NUS_TerrainBase::Goal::VICINITY,
					NUS_TerrainBase::Task::INDIVIDUAL);

				task.goal.obj = target;
				//task.goal.loc.x = target->get_int("x");
				//task.goal.loc.y = target->get_int("y");

				task.goal.distance = distance;

				ExtraObjectData newData;
				newData.state = ExtraObjectData::ATTACK;

				if (currentOrder->targetUnit) {
					newData.targetUnit = currentOrder->targetUnit;
				}
				else {
					if (currentOrder->targetBase) {
						newData.targetBase = currentOrder->targetBase;
					}
					else ERR("Attack: No target set");
				}

				FORALL(unitVector,it) {
					ourPlayer->getOrtsUnits()[*it] = newData;

					task.objs.clear();
					task.objs.insert((*it)->getObject());
					//if (debug) std::cout <<  "sending group to attack by moving nearer" << std::endl;
					doPath(task);		
				}
			}
		}
		
	}

//update the path, move to next waypoint, if all units has reach the current waypoint
//called whenever we got a path success msg
	bool SimulationAI::updatePath(Group* group){

		bool needStrategyUpdate = false;
		
		//std::cout<< "updating path" << std::endl;
		
		bool mode;
		::Options::get("-waypoint", mode);

		if (group->getOrder() && group->getOrder()->getActionType() == "MOVE") {
			
			if(group->getOrder()->isCompleted()) ERR("Order is already completed");
			
			if(mode){	//we are going from waypoint to waypoint
				Player* ourPlayer = getMe();
				if(ourPlayer == NULL) ERR("umm, weirdness, our player should exist always");

				bool allDone = true;

				FORALL(group->getUnits(), it) {
					if(!ourPlayer->getOrtsUnits()[*it].isDone) {
						allDone = false;
						break;
					}
				}

				if(allDone){
					group->getOrder()->goToNextWayPoint();
	
					if(group->getOrder()->isCompleted()){		//if this path is done
						needStrategyUpdate = true;
					}
					else{
						Vector3D dest = group->getOrder()->getDestination();
						//move to the next way point
						FORALL(group->getUnits(), it){
							UnitVector unitVector;
							unitVector.push_back(*it);
							doPath(ourPlayer,unitVector, dest, group->getExtents());
						}
					}
				} //else we wait for all unit to reach its current way point first
			}
			else{	//we are going to the final destination straight
				group->getOrder()->setCompleted(true);	//we are completed already
				needStrategyUpdate = true;
			}
		} 
		else {
			if (group->getOrder() && group->getOrder()->getActionType() == "ATTACK") {		
				GameObj *target = group->getOrder()->getTargetObject();
				uint4 weaponRange = group->getMaxAttackRange();
				real8 groupDistance = group->distance_to(target);
				Order* currentOrder = group->getOrder();
				
				Player* ourPlayer = getMe();
				if(ourPlayer == NULL) ERR("umm, weirdness, our player should exist always");
				
				bool allWithinRange = true;
				
				Vector<GameObj::WPtr> foo;
				FORALL(group->getUnits(), it) {
					ourPlayer->getOrtsUnits()[*it].isDone = true;
					
					if ((*it)->distance_to(target) < weaponRange) {
						foo.clear();
						foo.push_back((*it)->getObject());
						GameObj *target = group->getOrder()->getTargetObject();
						doCombat(foo, target);
					} 
					else 
						allWithinRange = false;
				}
				// attack as a squad (does not seem to work well)
				/*
				if (!foo.empty()) {
					GameObj *target = group->getOrder()->getTargetObject();
					doCombat(foo, target);
				}
				*/
				if (allWithinRange) {
					group->getOrder()->setCompleted(true);		
				}
				needStrategyUpdate = true;
			}
		}
		
		return needStrategyUpdate;
	}

void SimulationAI::executeAttack(Group* currGroup, int phase){

	Player* ourPlayer = getMe();

	if(ourPlayer == NULL) ERR("umm, weirdness, our player should exist always");

	//if (debug) std::cout <<  "SimulationAI::executeAttack: starting" << std::endl;

	if (currGroup->getOrder() == NULL) ERR("SimulationAI::executeAttack: order is null");
	if (currGroup->getOrder()->getTargetObject() == NULL) ERR("SimulationAI::executeAttack: target is NULL");

	if (currGroup->getOrder()->getTargetObject()->is_dead()) {
		//if (debug) std::cout <<  "SimulationAI::executeAttack: trying to attack dead object" << std::endl;
		currGroup->getOrder()->setCompleted(true); //ERR("Trying to attack a dead object");
		return;
	}

	Order* currentOrder = currGroup->getOrder();
	GameObj* currObj = currGroup->getUnits()[0]->getObject();
	GameObj* target = currGroup->getOrder()->getTargetObject();
		
	uint4 distance = 0;
	
	if (phase == 0) {
		uint4 weaponRange = currObj->component("weapon")->get_int("max_ground_range")/2;
		uint4 objRadius = currObj->get_int("radius")*4;
		distance = objRadius;
		
		if (distance > weaponRange) distance = weaponRange;
	}
	
	NUS_TerrainBase::Task task(NUS_TerrainBase::Goal::OBJ, NUS_TerrainBase::Goal::VICINITY,
		NUS_TerrainBase::Task::INDIVIDUAL);
		
	task.goal.obj = target;
	//task.goal.loc.x = target->get_int("x");
	//task.goal.loc.y = target->get_int("y");	
	task.goal.distance = distance;

	ExtraObjectData newData;
	newData.state = ExtraObjectData::ATTACK;

	if (currentOrder->targetUnit) {
		newData.targetUnit = currentOrder->targetUnit;
	}
	else {
		if (currentOrder->targetBase) {
			newData.targetBase = currentOrder->targetBase;
		}
		else ERR("Attack: No target set");
	}
	
	FORALL(currGroup->getUnits(), it) {			
		currObj = (*it)->getObject();
		ourPlayer->getOrtsUnits()[(*it)] = newData;

		task.objs.clear();
		task.objs.insert(currObj);
		//if (debug) std::cout <<  "sending group to attack by moving nearer" << std::endl;
		doPath(task);		
	}

}

sint4 SimulationAI::doCombat(Vector<GameObj::WPtr> units, GameObj::WPtr target) {
	//if (debug) std::cout <<  "attack target" << endl;
	return call_handlers(NUS_SquadCombatEvent(FROM, NUS_SquadCombatAI::ATTACK_TARGET, units, target));
}

sint4 SimulationAI::stopCombat(Vector<GameObj::WPtr> units) {
	//if (debug) std::cout <<  "stop attack" << endl;
	return call_handlers(NUS_SquadCombatEvent(FROM, NUS_SquadCombatAI::STOP_ATTACK, units));
}

	//get my player
Player* SimulationAI::getMe(){
	const sint4 me = gsm.get_game().get_client_player();

	return simulationState->getPlayer(me);
}

}
