// $Id: NUS_SquadCombatAI.C,v 1.4 2007/03/24 22:55:55 orts_furtak Exp $

// This is an ORTS file (c) Tim Furtak licensed under the GPL

#include "NUS_SquadCombatAI.H"
#include "GameStateModule.H"
#include "Game.H"
//#include "GameObj.H"
#include "GameChanges.H"
#include "Profiler.H"
#include "Options.H"

using namespace std;

const sint4 NUS_SquadCombatAI::FROM = EventFactory::new_who();
const sint4 NUS_SquadCombatAI::SET_PRIORITY = EventFactory::new_what();
const sint4 NUS_SquadCombatAI::ATTACK_TARGET = EventFactory::new_what();
const sint4 NUS_SquadCombatAI::ATTACK_LOCATION = EventFactory::new_what();
const sint4 NUS_SquadCombatAI::ATTACK_FREELANCE = EventFactory::new_what();
const sint4 NUS_SquadCombatAI::MOVE_LOCATION = EventFactory::new_what();
const sint4 NUS_SquadCombatAI::STOP_ATTACK = EventFactory::new_what();

const sint4 NUS_SquadCombatAI::ENEMY_IN_RANGE = EventFactory::new_what();
const sint4 NUS_SquadCombatAI::TARGET_DESTROYED = EventFactory::new_what();
const sint4 NUS_SquadCombatAI::AT_LOCATION = EventFactory::new_what();

const real8 NUS_SquadCombatAI::overkill_percent;

//===================================================================

void NUS_SquadCombatAI::add_options()
{
  ::Options o("Squad");
  o.put("-localstrategy", 0, "target strategy selection using local rules");
  o.put("-globalstrategy", 0, "target strategy selection using global computation");
}

//===================================================================

NUS_SquadCombatAI::ShooterData::ShooterData() : dir(0,0), ndir(0,0)
{
  x = y = -1;
  min_damage = max_damage = avg_damage = -1;
  range = -1;
  hp = -1;
  cur_target = 0;
  old_target = 0;
  support = 0.0;
  squad = 0;
  mode = PASSIVE;
}

//===================================================================

NUS_SquadCombatAI::GroupData::GroupData(sint4 width, sint4 height) :
  attack_influence(width, height, 8),
  target_influence(width, height, 8),
  occupied_influence(width, height, 2)
{
}

//===================================================================

NUS_SquadCombatAI::NUS_SquadCombatAI(GameStateModule &gsm_) : gsm(gsm_), terrain_field(gsm_)
{
  gsm.add_handler(this);

  // Default Values
  enabled_attack = false;
  enabled_motion = false;
  friendlies = enemies = neutrals = 0;

  strategy = GLOBAL_STRATEGY;
  target_strategy = SHOOT_IMPORTANT;

  sint4 local_strategy, global_strategy;
  ::Options::get("-localstrategy", local_strategy);
  ::Options::get("-globalstrategy", global_strategy);

  if (local_strategy) {
    target_strategy = (TargetSelectionStrategy)local_strategy;
    strategy = LOCAL_STRATEGY;
  } else if (global_strategy) {
    target_strategy = (TargetSelectionStrategy)global_strategy;
    strategy = GLOBAL_STRATEGY;
  }

}

//-------------------------------------------------------------------

NUS_SquadCombatAI::~NUS_SquadCombatAI()
{
  FORU (i, groups.size()) {
    delete groups[i];
  }
}

//===================================================================

void NUS_SquadCombatAI::compute_squad_motion()
{
  const real4 move_mult = 10;
  const real4 attack_mult = 50;

  if (0)
  {
    Squad *sq = new Squad();
    sq->objective = Squad::FREELANCE;
    
    FORALL (friendlies->units, u) {
      GameObj::WPtr g = *u;
      ShooterData &sd = shooter_data[g];
      if (sd.squad == 0) {
        add_to_squad(g, sq);
      }
    }
    squads.push_back(sq);
    update_squads();
  }

  // first pass: compute desired direction ignoring countervailing influences
  if (enemies)
    compute_support(enemies->units);
  
  FORALL (squads, s) {
    
    std::set<GameObj::WPtr> finished;

    Squad &squad = **s;
    if (squad.objective == Squad::ATTACK_TARGET) {
      squad.target->get_center(squad.x, squad.y);
    }

    sint4 tx = squad.x;
    sint4 ty = squad.y;

    FORALL (squad.units, u) {

      GameObj::WPtr g = *u;
      ShooterData &sd = shooter_data[g];
      
      // goal direction

      Vec2<real4> force1(0,0);
      if (squad.objective == Squad::ATTACK_TARGET ||
          squad.objective == Squad::ATTACK_LOCATION ||
          squad.objective == Squad::MOVE_TO_POINT) {
        force1 = Vec2<real4>(tx - sd.x, ty - sd.y);
      }

      // misc. combat goals

      Vec2<real4> force2 = pro_vec(sd.x, sd.y, sd.dir.x, sd.dir.y); // using old dir

      if (squad.objective == Squad::MOVE_TO_POINT) {
        force2 *= move_mult;
      }
      if (squad.objective == Squad::FREELANCE) {
        force2 *= attack_mult;
      }
      if (squad.objective == Squad::ATTACK_TARGET ||
          squad.objective == Squad::ATTACK_LOCATION) {
        //        cout << "dir = " << sd.dir.x << " " << sd.dir.y << endl;
        //        cout << "pro vec = " << force2.x << " " << force2.y << endl;
        force2 *= attack_mult;
      }


      sd.dir = force1 + force2;
      sd.ndir = sd.dir;
      //      if (sd.ndir.length() > 0.000001) sd.ndir.normalize();
      if (sd.ndir.length() < 0.0001) {
        //        finished.insert(g);
      } else {
        sd.ndir.normalize();
      }
    }

    FORALL (finished, f) {
      squad.units.erase(*f);
    }
  }

  FORU_DYN (i, squads.size()) {
    if (squads[i]->units.empty()) {
      squads.erase(squads.begin()+i);
      --i;
    }
  }

  compute_support(friendlies->units);

  // second pass: compute opposing enemy influences

  FORALL (squads, s) {

    Squad &squad = **s;

    FORALL (squad.units, u) {

      GameObj::WPtr g = *u;
      ShooterData &sd = shooter_data[g];
      
      const Vec2<real4> &force1 = sd.dir;
      
      Vec2<real4> force2 = anti_vec(sd.x, sd.y, sd.ndir.x, sd.ndir.y);
      
      if (squad.objective == Squad::MOVE_TO_POINT) {
        force2 *= move_mult;
      }
      if (squad.objective == Squad::FREELANCE) {
        force2 *= attack_mult;
      }
      if (squad.objective == Squad::ATTACK_TARGET ||
          squad.objective == Squad::ATTACK_LOCATION) {
        force2 *= attack_mult;
        //        cout << "f1 = " << force1.x << " " << force1.y << " " << force1.length() << endl;
        //        cout << "f2 = " << force2.x << " " << force2.y << " " << force2.length() << endl;
        //        cout << "su = " << sd.support << endl;
      }


      // now succumb to peer pressure and make poor life choices
  
     Vec2<real4> force = force1 * 1 + force2 * (1.0 / (1.0 + sd.support * 2));
      //      force *= 10.0;

      real4 len = force.length();
      Vec2<real4> nforce = force;
      if (len > 0.001) nforce /= len;

      //      cout << "force1 = " << force1.x << " " << force1.y << endl;

      Vec2<real4> force3 = spacing_vec(sd.x, sd.y, force.x, force.y); // unnormalized
	
      // decrease spacing concerns when near the target
      if (0 && len < 32.0) {
        real4 scale = len / 32.0;
        cout << "scale = " << scale << endl;
        force3 *= scale;
      }

      //      cout << "force = " << force.x << " " << force.y << endl;
      //      cout << "spacing vec = " << force3.x << " " << force3.y << endl;

      force += force3;

      //      real4 l2 = force.length();
      //      cout << "l2 = " << l2 << endl;

      if (len > 8) {
        Vec2<real4> deflection = terrain_field.field(sd.x, sd.y, force.x, force.y);
        force += deflection*3;
      }


      if (len > 3.0) {
        
        Vector<sint4> params;
        params.push_back(sd.x + (sint4)force.x);
        params.push_back(sd.y + (sint4)force.y);
	if (enabled_motion)
	  g->set_action("move", params);
      } else {
        //        finished.insert(g);
      }

    }
  }

}

//-------------------------------------------------------------------

void NUS_SquadCombatAI::compute_support(const std::set<GameObj::WPtr> &units)
{
  FORALL (units, u) {

    GameObj::WPtr g = *u;
    ShooterData &sd = shooter_data[g];
    
    real4 s = 0.0;
    
    FORALL (units, it2) {
      if (u == it2) continue;
      ShooterData &sd2 = shooter_data[*it2];
      s += support(Vec2<real4>(sd.x, sd.y), Vec2<real4>(sd2.x, sd2.y));
    }

    sd.support = s;
  }
}

//-------------------------------------------------------------------

void NUS_SquadCombatAI::compute_support2(const std::set<GameObj::WPtr> &units)
{
  FORALL (units, u) {

    GameObj::WPtr g = *u;
    ShooterData &sd = shooter_data[g];
    
    real4 s = 0.0;
    
    FORALL (units, it2) {
      if (u == it2) continue;
      ShooterData &sd2 = shooter_data[*it2];
      s += support(Vec2<real4>(sd.x, sd.y), sd.ndir,
                   Vec2<real4>(sd2.x, sd2.y), sd2.ndir);
    }

    sd.support = s;
  }
}

//===================================================================

void NUS_SquadCombatAI::add_to_squad(GameObj::WPtr g, Squad *s, NUS_SquadCombatAI::ShooterData::Mode m)
{
  ShooterData &sd = shooter_data[g];
  if (sd.squad) {
    sd.squad->units.erase(g);
  }
  sd.squad = s;
  sd.mode = m;
  s->units.insert(g);
}

//-------------------------------------------------------------------

void NUS_SquadCombatAI::update_squads()
{
  FORU_DYN (i, squads.size()) {
    if (squads[i]->units.empty()) {
      squads.erase(squads.begin()+i);
      --i;
    }
  }
}

//===================================================================

bool NUS_SquadCombatAI::handle_event(const Event &e)
{
  if (e.get_what() == GameStateModule::VIEW_MSG) {
    update(gsm.get_changes());
    
    compute_attack();
    compute_squad_motion();
    
    return true;
  }

  //-------------------------------------

  if (e.get_what() == NUS_SquadCombatAI::SET_PRIORITY) {
    const NUS_ShooterPriorityEvent &spe = static_cast<const NUS_ShooterPriorityEvent&>(e);
    ShooterData &sd = shooter_data[spe.obj];
    if (spe.priority == NUS_ShooterPriorityEvent::ACTIVE) sd.mode = ShooterData::ACTIVE;
    if (spe.priority == NUS_ShooterPriorityEvent::PASSIVE) sd.mode = ShooterData::PASSIVE;
  }

  //-------------------------------------

  if (e.get_what() == NUS_SquadCombatAI::ATTACK_TARGET) {

    const NUS_SquadCombatEvent &sce = static_cast<const NUS_SquadCombatEvent&>(e);
    if (!sce.target.good()) return false;

    Squad *sq = new Squad();
    sq->target = sce.target;
    sq->target->get_center(sq->x, sq->y);
    sq->objective = Squad::ATTACK_TARGET;

    //    cout << "attack target " << sce.units.size() << endl;

    FORALL (sce.units, u) {
      GameObj::WPtr g = *u;
      add_to_squad(g, sq);
    }
    
    update_squads();
    squads.push_back(sq);
    
    return true;
  }

  //-------------------------------------

  if (e.get_what() == NUS_SquadCombatAI::ATTACK_LOCATION) {

    cout << "attack location" << endl;

    const NUS_SquadCombatEvent &sce = static_cast<const NUS_SquadCombatEvent&>(e);

    Squad *sq = new Squad();
    sq->x = sce.x;
    sq->y = sce.y;
    sq->target = 0;
    sq->objective = Squad::ATTACK_LOCATION;

    FORALL (sce.units, u) {
      GameObj::WPtr g = *u;
      add_to_squad(g, sq);
    }

    update_squads();
    squads.push_back(sq);

    return true;
  }

  //-------------------------------------

  if (e.get_what() == NUS_SquadCombatAI::ATTACK_FREELANCE) {
    //cout << "freelancing" << endl;
    REM("ATTACK_FREELANCE Event\n");
    const NUS_SquadCombatEvent &sce = static_cast<const NUS_SquadCombatEvent&>(e);

    Squad *sq = new Squad();
    sq->x = 0;
    sq->y = 0;
    sq->target = 0;
    sq->objective = Squad::FREELANCE;

    FORALL (sce.units, u) {
      GameObj::WPtr g = *u;
      add_to_squad(g, sq);
    }

    update_squads();
    squads.push_back(sq);

    return true;
  }

  //-------------------------------------

  if (e.get_what() == NUS_SquadCombatAI::MOVE_LOCATION) {
    cout << "move to loc" << endl;
    
    const NUS_SquadCombatEvent &sce = static_cast<const NUS_SquadCombatEvent&>(e);

    Squad *sq = new Squad();
    sq->x = sce.x;
    sq->y = sce.y;
    sq->target = 0;
    sq->objective = Squad::MOVE_TO_POINT;

    FORALL (sce.units, u) {
      GameObj::WPtr g = *u;
      add_to_squad(g, sq);
    }

    update_squads();
    squads.push_back(sq);

    return true;
  }

  //-------------------------------------

  return false;
}

//===================================================================

NUS_SquadCombatAI::GroupData* NUS_SquadCombatAI::get_group(GameObj::WPtr gob)
{
  assert(gob.good());

  sint4 me = gsm.get_game().get_client_player();
  sint4 npc = gsm.get_game().get_player_num();
  sint4 owner = *gob->sod.owner;

  if (owner == me) return friendlies;
  if (owner == npc) return neutrals;
  return enemies;
}

//===================================================================

void NUS_SquadCombatAI::update(const GameChanges &changes)
{
  sint4 npc = gsm.get_game().get_player_num()+1;
  sint4 tile_points = gsm.get_game().get_tile_points();

  width = gsm.get_game().get_map().get_width() * tile_points;
  height = gsm.get_game().get_map().get_height() * tile_points;
  game_tick = gsm.get_game().tick();

  if (groups.size() == 0) {
    FORU (i, npc) {
      groups.push_back(new GroupData(width, height));
    }
    if (npc == 2) {
      friendlies = groups[0];
      enemies = 0;
      neutrals = groups[1];
    }
    else if (npc >= 3) {
      friendlies = groups[0];
      enemies = groups[1];
      neutrals = groups[2];
    }
    else {
      cout << "npc = " << npc << endl;
      ERR("buh?");
    }
  }

  FORALL (changes.new_objs, i) {
    GameObj::WPtr gob = (*i)->get_GameObj();
    if (gob->get_shape() != Object::SHAPE_UNDEF)
      get_group(gob)->units.insert(gob);
  }

  FORALL (changes.vanished_objs, i) {
    GameObj::WPtr gob = (*i)->get_GameObj();
    GroupData *gd = get_group(gob);
    gd->units.erase(gob);
    shooter_data.erase(gob);

    FORALL (squads, s) {
      Squad &squad = **s;
      if (squad.objective == Squad::ATTACK_TARGET) {
        FIND (shooter_data, it, squad.target);
        if (it == shooter_data.end()) {
          squad.units.clear();
          continue;
        }
      }
      squad.units.erase(gob);
    }
  }

  FORALL (changes.dead_objs, i) {
    GameObj::WPtr gob = (*i)->get_GameObj();
    GroupData *gd = get_group(gob);
    gd->units.erase(gob);
    shooter_data.erase(gob);

    FORALL (squads, s) {
      Squad &squad = **s;
      if (squad.objective == Squad::ATTACK_TARGET) {
        FIND (shooter_data, it, squad.target);
        if (it == shooter_data.end()) {
          squad.units.clear();
          continue;
        }
      }
      squad.units.erase(gob);
    }
  }

  update_units();

  compute_attack_graph();
  recompute_targets();
  compute_influence();
}

//===================================================================

void NUS_SquadCombatAI::update_units()
{
  FORALL (groups, g) {
    FORALL ((*g)->units, uu) {
      GameObj::WPtr u = *uu;
      ShooterData &sd = shooter_data[u];
      
      const sintptr *hp = u->get_int_ptr("hp");
      sd.hp = hp ? *hp : -1;
      
      const sintptr *rad = u->get_int_ptr("radius");
      sd.r = rad ? *rad : -1;
      
      u->get_center(sd.x, sd.y);
      
      const sintptr *arm = u->get_int_ptr("armor");
      sd.armor = arm ? *arm : 0;

      sd.zcat = u->get_zcat(0);

      ScriptObj *weapon = u->component("weapon");
      if (!weapon) {
        sd.min_damage = sd.max_damage = sd.range = sd.time_to_shoot = -1;
        continue;
      }
      
      const sintptr *r = weapon->get_int_ptr("max_ground_range");
      sd.range = r ? *r : -1;

      const sintptr *air_r = weapon->get_int_ptr("max_air_range");
      sd.air_range = air_r ? *air_r : -1;

      const sintptr *min = weapon->get_int_ptr("min_damage");
      sd.min_damage = min ? *min : -1;
      
      const sintptr *max = weapon->get_int_ptr("max_damage");
      sd.max_damage = max ? *max : -1;
      
      sd.avg_damage = (sd.min_damage + sd.max_damage) / 2;

      const sintptr *los = weapon->get_int_ptr("line_of_sight");
      sd.line_of_sight = los ? *los : false; //fixme: Should default los be false or true?

      const sintptr *lf = weapon->get_int_ptr("last_fired");
      const sintptr *cd = weapon->get_int_ptr("cooldown");
      
      if ( lf && cd ) {
        sd.time_to_shoot = *cd - (game_tick - (uint4)*lf) ;
        if (sd.time_to_shoot < 0)
          sd.time_to_shoot = 0;
      }
      else
        sd.time_to_shoot = -1;
    }
  }
}

//===================================================================

/** Compute incoming and outgoing edges for friendly and enemy units. */

void NUS_SquadCombatAI::compute_attack_graph()
{
  PROFILE("NUS_SquadCombatAI - compute attack graph");
  if (!enabled_attack) return;

  FORALL (shooter_data, it) {
    it->second.outgoing.clear();
    it->second.incoming.clear();
  }

  if (friendlies && enemies) {
    compute_attack_graph(friendlies, enemies);
    compute_attack_graph(enemies, friendlies);
  }
}

//-------------------------------------------------------------------

void NUS_SquadCombatAI::compute_attack_graph(GroupData *ff, GroupData *ee)
{
  Game &game = gsm.get_game();

  FORALL (ff->units, f) {

    ShooterData &sdf = shooter_data[*f];
    if ((sdf.range < 0) && (sdf.air_range < 0)) continue;

    FORALL (ee->units, e) {

      ShooterData &sde = shooter_data[*e];
      sint4 range = 0;
      if (sde.zcat == GameObj::ON_LAND){
	range = sdf.range;
      }
      else if (sde.zcat == GameObj::IN_AIR){
	range = sdf.air_range;
      }
 

      real8 rdist = game.distance(**f, **e);
      sint4 sdist = (sint4)floor(rdist);
      if (sdist > range) continue;
      
      bool can_see = game.can_see_obj(**f, **e);
      if (sdf.line_of_sight && !can_see) continue;
      
      sdf.outgoing.push_back(*e);
      sde.incoming.push_back(*f);
    }
  }
}

//-------------------------------------------------------------------

/** [s]dir should be normalized before calling. */

real4 NUS_SquadCombatAI::support(const Vec2<real4> &pos, const Vec2<real4> &dir,
                             const Vec2<real4> &spos, const Vec2<real4> &sdir)
{
  Vec2<real4> dpos = spos - pos;
  real4 dist = dpos.length();

  real4 dot = dir.dot(sdir);
  if (dot <= 0.0) return 0.0;

  real4 val = (48.0 - dist) / 48.0;
  if (val <= 0.0) return 0.0;

  return dot * val;
}

//-------------------------------------------------------------------

real4 NUS_SquadCombatAI::support(const Vec2<real4> &pos, const Vec2<real4> &spos)
{
  Vec2<real4> dpos = spos - pos;
  real4 dist = dpos.length();

  real4 val = (48.0 - dist) / 48.0;
  if (val <= 0.0) return 0.0;

  return val;
}

//-------------------------------------------------------------------

/** Determine shooting actions to take for all units.
    Units that aren't part of a squad and that are currently PASSIVE should
    attack with low priority.
*/

void NUS_SquadCombatAI::compute_attack()
{
  if (!enabled_attack) return;

  Game &game = gsm.get_game();

  FORALL (friendlies->units, u) {

    GameObj *shooter = *u;
    //    if (shooter->action_already_set()) continue;

    ShooterData &sd = shooter_data[shooter];
    GameObj *targ = sd.cur_target;
    if (!targ) continue;

    ScriptObj *weapon = shooter->component("weapon");
    if (!weapon) continue;
    const sintptr *shooting = weapon->get_int_ptr("shooting");
    if (!shooting) continue;

    if (*shooting && targ == sd.old_target) continue;
    //    sd.cur_target = targ;

    Vector<sint4> args;
    args.push_back(game.get_cplayer_info().get_id(targ));

    sint4 priority = sd.mode == ShooterData::ACTIVE ? 3 : 2;
    
    //    cout << " TargetID: " << game.get_cplayer_info().get_id(targ) << " targ name: " << targ->bp_name() << " " << (void*)targ << "\n";
    weapon->set_action("attack", args, priority);
  }
}

//===================================================================

void NUS_SquadCombatAI::compute_influence()
{
  FORALL (groups, g) {
    //    (*g)->attack_influence.clear();
    //    (*g)->target_influence.clear();
    //    (*g)->occupied_influence.clear();

    (*g)->attack_field.clear();
    (*g)->target_field.clear();
    (*g)->occupied_field.clear();
    (*g)->motion_field.clear();
  }

  compute_influence(friendlies);
  compute_influence(enemies);

  compute_occupation(friendlies);
  compute_occupation(enemies);
  compute_occupation(neutrals);
}

//-------------------------------------------------------------------

void NUS_SquadCombatAI::compute_influence(GroupData *gd)
{
  if (!gd) return;

  FORALL (gd->units, u) {
    
    ShooterData &sd = shooter_data[*u];
    
    real4 tx = sd.x;
    real4 ty = sd.y;

    if (sd.range < 0) continue;
    real4 range = sd.range;

    sint4 hp = max(sd.hp, sd.min_damage);

    real4 q_value = 4.0 + (real4)sd.avg_damage / (real4)hp;
    real4 q_mult = 1.0 + (0.1 * sd.incoming.size());
    q_value *= q_mult;

    //    gd->attack_influence.add_aacircle(tx, ty, range, 0.1);
    //    gd->target_influence.add_aacircle(tx, ty, range+32, q_value*0.3);

    //    cout << "q value = " << q_value << endl;

    gd->attack_field.add_emitter(new VectorField::PointEmitter2(tx, ty, range, range+32, 1.0 + sd.support));
    gd->target_field.add_emitter(new VectorField::PointEmitter2(tx, ty, range+128, range+256, q_value));
  }
}

//-------------------------------------------------------------------

void NUS_SquadCombatAI::compute_occupation(GroupData *gd)
{
  if (!gd) return;

  FORALL (gd->units, u) {
    
    ShooterData &sd = shooter_data[*u];
    
    real4 tx = sd.x;
    real4 ty = sd.y;

    VectorField::OrthoEmitter *orth = new VectorField::OrthoEmitter(tx, ty, sd.r + 4, sd.r + 16, 1.0);
    gd->occupied_field.add_emitter(orth);

    //    gd->occupied_influence.add_aacircle(tx, ty, 6.0, 1.0);

#if 1
    VectorField::Emitter *occ = new VectorField::GaussianEmitter(tx, ty, sd.r+8, sd.r + 16, 10, 1.0);
    //    VectorField::Emitter *occ = new VectorField::GaussianEmitter(tx, ty, 32, 20, 2.0);
    if (fabs(sd.dir.x) > 0.001 || fabs(sd.dir.y) > 0.001) {
      VectorField::Emitter *dir = new VectorField::DirMultiplier(tx, ty, sd.dir.x, sd.dir.y, occ);
      gd->motion_field.add_emitter(dir);
    } else {
      gd->motion_field.add_emitter(occ);
    }
#endif
#if 1
    VectorField::Emitter *e2 = new VectorField::GaussianEmitter(tx, ty, sd.r+8, sd.r + 12, 10, 0.1);
    gd->motion_field.add_emitter(e2);
#endif
  }
}

//===================================================================

Vec2<real4> NUS_SquadCombatAI::pro_vec(sint4 x, sint4 y, real4 dx, real4 dy)
{
  if (!enemies) return Vec2<real4>(0,0);
  assert(friendlies);

  Vec2<real4> enemy_target = enemies->target_field.field(x, y, dx, dy);
  Vec2<real4> friend_attack = friendlies->attack_field.field(x, y, dx, dy);
  Vec2<real4> friend_target = friendlies->target_field.field(x, y, dx, dy);

  //  cout << "target val = " << enemy_target.length() << endl;

  if (fabs(enemy_target.x) > 0.1 && fabs(enemy_target.y) > 0.1) enemy_target.normalize();
  if (fabs(friend_attack.x) > 0.1 && fabs(friend_attack.y) > 0.1) friend_attack.normalize();
  if (fabs(friend_target.x) > 0.1 && fabs(friend_target.y) > 0.1) friend_target.normalize();

  //  return (occupied * 10);

  return (enemy_target  * -1.0) +
         (friend_attack * -0.2) +
         (friend_target * -0.3);
}

//===================================================================

Vec2<real4> NUS_SquadCombatAI::spacing_vec(sint4 x, sint4 y, real4 dx, real4 dy)
{
  Vec2<real4> dir(dx, dy);
  real4 len = dir.length();
  if (len > 0.001) {
    dir /= len;
  } else {
    len = 1.0;
  }

  Vec2<real4> occupied;
  FORALL (groups, g) {
    occupied += (*g)->occupied_field.field(x, y, dir.x, dir.x);
  }

  Vec2<real4> motion;
  FORALL (groups, g) {
    motion += (*g)->motion_field.field(x, y, dir.x, dir.y);
  }

  return ((motion * 0.2) + (occupied * 0.5)) * len;
}

//===================================================================

Vec2<real4> NUS_SquadCombatAI::anti_vec(sint4 x, sint4 y, real4 dx, real4 dy)
{
  if (!enemies) return Vec2<real4>(0,0);

  Vec2<real4> enemy_attack = enemies->attack_field.field(x, y, dx, dy);

  //  cout << "danger val = " << enemy_attack.length() << endl;

  return (enemy_attack * 1.0);
}

//===================================================================
/*
void NUS_SquadCombatAI::remove_obj(const GameObj::WPtr gob)
{

}
*/
//===================================================================
//===================================================================

NUS_SquadCombatEvent::NUS_SquadCombatEvent(sint4 who, sint4 what, const Vector<GameObj::WPtr> &u,
                                   sint4 x_, sint4 y_)
  : Event(who, what), units(u), x(x_), y(y_), target(0), objective(ATTACK_LOCATION)
{
}

//-------------------------------------------------------------------

NUS_SquadCombatEvent::NUS_SquadCombatEvent(sint4 who, sint4 what, const Vector<GameObj::WPtr> &u, GameObj::WPtr targ)
  : Event(who, what), units(u), x(0), y(0), target(targ), objective(ATTACK_TARGET)
{
}

//-------------------------------------------------------------------

NUS_SquadCombatEvent::NUS_SquadCombatEvent(sint4 who, sint4 what, const Vector<GameObj::WPtr> &u)
  : Event(who, what), units(u), x(0), y(0), target(0), objective(STOP)
{
}

//===================================================================

