// $Id: NUS_Strategies.cpp,v 1.1 2006/08/10 17:16:02 nicholas Exp $


#include "NUS_Strategies.h"

#include "NUS_SimulationAI.H"
#include "NUS_SimulationState.h"

#include "GameStateModule.H"
#include "Game.H"
#include "GameChanges.H"
#include "GameObj.H"
//#include "GfxModule.H"
#include "NUS_TerrainModule.H"
#include "TerrainClassifier.H"
#include "NUS_Order.h"
#include "Options.H"

static const bool debug = false;

namespace NUS_Simul { 

#define DEFINE_TASK(name) \
	name::name(uint4 playerID, \
		SimulationState* state, \
		GroupVector& groupVector, \
		TaskCreateFunctionPtr f, \
		std::string n, \
		real8 timeout, \
		MoveStruct* extraData): \
	Task(playerID,state,groupVector,f,n,timeout,extraData){ \
		exec();	\
		} \
		name::name(Task* parent, \
		GroupVector& groupVector, \
		TaskCreateFunctionPtr function, \
		std::string n, \
		real8 timeout, \
		MoveStruct* extraData) : \
	Task(parent,groupVector,function,n,timeout,extraData){ \
		exec(); \
		} \
		name::name(const name& t) : Task(t){} \
	Task* name::clone(SimulationState* state, Task* newParent){ \
		name* newTask = new name(*this); \
		newTask->parent = newParent; \
		newTask->state = state; \
	FORALL(children,it){ \
		newTask->children.push_back((*it)->clone(state,newTask)); \
		} \
		return newTask;	\
		} \
		bool name::exec()

// ----------------
// Basic Tasks

	DEFINE_TASK(SpreadAttack) {
		UnitVector units;	//all units in all newGroups
		GroupVector temp;
		
		if (debug) std::cout << "SpreadAttack: Start by player " << playerID << std::endl;
		
		switch(lineNum) {
			case 0:

			temp = groups;
			FORALL(temp, it) {
				Base* nearestBase = state->getNearestEnemyBaseOfGroup(playerID, (*it)->getUnits());
				if(nearestBase != NULL) {
					GroupVector gv;
					gv.push_back(*it);
					new Attack(this, gv, function, "Attack", 0.0, new MoveStruct(nearestBase));
				} else return true; //ERR("No base to attack?");
			}
			default:
			return false;
		}
		return false;
	}

	DEFINE_TASK(Join) {
		assert(extraData == NULL); // extradata should be NULL for this task
		UnitVector units;

		if (debug) std::cout << "Join: Start" << std::endl;

		switch(lineNum) {
			case 0:{
				if (debug) std::cout << "Join: creating newGroups consisting of singleton units" << std::endl;
				
				// algo: get average position, then get nearest group to average point
				
				Vector3D average(0.0, 0.0, 0.0);
				int totalNumUnits = 0;
				
				FORALL(groups, it){
					average += (*it)->getPosition() * (double)(*it)->getNumUnits();
					totalNumUnits += (*it)->getNumUnits();
				}	
				
				average /= (double) totalNumUnits;
				
				real8 nearest = 2e8;
				Group *nearestGroup = NULL;
				
				FORALL (groups, it) {
					real8 dist = ((*it)->getPosition() - average).length();
					if (dist < nearest) {
						nearest = dist;
						nearestGroup = (*it);
					}
				}
				
				NUS_TerrainBase::Loc center_loc(nearestGroup->getPosition().x, nearestGroup->getPosition().y);
								
				GroupVector &farAwayGroups(groups);
			
				if(state->findspot){
					GameStateModule & gsm = state->simAI->get_gsm();
					center_loc = find_a_spot(center_loc, gsm, state);
				}

				if (!farAwayGroups.empty()) {
					real8 timeoutJoin;
					::Options::get("-timeoutJoin", timeoutJoin);
					
					if (debug) std::cout << "Join: Pathfinding for individual units for Join" << std::endl;
					new MoveToLocation(this, farAwayGroups, function, "MoveToLocation", timeoutJoin, new MoveStruct(center_loc));
					++lineNum;
					break;
				} else 
					++lineNum; // close enough, proceed to next step
			}
			case 1: {
				//re-group all the newGroups into one group after joining	
				//add all units from all newGroups into a unit vector and remove the old newGroups
				UnitVector units;
			
				FORALL(groups, it){
					FORALL((*it)->getUnits(), it2){
						units.push_back((*it2));
					}
				}
			
				removeAllGroups();
			
				Group *group = new Group(playerID, units);
				addGroup(group);
				
				++lineNum;
				return true;
			}
			default:
			return true;
		}
		return false;
	}

	DEFINE_TASK(Attack) {
		assert(extraData != NULL);
		NUS_TerrainBase::Loc targetLoc;
		GameObj *target;
		
		if (debug) std::cout << "Attack: Start" << std::endl;

		switch(lineNum) {
			case 0:
			// everyone should arrive at the same time

			//get the location of the target
			if (extraData->unit) {
				targetLoc.x = sint4(extraData->unit->getPosition().x);
				targetLoc.y = sint4(extraData->unit->getPosition().y);
				target = NULL;
			}
			else {
				if (extraData->base) {
					targetLoc.x = sint4(extraData->base->getPosition().x);
					targetLoc.y = sint4(extraData->base->getPosition().y);
					target = extraData->base->getObject();
				}
				else ERR("Attack: No target set");
			}
			
			//if (debug) std::cout << "Target Loc: (" << targetLoc.x << "," << targetLoc.y << ")" << std::endl;

			FORALL(groups, it){
				NUS_TerrainBase::Loc startPt((*it)->getPosition().x, (*it)->getPosition().y);
				WayPointVector wayPoints;

				real8 length = 0.0;
				
				if (target) 
					length = state->simAI->planPath(startPt, target, (*it)->getExtents(), wayPoints);
				else
					length = state->simAI->planPath(startPt, targetLoc, (*it)->getExtents(), wayPoints);
				
				Order* order = new Order(wayPoints, (*it)->getMaxSpeed());
				
				if (extraData->unit) {
					order->setTarget(extraData->unit);
				}
				else {
					if (extraData->base) {
						order->setTarget(extraData->base);
					}
					else ERR("Attack: No target set");
				}
				
				order->setActionType("ATTACK");
				//order->setActionType("IDLE");
				
				(*it)->setOrder(order);
			}
		
			++lineNum;
			break;
			default:
				return true;
		}
		return false;
	}

	DEFINE_TASK(MoveToLocation) {
		assert(extraData != NULL);

		if (debug) std::cout << "MoveToLocation: Start" << std::endl;
		
		switch(lineNum) {
			case 0:
	// everyone should arrive at the same time
			FORALL(groups, it){
				NUS_TerrainBase::Loc startPt((*it)->getPosition().x, (*it)->getPosition().y);
				WayPointVector wayPoints;
				
				real8 length = state->simAI->planPath(startPt, extraData->pos, (*it)->getExtents(), wayPoints);
				//std::cout << "planned path" << std::endl;
				Order* order = new Order(wayPoints, (*it)->getMaxSpeed());
				/*		FORALL(wayPoints, it2){
					std::cout << it2->toString() <<"\t" ;
				}
				std::cout << std::endl;*/
				order->setActionType("MOVE");
				(*it)->setOrder(order);	
			}

			++lineNum;
			break;
			default:
			return true;
		}
		return false;
	}

// ----------------
// Hunter 

DEFINE_TASK(Hunter) {
	UnitVector units;
	
	if (debug) std::cout << "Hunter: Start" << std::endl;
	
	switch (lineNum){
					
		case 0:
			// make newGroups of X units(in this case, 5).  kind of a dumb way to do it... basically, just take one unit, find X-1
			// closest units, finding their average position, and then joining at the unit closest to that position
			
			if (debug) std::cout << "Hunter: Start Join" << std::endl;
			
			while(groups.size() > 0) {
				
				GroupVector newGroupsToJoin;
				GroupVector remainingGroups(groups);
				
				int numUnitsInGroups = groups[0]->getNumUnits();
				Vector3D leaderPosition = groups[0]->getPosition();
				
				newGroupsToJoin.push_back(groups[0]);
				remainingGroups.erase(remainingGroups.begin());
				
				while (newGroupsToJoin.size() < groups.size() && numUnitsInGroups < state->groupSize) {
					real8 nearest = 2e8;
					Group *nearestGroup = NULL;
					GroupVector::iterator nit;
					
					FORALL (remainingGroups, it) {
						real8 dist = ((*it)->getPosition() - leaderPosition).length();
						if (dist < nearest) {
							nearest = dist;
							nearestGroup = (*it);
							nit = it;
						}
					}
					
					numUnitsInGroups += nearestGroup->getNumUnits();
					newGroupsToJoin.push_back(nearestGroup);
					remainingGroups.erase(nit);
				}
				
				new JoinAndHunt(this, newGroupsToJoin, function, "JoinAndHunt", 0.0);				
			}
			
			if (debug) std::cout << "Hunter: Join End" << std::endl;

			lineNum++;
			break;
		case 1:
			if (debug) std::cout << "Hunter: Start Spread Attack" << std::endl;
			//new MassAttack(this, objects, function, name, 0.0);
			new SpreadAttack(this, groups, function, name, 0.0);

			lineNum++;
			break;
			default:
				return true;
	}
	return false;
}

DEFINE_TASK(JoinAndHunt){
	assert(extraData == NULL);

	if (debug) std::cout << "JoinAndHunt: Start" << std::endl;
	
	switch (lineNum){
		case 0:{
			if (debug) std::cout << "JoinAndHunt: Start Join" << std::endl;

			if(state->smartJoin){
				new SmartJoin(this, groups, function, name, 0.0); 
			}else{
				new Join(this, groups, function, "Join", 0.0);
			}
			lineNum++;
			break;
		}
		case 1:{		
			//note: the is only 1 group	
			if (debug) std::cout << "JoinAndHunt: Start attack" << std::endl;

			if(groups.size()!=1) ERR(("JoinAndHunt cannot have more than 1 group ??"));

			if (debug) std::cout << "JoinAndHunt: Find get nearest enemy" << std::endl;
			
			Unit* target = state->getNearestEnemyUnitOfGroup(playerID, groups[0]->getUnits());

			if (debug) std::cout << "JoinAndHunt: Done finding get nearest enemy" << std::endl;

			if (target != NULL){
				//if(state->useMoveForAttack){
					//new MoveToObject(this, newGroups, function,name,0.0, new MoveStruct(target));
				//}else{
					new Attack(this, groups, function, "Attack",0.0, new MoveStruct(target));
				//}
					} else return true;
			break;
		}
		default:
		return true;
	}
	return false;
}

// ----------------
// Hunter 2

DEFINE_TASK(Hunter2) {
	UnitVector units;
	
	if (debug) std::cout << "Hunter2: Start" << std::endl;
	
	switch (lineNum){
					
		case 0:
			// make newGroups of X units(in this case, 5).  kind of a dumb way to do it... basically, just take one unit, find X-1
			// closest units, finding their average position, and then joining at the unit closest to that position
			
			if (debug) std::cout << "Hunter2: Start Join" << std::endl;
			
			while(groups.size() > 0) {
				
				GroupVector newGroupsToJoin;
				GroupVector remainingGroups(groups);
				
				int numUnitsInGroups = groups[0]->getNumUnits();
				Vector3D leaderPosition = groups[0]->getPosition();
				
				newGroupsToJoin.push_back(groups[0]);
				remainingGroups.erase(remainingGroups.begin());
				
				while (newGroupsToJoin.size() < groups.size() && numUnitsInGroups < state->groupSize) {
					real8 nearest = 2e8;
					Group *nearestGroup = NULL;
					GroupVector::iterator nit;
					
					FORALL (remainingGroups, it) {
						real8 dist = ((*it)->getPosition() - leaderPosition).length();
						if (dist < nearest) {
							nearest = dist;
							nearestGroup = (*it);
							nit = it;
						}
					}
					
					numUnitsInGroups += nearestGroup->getNumUnits();
					newGroupsToJoin.push_back(nearestGroup);
					remainingGroups.erase(nit);
				}
				
				new JoinAndHunt2(this, newGroupsToJoin, function, "JoinAndHunt2", 0.0);
			}
			
			if (debug) std::cout << "Hunter2: Join End" << std::endl;

			lineNum++;
			break;
		case 1:
			if (debug) std::cout << "Hunter2: Start Spread Attack" << std::endl;
			//new MassAttack(this, objects, function, name, 0.0);
			new SpreadAttack(this, groups, function, name, 0.0);

			lineNum++;
			break;
			default:
				return true;
	}
	return false;
}

DEFINE_TASK(JoinAndHunt2){
	assert(extraData == NULL);

	if (debug) std::cout << "JoinAndHunt2: Start" << std::endl;
	
	switch (lineNum){
		case 0:{
			if (debug) std::cout << "JoinAndHunt2: Start Join" << std::endl;

			if(state->smartJoin){
				new SmartJoin(this, groups, function, name, 0.0);  
			}else{
				new Join(this, groups, function, "Join", 0.0);
			}
			lineNum++;
			break;
		}
		case 1:{			
			if (debug) std::cout << "JoinAndHunt2: Start attack" << std::endl;

			if(groups.size()!=1) ERR(("JoinAndHunt cannot have more than 1 group ??"));

			if (debug) std::cout << "JoinAndHunt2: Find get nearest enemy" << std::endl;

			Group *group = state->getNearestGroupOfGroup(playerID, groups[0]);
			
			if (group) {
				if ( (group->getPosition()-groups[0]->getPosition()).length() < 100.0 ) {
					if (debug) std::cout << "Nearby friendly group, let's merge!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" << std::endl;
					merge(groups[0], group);
				}
			}			
			
			Unit* target = state->getAverageNearestEnemyUnitOfGroup(playerID, groups[0]->getUnits());
			//Unit* target = state->getNearestEnemyUnitOfCGOfGroup(playerID, groups[0]->getUnits());

			if (debug) std::cout << "JoinAndHunt2: Done finding get nearest enemy" << std::endl;

			if (target != NULL){
				//if(state->useMoveForAttack){
					//new MoveToObject(this, newGroups, function,name,0.0, new MoveStruct(target));
				//}else{
					new Attack(this, groups, function, "Attack",0.0, new MoveStruct(target));
				//}
					} else return true;
			break;
		}
		default:
		return true;
	}
	return false;
}

DEFINE_TASK(JoinAndHunt3){
	assert(extraData == NULL);
	Group *group;
	if (debug) std::cout << "JoinAndHunt3: Start" << std::endl;
	//Group *group;
	
	switch (lineNum){
		case 0:{
			if (debug) std::cout << "JoinAndHunt3: Start Join" << std::endl;

			if(state->smartJoin){
				new SmartJoin(this, groups, function, name, 0.0); 
			}else{
				new Join(this, groups, function, "Join", 0.0);
			}
			lineNum++;
			break;
		}
		case 1:{		
			//note: the is only 1 group	
			if (debug) std::cout << "JoinAndHunt3: Start attack" << std::endl;

			if(groups.size()!=1) ERR(("JoinAndHunt3 cannot have more than 1 group ??"));

			if (debug) std::cout << "JoinAndHunt3: Find get nearest enemy" << std::endl;
			
			Unit* target = state->getNearestWeakEnemyUnitOfGroup(playerID, groups[0]->getUnits(), groups[0]->getNumUnits());

			if (debug) std::cout << "JoinAndHunt3: Done finding get nearest enemy" << std::endl;

			if (target != NULL){
				//if(state->useMoveForAttack){
					//new MoveToObject(this, newGroups, function,name,0.0, new MoveStruct(target));
				//}else{
					new Attack(this, groups, function, "Attack",0.0, new MoveStruct(target));
				//}
			} else{
				std::cout <<"JoinAndHunt3 no weaker enemy influence, move to the  nearest group" <<std::endl;
				Group* nearestGroup = state->getNearestGroupOfGroup(playerID, groups[0]);
				if(nearestGroup){
					NUS_TerrainBase::Loc move_loc(nearestGroup->getPosition().x, nearestGroup->getPosition().y);
					if(!groups[0]->getPosition().x == move_loc.x || !groups[0]->getPosition().y == move_loc.y  ){
						std::cout << "trying to move to another group location" << std::endl;
						new MoveToLocation(this,groups,function,name,0.0,new MoveStruct(move_loc));
						}else {
							std::cout <<"merged!!" << std::endl;
							merge(nearestGroup, groups[0]);
						}
					}else ERR("no nearest group found");
				lineNum++;
			} 
		}
			break;
			case 2:
			group = state->getNearestGroupOfGroup(playerID, groups[0]);

			if (group) {
				if ( (group->getPosition()-groups[0]->getPosition()).length() < 100.0 ) {
					if (debug) std::cout << "Nearby friendly group, let's merge!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" << std::endl;
					merge(group, groups[0]);
					}else ERR("Friendly group to join is not near");
			}
		
		default:
		return true;
	}
	return false;
}


// ----------------
// Hunter 3

DEFINE_TASK(Hunter3) {
	UnitVector units;
	
	if (debug) std::cout << "Hunter3: Start" << std::endl;
	
	switch (lineNum){
					
		case 0:
			// make newGroups of X units(in this case, 5).  kind of a dumb way to do it... basically, just take one unit, find X-1
			// closest units, finding their average position, and then joining at the unit closest to that position
			
			//add all units from all newGroups into a unit vector

			FORALL(groups, it){
				FORALL((*it)->getUnits(), it2){
					units.push_back((*it2));
				}
			}

			removeAllGroups();
			
			if (debug) std::cout << "Hunter3: Start Join" << std::endl;
			
			// make newGroups of X units(in this case, 5).  kind of a dumb way to do it... basically, just take one unit, find X-1
			// closest units, finding their average position, and then joining at the unit closest to that position
			while(units.size()>0){
				UnitVector unitsToJoin;
				state->getNearestObjectsToObject(unitsToJoin,units,units[0],state->groupSize);
				
				//create vector of new newGroups consisting of single units to join
				//add the group to Task
				GroupVector newGroupsToJoin;
				FORALL(unitsToJoin, it){
					Group* group = new Group(playerID, *it);
					addGroup(group);
					
					newGroupsToJoin.push_back(group);
				}
				
				new JoinAndHunt3(this, newGroupsToJoin, function, "JoinAndHunt3", 0.0);
				
				// remove unitToJoin from units
				FORALL(unitsToJoin, it) {
					Vector<Unit*>::iterator uit = std::find(units.begin(), units.end(), *it);
					
					if (uit != units.end())
						units.erase(uit);
				}
			}
			
			if (debug) std::cout << "Hunter3: Join End" << std::endl;

			lineNum++;
			break;
		case 1:
			if (debug) std::cout << "Hunter3: Start Spread Attack" << std::endl;
			//new MassAttack(this, objects, function, name, 0.0);
			new SpreadAttack(this, groups, function, name, 0.0);

			lineNum++;
			break;
			default:
				return true;
	}
	return false;
}

DEFINE_TASK(JoinSpreadAttack) {
	
	UnitVector units;
	
	if (debug) std::cout << "JoinSpreadAttack: Start" << std::endl;
	
	switch (lineNum){
		case 0:
			// make newGroups of X units(in this case, 5).  kind of a dumb way to do it... basically, just take one unit, find X-1
			// closest units, finding their average position, and then joining at the unit closest to that position
			
			if (debug) std::cout << "JoinSpreadAttack: Start Join" << std::endl;
			
			// make newGroups of X units(in this case, 5).  kind of a dumb way to do it... basically, just take one unit, find X-1
			// closest units, finding their average position, and then joining at the unit closest to that position
			while(groups.size() > 0) {
				
				GroupVector newGroupsToJoin;
				GroupVector remainingGroups(groups);
				
				int numUnitsInGroups = groups[0]->getNumUnits();
				Vector3D leaderPosition = groups[0]->getPosition();
				
				newGroupsToJoin.push_back(groups[0]);
				remainingGroups.erase(remainingGroups.begin());
				
				while (newGroupsToJoin.size() < groups.size() && numUnitsInGroups < state->groupSize) {
					real8 nearest = 2e8;
					Group *nearestGroup = NULL;
					GroupVector::iterator nit;
					
					FORALL (remainingGroups, it) {
						real8 dist = ((*it)->getPosition() - leaderPosition).length();
						if (dist < nearest) {
							nearest = dist;
							nearestGroup = (*it);
							nit = it;
						}
					}
					
					numUnitsInGroups += nearestGroup->getNumUnits();
					newGroupsToJoin.push_back(nearestGroup);
					remainingGroups.erase(nit);
				}
				
				new JoinAndAttack(this, newGroupsToJoin, function, "JoinAndAttack", 0.0);
			}
			
			if (debug) std::cout << "JoinSpreadAttack: Join End" << std::endl;

			lineNum++;
			break;
		case 1:
			if (debug) std::cout << "JoinSpreadAttack: Start Spread Attack" << std::endl;
			//new MassAttack(this, objects, function, name, 0.0);
			new SpreadAttack(this, groups, function, name, 0.0);

			lineNum++;
			break;
			default:
				return true;
	}
	return false;
}


/*
	DEFINE_TASK(MoveToObject) {
		assert(extraData != NULL);
		assert(extraData->obj != NULL);

		switch(lineNum) {
			case 0:
	// everyone should arrive at the same time
			FORALL(objects,it) {
				if(*it == extraData->obj){ // the center object does not need to move
					ExtraObjectData newData;
				newData.isDone = true;
				Player* currPlayer = state->getPlayer(playerID);
				currPlayer->getOrtsUnits()[*it] = newData;
				continue;
			}
			NUS_TerrainBase::Task task(NUS_TerrainBase::Goal::OBJ, NUS_TerrainBase::Goal::TOUCH,
				NUS_TerrainBase::Task::INDIVIDUAL);

			ExtraObjectData newData;
			newData.target = extraData->obj;
			newData.state = ExtraObjectData::MOVE;

			Player* currPlayer = state->getPlayer(playerID);

			currPlayer->getOrtsUnits()[*it] = newData;

			task.goal.obj = extraData->obj;
			task.objs.insert(*it);
			state->simAI->doPath(task);
		}
		++lineNum;
		break;
		default:
		return false;
	}
	return false;
}
*/
/*
DEFINE_TASK(DefendObject) {
	assert(extraData != NULL);

	switch(lineNum) {
		case 0:
	// everyone should arrive at the same time
		FORALL(objects,it){
		// this should NEVER happen, since you should not be defending an object that is getting orders
		// to defend itself!
			if(*it == extraData->obj){
				ExtraObjectData newData;
				newData.isDone = true;
				Player* currPlayer = state->getPlayer(playerID);
				currPlayer->getOrtsUnits()[*it] = newData;
				continue;
			}
			NUS_TerrainBase::Task task(NUS_TerrainBase::Goal::OBJ, NUS_TerrainBase::Goal::TOUCH,
				NUS_TerrainBase::Task::INDIVIDUAL);

			ExtraObjectData newData;
			newData.target = extraData->obj;
			newData.state = ExtraObjectData::DEFEND;

			Player* currPlayer = state->getPlayer(playerID);

			currPlayer->getOrtsUnits()[*it] = newData;

			task.goal.obj = extraData->obj;
			task.objs.insert(*it);
			state->simAI->doPath(task);
		}
		++lineNum;
		break;
		default:
		return false;
	}
	return false;
}
*/

/*
DEFINE_TASK(HalfUnitsBaseDefence_MassAttack){

	switch(lineNum) {
		case 0:{
			Player* currPlayer = state->getPlayer(playerID);
	// make sure that we have enough units to do anything.. if not, then just don't defend bases
			uint4 numUnits = currPlayer->getOrtsUnits().size();
			uint4 numBases = currPlayer->getOrtsBases().size();
			if(numUnits/2 >= numBases){
				uint4 numUnitsPerBase = (numUnits/2) / numBases;
				FORALL(currPlayer->getOrtsBases(),it){
					ObjectVector unitsToDefendBase;
					state->getNearestObjectsToObject(unitsToDefendBase,objects,it->first,numUnitsPerBase);
					new DefendObject(this,unitsToDefendBase,function,name,0.0,new MoveStruct(it->first));

				}
			}
	// the rest of the units that are not defending go join up and attack!
			new MassAttack(this,objects,function,name,0.0);
			++lineNum;
			break;
		}
		default:
		return false;
	}
	return false;
}
*/
/*
DEFINE_TASK(TouchObj) {
	assert(extraData != NULL);
	//assert(extraData->obj != NULL);

	switch(lineNum){
		case 0:
		if (extraData->obj != NULL){
			if (extraData->obj->getMaxSpeed() > 0 && *extraData->obj->sod.owner == (sint4)playerID){
				NUS_TerrainBase::Loc center_loc;
				extraData->obj->get_center(center_loc.x, center_loc.y);
				new MoveToLocation(this,objects,function,name,0.0,new MoveStruct(center_loc));
			}
			else{
				new MoveToObject(this, objects, function,name,0.0, new MoveStruct(extraData->obj));
			}
		}
		++lineNum;
		break;
		default:
		return false;
	}
	return false;
}
*/


DEFINE_TASK(JoinAndMoveToObj) {
	
	assert(extraData != NULL); // extradata should not be NULL for this task

	assert(extraData->pos != NUS_TerrainBase::Loc(-1, -1));
	UnitVector units;

	switch(lineNum) {
		case 0:{
			FORALL(groups, it){
				FORALL((*it)->getUnits(), it2){
					units.push_back((*it2));
				}
			}
			
			Unit* centerObject = state->getNearestObject(units, state->getCenterOfObjects(units));
			assert(centerObject != NULL);// takje this assert OUT later
			NUS_TerrainBase::Loc center_loc = centerObject->getPosition();
			
			GroupVector &farAwayGroups(groups);
			
		/*	FORALL(groups, it){
				NUS_TerrainBase::Loc groupLoc(sint4((*it)->getPosition().x),sint4((*it)->getPosition().y));
				
				if (sqrt((groupLoc.x-center_loc.x)*(groupLoc.x-center_loc.x) \
					+ (groupLoc.y-center_loc.y)*(groupLoc.y-center_loc.y)) > 100.0)
					farAwayGroups.push_back(*it);
			}*/
			
			
			
			if (!farAwayGroups.empty())
			{
				new MoveToLocation(this,farAwayGroups,function,name,0.0,new MoveStruct(center_loc));

				++lineNum;
				break;
				} else ++lineNum;
		}
		case 1:
		new MoveToLocation(this,groups,function,name,0.0,new MoveStruct(extraData->pos));

		++lineNum;
		break;
		default:
		return false;
	}
	return false;
}



DEFINE_TASK(SmartJoin) {
	
	assert(extraData == NULL); // extradata should be NULL for this task

	if (debug) std::cout << "SmartJoin: Start" << std::endl;

	switch(lineNum) {
		case 0:{
			UnitVector units;
			//add all units from all newGroups into a unit vector

			FORALL(groups, it){
				FORALL((*it)->getUnits(), it2){
					units.push_back((*it2));
				}
			}

			Unit* centerObject = state->getNearestObject(units, state->getCenterOfObjects(units));
			//assert(centerObject != NULL && centerObject->get_shape() == Object::CIRCLE);// takje this assert OUT later
			NUS_TerrainBase::Loc center_loc;
			//centerObject->get_center(center_loc.x, center_loc.y);
			center_loc = centerObject->getPosition();

			/*
			ignore this for now
			if(state->findspot){
				GameStateModule & gsm = state->simAI->get_gsm();
				center_loc = find_a_spot(center_loc, gsm);
			}*/
			removeAllGroups();
			
			while(!units.empty()) {
				if (debug) std::cout << "SmartJoin: start JoinAndMoveToObj" << std::endl;
				
				UnitVector unitsToJoin;

				state->getNearestObjectsToObject(unitsToJoin,units,units[0],state->smartJoinSize);
				
				GroupVector newGroupsToJoin;
				FORALL(unitsToJoin, it){
					Group* group = new Group(playerID, *it);
					addGroup(group);
					
					newGroupsToJoin.push_back(group);
				}
				
				new JoinAndMoveToObj(this,newGroupsToJoin,function,name,0.0, new MoveStruct(center_loc));
				
				// remove unitToJoin from units
				FORALL(unitsToJoin, it) {
					Vector<Unit*>::iterator uit = std::find(units.begin(), units.end(), *it);
					
					if (uit != units.end())
						units.erase(uit);
				}
			}
		}
		case 1: {
			//re-group all the newGroups into one group after joining	
			//add all units from all newGroups into a unit vector and remove the old newGroups
			UnitVector units;

			FORALL(groups, it){
				FORALL((*it)->getUnits(), it2){
					units.push_back((*it2));
				}
			}

			removeAllGroups();

			Group *group = new Group(playerID, units);
			addGroup(group);

			++lineNum;
			return true;
		}
		default:
		return true;
	}
	return false;
}


DEFINE_TASK(JoinAndAttack){
	
	assert(extraData == NULL);

	if (debug) std::cout << "JoinAndAttack: Start" << std::endl;

	switch (lineNum){
		case 0:{
			if (debug) std::cout << "JoinAndAttack: Start Join" << std::endl;

			if(state->smartJoin){
				new SmartJoin(this, groups, function, "Smart Join", 0.0);  //no smart joins for now
			}else{
				new Join(this, groups, function, "Join", 0.0);
			}
			lineNum++;
			break;
		}
		case 1:{			
			if (debug) std::cout << "JoinAndAttack: Start attack" << std::endl;
			if (debug) std::cout << "JoinAndAttack: Find get nearest enemy base" << std::endl;

			Group *group = state->getNearestGroupOfGroup(playerID, groups[0]);
			
			if (group) {
				if ( (group->getPosition()-groups[0]->getPosition()).length() < 75.0 ) {
					if (debug) std::cout << "Nearby friendly group, let's merge!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" << std::endl;
					merge(groups[0], group);
				}
			}			
			
			Base* target = state->getNearestEnemyBaseOfGroup(playerID, groups[0]);

			if (debug) std::cout << "JoinAndAttack: Done finding get nearest enemy base" << std::endl;

			if (target != NULL){
				//if(state->useMoveForAttack){
					//new MoveToObject(this, newGroups, function,name,0.0, new MoveStruct(target));
				//}else{
					new Attack(this, groups, function, "Attack",0.0, new MoveStruct(target));
				//}
					} else return true;
			break;
		}
		default:
		return true;
	}
	return false;
}


/*
	DEFINE_TASK(MassAttack) {
	// ObjectVector temp;
		switch(lineNum){

			case 0:
			if(state->smartJoin){
				new SmartJoin(this, objects, function, name, 0.0);
			}else{
				new Join(this, objects, function, name, 0.0);
			}

			++lineNum;
			break;
			case 1:{
	// need a copy of this vector, since the constructor for tasks modifies the original by removing objects from it
				GameObj* nearestBase = state->getNearestEnemyBaseOfGroup(playerID,objects);
		//temp = objects;

				if(nearestBase != NULL){
					new Attack(this,objects,function,name,0.0,new MoveStruct(nearestBase));
				}
				break;
			}
			default:
			return false;
		}
		return false;
	}
*/


//1024x768
//112
float Task::dist2dCompare(NUS_TerrainBase::Loc a, NUS_TerrainBase::Loc b)
{
	return (a.x-b.x)*(a.x-b.x) + (a.y-b.y)*(a.y-b.y);
}

float Task::dist2dCompare(float a,float b,float x,float y)
{
	return (a-x)*(a-x) + (b-y)*(b-y);
}

float Task::dist2d(NUS_TerrainBase::Loc a, NUS_TerrainBase::Loc b)
{
	return sqrt((a.x-b.x)*(a.x-b.x) + (a.y-b.y)*(a.y-b.y));
}  

NUS_Simul::Unit* Task::GetNearestEnemy(NUS_TerrainBase::Loc pos)
{
	float nearest =  10000000;
	Unit* enemyunit = 0;
	
	FORALL(state->getPlayers(),it) 
	{
		if(it->first == playerID) continue; // don't check our own units
		FORALL((it->second)->getOrtsUnits(),it2)
		{
			Unit* unit = it2->first;
			if (unit->getObject()->component("weapon") == NULL) continue;
			float dist = dist2dCompare(pos, unit->getPosition());
			if (dist<nearest)
			{
				nearest=dist;
				enemyunit = unit;
			}
			
			
		}			
	}
	
	return enemyunit;	
}

Vector3D Task::ToVector3D(NUS_TerrainBase::Loc pos)
{
	Vector3D position;
	position.x=pos.x; position.y=pos.y;
	return 	position;
}

NUS_TerrainBase::Loc Task::ToTerrainBase(Vector3D vec3)
{
	NUS_TerrainBase::Loc position;
	position.x=vec3.x; position.y=vec3.y;
	return 	position;
} 

bool Task::OffMapLimits(Vector3D destination)
{
	if (destination.x>1014) {return true;}
	if (destination.x<10) {return true;}
	if (destination.y>758) {return true;}
	if (destination.y<10) {return true;}
	return false;
}

bool Task::TestPtClear(float x, float y, Unit* ourunit)
{
			float dismissalradius = 9.5f*9.5f;
			//float dismissalradius = 12.5f*12.5f;
			FORALL(state->getPlayers(),it) 
			{
				FORALL((it->second)->getOrtsUnits(),it2)
				{
					Unit* otherunit = it2->first;
					//if (otherunit->getObject()->component("weapon") == NULL) dismissalradius=10.0f*10.0f;
					if (ourunit->getObject()!=otherunit->getObject())
					{
						NUS_TerrainBase::Loc posTB = otherunit->getPosition(); // TerrainBase::Loc
						float a= posTB.x;
						float b= posTB.y;
						if (dist2dCompare(x,y,a,b)<dismissalradius)
						{
							return false;
						}
					}
					//if (otherunit->getObject()->component("weapon") == NULL) dismissalradius=8.0f*8.0f;
				}
			}
			return true;
}

//bool Task::Prod(float x, float y, float heading, float speed, Unit* ourunit)
//{
	//for (int i=0; i<
//}



int Task::turnangle (float curDirection, float desDirection, float turnspeed)
{
	
		int dir, turndirection;
		float turnwidth;  //dir : 1= anticlockwise, dir: 0= clockwise
		
		turndirection=3;

		turnwidth = desDirection - curDirection;
		dir = 1;
		if ( turnwidth < 0) 
		{
			turnwidth = curDirection - desDirection; 
			dir = 0;
		}
		
	  if (turnwidth < (turnspeed * 2.0f))  
		{
			turndirection = 3;
		}
	  else
		{
				

				
		  if (dir==1) 
		  {
			if ((6.28f-turnwidth)< turnwidth)  
				{
					turndirection = -1 ;
				}
			else
				{
					turndirection = 1 ;
				}
		  }


		  if (dir==0) 
		  {
			if ((6.28f-turnwidth)< turnwidth)  
				{
					turndirection = 1;
				}
			else
				{
					turndirection = -1;
				}
		  }

		
		}
	  
	  
	  return turndirection;

}

Vector3D Task::MTFA(float x, float y, float heading, float speed, Unit* ourunit)
{
	/*
	float dir=1;
	int leftcount, rightcount;
	leftcount=rightcount=0;
			FORALL(state->getPlayers(),it) 
			{
				FORALL((it->second)->getOrtsUnits(),it2)
				{
					Unit* otherunit = it2->first;
					//if (otherunit->getObject()->component("weapon") == NULL) dismissalradius=10.0f*10.0f;
					if (ourunit->getObject()!=otherunit->getObject())
					{
						float angle = angfrom(x,y, otherunit->getPosition().x, otherunit->getPosition().y);
						if (turnangle(heading,angle,0.0001f)==-1) leftcount+=1; else rightcount+=1; 		 
					}
				}
			}
			if (leftcount>rightcount) dir=-1;
	*/
	for (float i=0.03f; i<2.8f; i+=0.03f)
	{
		Vector3D testpt;
		testpt.x = x+sinf(heading+i)*speed;
		testpt.y = y+cosf(heading+i)*speed;
		testpt.z =0;
		
		if (TestPtClear(testpt.x, testpt.y, ourunit)==true) return testpt;
		
		testpt.x = x+sinf(heading-i)*speed;
		testpt.y = y+cosf(heading-i)*speed;
		testpt.z =0;
		
		if (TestPtClear(testpt.x, testpt.y, ourunit)==true) return testpt;
	} 
	return Vector3D(x,y,0);
}



void Task::MoveTo(Group* unit, Vector3D destination, float recommendedspeed)
{
	Vector3D newpos = destination;
	Vector3D direction = (newpos - unit->getPosition()).normalized();
	Vector3D hundredforwardpos = unit->getPosition() + direction * 100.0f; 
	bool slide=OffMapLimits(hundredforwardpos);
	
	
	
	
	
	
	if (slide==true)
	{
		Vector3D newpos = destination;
		float angle = angfrom(unit->getPosition().x, unit->getPosition().y, destination.x, destination.y);
		newpos.x = unit->getPosition().x+sinf(angle+3.14f)*100.0f;
		newpos.y = unit->getPosition().y+cosf(angle+3.14f)*100.0f;
		for (float i=0.03f; i<3.14f; i+=0.1f)
		{
			
			newpos.x = unit->getPosition().x+sinf(angle+i)*100.0f;
			newpos.y = unit->getPosition().y+cosf(angle+i)*100.0f;
			if (OffMapLimits(newpos)==false) 
			{
				break;
			}
			newpos.x = unit->getPosition().x+sinf(angle-i)*100.0f;
			newpos.y = unit->getPosition().y+cosf(angle-i)*100.0f;
			if (OffMapLimits(newpos)==false) 
			{
				break;
			}
		}
	}
	
	
	if (newpos.x>1019) {newpos.x=1019;}
	if (newpos.x<5) {newpos.x=5;}
	if (newpos.y>763) {newpos.y=763;}
	if (newpos.y<5) {newpos.y=5;}
	
	
	
	if (unit->getOrder()) unit->getOrder()->executed = false;
	
	
	
	Unit* ourunit = unit->getUnits()[0];
	Vector3D ourunitpos = ToVector3D(ourunit->getPosition());
	
	float MaxSpeed = (float) ourunit->getObject()->get_int("max_speed");
	Vector3D NextPos = ourunitpos+ (newpos-ourunitpos).normalized()*MaxSpeed;
	
	/*
			FORALL(state->getPlayers(),it) 
			{
				FORALL((it->second)->getOrtsUnits(),it2)
				{
					Unit* otherunit = it2->first;
					if (ourunit->getObject()!=otherunit->getObject())
					{
						NUS_TerrainBase::Loc posTB = otherunit->getPosition(); // TerrainBase::Loc
						Vector3D otherunitpos = ToVector3D(posTB);
						if ((otherunitpos-NextPos).length()<8.1f)
						{
							//std::cout<<"collllide "<<(otherunitpos-NextPos).length();
							Vector3D tootherunit = otherunitpos-ourunitpos;
							tootherunit.z =0;
							tootherunit = tootherunit.normalized();
							Vector3D direction = newpos-ourunitpos;
							direction.z =0;
							direction = direction.normalized();
							
							if (direction.dot(tootherunit)>0)
							{
								newpos.x = ourunitpos.x - tootherunit.y*10.0f;
								newpos.y = ourunitpos.y + tootherunit.x*10.0f;
							}
							else
							{
								newpos.x = ourunitpos.x + tootherunit.y*10.0f;
								newpos.y = ourunitpos.y - tootherunit.x*10.0f;
							}
						}
					}
				}			
			}
			*/
	if (TestPtClear(NextPos.x, NextPos.y,ourunit)==false)
	{
		newpos = MTFA(ourunitpos.x, ourunitpos.y, angfrom(ourunitpos.x, ourunitpos.y, NextPos.x, NextPos.y), MaxSpeed, ourunit);
	}
			
	//new MoveToLocation(this,myGroupVector,function,name,0.0,new MoveStruct(ToTerrainBase(newpos)));
	// Assume group of 1 unit
	/*NUS_TerrainBase::Loc startPt(unit->getPosition().x, unit->getPosition().y);
	WayPointVector wayPoints;
	
	real8 length = state->simAI->planPath(startPt, NUS_TerrainBase::Loc(newpos.x,newpos.y), unit->getExtents(), wayPoints);
	Order* order = new Order(wayPoints, unit->getMaxSpeed());
	order->setActionType("MOVE");
	unit->setOrder(order);
	std::cout << "trying to move" << std::endl;
	*/
	    Vector<sint4> params;
        params.push_back(newpos.x);
        params.push_back(newpos.y);
        GameObj *gob = unit->getUnits()[0]->getObject();
        params.push_back(*gob->sod.max_speed*recommendedspeed);

        gob->set_action("move", params);
	
} 

float Task::angfrom (float x1, float y1, float x2, float y2)   
{     
	float hdist, vdist, angle;
	hdist = (x1 - x2);
	vdist = (y1 - y2);

	


    if (vdist == 0.0f )
	{
        if (hdist < 0.0f)  {angle = 4.71238898f;} else {angle = 1.570796327f;}
	}   
    else
	{
		angle = (float) atan(hdist / vdist) + 3.141592654f;		
		if (vdist > 0.0f) {angle = angle + 3.14592654f;}
	}
	

	angle = angle + 3.14592654f;
	if (angle > 6.283185307f) {angle = angle - 6.283185307f;}
	if (angle < 0.0f) {angle = angle + 6.283185307f;}
	return angle;

}

struct TravoUnit
{
	Vector3D vec;
};

void Task::Push(Group* me, Vector3D dest)
{
	Vector3D mypos = me->getPosition();
	Vector3D newpos = mypos + ((dest-mypos).normalized()*2.0f);
	
	float heading = angfrom(mypos.x, mypos.y, dest.x, dest.y);
	
	GroupVector temp(groups);
	FORALL(temp, i)
	{
		Group* unit = *i;
		if (unit!=me)
		{
			if (dist2dCompare(mypos.x, mypos.y, unit->getPosition().x, unit->getPosition().y)<(30.0f*30.0f))
			{
					Vector3D offset = newpos-unit->getPosition();
					float angler = angfrom(mypos.x, mypos.y, unit->getPosition().x, unit->getPosition().y); 
					if (angdist(heading, angler)<0.175f)
					{
						unit->pushed=true;
						MoveTo(unit, unit->getPosition()-(offset*2.0f),1.0f);
					}
			}
			
		}
	}
	
	/*
	GroupVector temp(groups);
	FORALL(temp, i)
	{
		Group* unit = *i;
		if (unit!=me)
		{
			if (unit->pushed==false)
			{
				Vector3D offset = newpos-unit->getPosition();
				if (offset.length2()<(10.0f*10.0f))
				{
					unit->pushed=true;
					unit->vec = offset * 10.0f;
					MoveTo(unit, unit->getPosition()-(offset*2.0f),1.0f);
					Push(unit,  unit->getPosition()-offset);
				}
			}
		}
	}*/
}

float Task::angdist (float curDirection, float desDirection)
{
	
		int dir;
		float turnwidth;  //dir : 1= anticlockwise, dir: 0= clockwise
		
		turnwidth = desDirection - curDirection;
		
		dir = 1;
		if ( turnwidth < 0) 
		{
			turnwidth = curDirection - desDirection; 
			dir = 0;
		}
		
	  if (turnwidth ==0)  
		{
			return 0;
		}
	  else
		{
				

				
		  if ((dir==1) && ((6.28f-turnwidth)< turnwidth))  
				{
					turnwidth = (6.28f-turnwidth);
				}
		  


		  if ((dir==0) && ((6.28f-turnwidth)< turnwidth))
				{
					turnwidth = (6.28f-turnwidth);
				}
		  

		
		}
	  
	  
	  return turnwidth;

}

int Task::HowManyInFrontOfMe(Group* me, Vector3D dest, float angwidth)
{
	int count=0;
	Vector3D mypos = me->getPosition();
	float heading = angfrom(mypos.x, mypos.y, dest.x, dest.y);
	
	GroupVector temp(groups);
	FORALL(temp, i)
	{
		Group* unit = *i;
		if (unit!=me)
		{
			if (dist2dCompare(mypos.x, mypos.y, unit->getPosition().x, unit->getPosition().y)<(16.0f*16.0f))
			{
					float angler = angfrom(mypos.x, mypos.y, unit->getPosition().x, unit->getPosition().y); 
					if (angdist(heading, angler)<angwidth)
					{
						count++;
					}
			}
			
		}
	}
	return count;
}

float Task::HowMuchFriendlyInfluence(Vector3D pos, Vector3D centerpos)
{
	float totaldistance=0;
	GroupVector temp(groups);
	FORALL(temp, i)
	{
		Group* unit = *i;
		float disttome = dist2dCompare(centerpos.x, centerpos.y, unit->getPosition().x, unit->getPosition().y);
		float dist = dist2dCompare(pos.x, pos.y, unit->getPosition().x, unit->getPosition().y);
		if (disttome<(70.0f*70.0f)) totaldistance+=dist;
	}
	return totaldistance;
}

static int TravoState=0;

static int Timer=0;

DEFINE_TASK(TravoAttack) {
	UnitVector units;
	
	
	
	if (debug) std::cout << "TravoAttack: Start" << std::endl;
	
	Timer++;
	//std::cout<<Timer<<std::endl;

			GroupVector temp(groups);
			
			int count=0;
			Vector3D avgpos=Vector3D(0,0,0);
			int maxhp=0;
			FORALL(temp, i)
			{
				Group* group = *i;
				Vector3D grouppos = group->getPosition();
				avgpos+=grouppos;
				count++;
				
				//group->vec=Vector3D(0,0,0);
				group->pushed=false;
				if (group->getUnits()[0]->getHitPoints()>maxhp)
				{
					maxhp = group->getUnits()[0]->getHitPoints();
				}
			}
			avgpos/=count;
			float maxdevfromavgpos = 0;
			
			Vector3D THEENEMYPOS = Vector3D(0,0,0);
			if (TravoState==1)
			{
				Unit* enemy = GetNearestEnemy(ToTerrainBase(avgpos));
				if (enemy!=0)
				{
					THEENEMYPOS = ToVector3D(enemy->getPosition());
				}	
			}
			
			//avg dist2enemy
			float avgdist2enemy=0;
			float mindist2enemy=100000000.0f;
			count=0;
			FORALL(temp, i)
			{
				Group* group = *i;
				Vector3D grouppos = group->getPosition();
				Unit* enemy = GetNearestEnemy(ToTerrainBase(avgpos));
				Vector3D enemypos = Vector3D(0,0,0);
				if (enemy!=0)
				{
					enemypos = ToVector3D(enemy->getPosition());
					float dist= (grouppos-enemypos).length();
					if (dist<mindist2enemy) mindist2enemy = dist;
					avgdist2enemy+=dist;
				}
				
				count++;
			}
			avgdist2enemy/=count;
			if (debug) std::cout<<"We have " <<count <<" units remaining"<<std::endl; 
			
			
			
			
			
			FORALL(temp, i)
			{
				Group* group = *i;
				
				
				Vector3D grouppos = group->getPosition();
				
				
				
				float devfromavgpos=(grouppos-avgpos).length();
				if (devfromavgpos>maxdevfromavgpos)
				{
					maxdevfromavgpos=devfromavgpos;
				}
				
				
				Unit* enemy =  GetNearestEnemy(ToTerrainBase(group->getPosition()));
				
				if (group->pushed==true) continue;
				
				if (enemy!=0)
				{
					int mylife = group->getUnits()[0]->getHitPoints();
					int enemylife = enemy->getHitPoints();
					Vector3D enemypos = ToVector3D(enemy->getPosition())+ enemy->velocity();
					Vector3D offset = enemypos - group->getPosition();
					float withdrawdist = 66 + ((maxhp-mylife)/4.0f);
					//if (group->getUnits()[0]->getHitPoints()<80) withdrawdist=100;
					
					float distance = offset.length();
					
					if (distance<withdrawdist)
					{
						group->pushed=true;
						MoveTo(group, group->getPosition() - offset, 1.0f);
						Push(group, group->getPosition() - offset); //MINUS
					} 
					else
					{
						float speed = 1.0f;
						
						if (distance>90.0f)
						{
							//if (distance<(avgdist2enemy-4.0f)) speed=0.8f;
							if (distance<=(mindist2enemy+4.0f)) speed=0.4f;
							if (distance<=(mindist2enemy)) speed=0.3f;
						}
						else if (distance>69.0f)
						{
							//Vector3D offset = enemypos - group->getPosition();
							//float offsetdist = offset.length();
							//if (distance<(avgdist2enemy)) speed=0.7f;
							if (distance<=(mindist2enemy+1.0f)) speed=0.8f;
							if (distance<=(mindist2enemy)) speed=0.7f;
							
							if ((distance+2.9f)<(enemypos+ enemy->velocity()- group->getPosition()).length()) speed = 1.0f;
							
							
							//if (offset.length()<(avgdist2enemy-5.0f)) speed=0.7f;
							//if (offset.length()<(avgdist2enemy-10.0f)) speed=0.6f;
						}
						else 
						{
							speed = 0.5f;
						}
						
						/*
						else if (distance>70.0f)
						{
							float localavgdist2enemy=0;
							float localmindist2enemy=100000000.0f;
							count=0;
							FORALL(temp, id)
							{
								Group* friendly = *id;
								if (friendly!=group)
								{
									if ((group->getPosition()-friendly->getPosition()).length2()<(50.0f*50.0f))
									{
										Vector3D grouppos = friendly->getPosition();
										Unit* enemy = GetNearestEnemy(ToTerrainBase(avgpos));
										Vector3D enemypos = Vector3D(0,0,0);
										if (enemy!=0)
										{
											enemypos = ToVector3D(enemy->getPosition());
											float dist= (grouppos-enemypos).length();
											if (dist<localmindist2enemy) localmindist2enemy = dist;
											localavgdist2enemy+=dist;
										}
										count++;
									}
								}
								
								
							}
							localavgdist2enemy/=count;
							
							
							
							if (distance<(localavgdist2enemy-4.0f)) speed=0.9f;
							if (distance<=(localmindist2enemy)) speed=0.8f;
							
						}*/
						
						
						
						if (HowManyInFrontOfMe(group, enemypos, 0.4f)<=0)
						{
							MoveTo(group, enemypos, speed);
						}
						else
						{
							
							float enemyangle = angfrom(enemypos.x, enemypos.y,group->getPosition().x, group->getPosition().y);
							Vector3D leftpos;
							leftpos.x = enemypos.x + sinf(enemyangle+0.1f)*(distance-4.0f);
							leftpos.y = enemypos.y + cosf(enemyangle+0.1f)*(distance-4.0f);
							leftpos.z=0;
							float leftinfluence = HowMuchFriendlyInfluence(leftpos,group->getPosition() );
							
							Vector3D rightpos;
							rightpos.x = enemypos.x + sinf(enemyangle-0.1f)*(distance-4.0f);
							rightpos.y = enemypos.y + cosf(enemyangle-0.1f)*(distance-4.0f);
							rightpos.z=0;
							float rightinfluence = HowMuchFriendlyInfluence(rightpos,group->getPosition() );
							
							if (leftinfluence>(rightinfluence+1000.0f)) MoveTo(group, leftpos, 1.0f); else MoveTo(group, rightpos, 1.0f);
							//slide around
						}
					}
				}
				
				/*
				group->vec*=0.9f;
				if (group->vec.length2()<0.1f)
				{ 
					group->vec= Vector3D(0,0,0);
				}
				else
				{
					MoveTo(group, group->getPosition()-group->vec, 1.0f);
				}
				*/

				
				
				
				
				//MoveTo(group, ToVector3D(enemy->getPosition()));
				
				switch (TravoState)
				{
					case 0: //gather all together
					{
						if (group->pushed==false) MoveTo(group, avgpos, 1.0f);
						//if (group->pushed==false) MoveTo(group, Vector3D(500,500,0));
						
						break;
					}
					
					case 1:
					{
						/*
						Unit* enemy = GetNearestEnemy(ToTerrainBase(group->getPosition()));
						if (enemy!=0)
						{
							Vector3D enemypos = ToVector3D(enemy->getPosition());
							MoveTo(group, enemypos);
						}
						break;
						*/
					}
				}
				
				
				/*
				
				
				
				
				FORALL(state->getPlayers(),it) 
				{
					if(it->first == playerID) continue;
					
					FORALL((it->second)->getOrtsUnits(),it2)
					{
						
						Unit* enemyunit = it2->first;
						if (enemyunit->getObject()->component("weapon") == NULL) continue;
						Vector3D enemypos = ToVector3D(enemyunit->getPosition());
						Vector3D dist = enemypos-grouppos; 
						
						
						if (enemyunit->getHitPoints()<80)
						{
							if (dist.length()<200.0f) 
							{
								MoveTo(group, grouppos-dist);
							}	
						}
						if (dist.length()<130.0f) 
						{
							MoveTo(group, grouppos-dist);
						}
						
						
					}
				}
				*/

			}
			
			if (Timer>170)
			{
				TravoState=1;
			}
			if (maxdevfromavgpos<40.0f)
			{
				TravoState=1;
			}
			/*
			std::cout<<(grouppos-avgpos).length()<<" " <<std::endl;
				if ((grouppos-avgpos).length()>40.0f)
				{
					TravoState=0;
				}
				else
				{
					TravoState=1;
				}
			*/

			
	
	
	return false;
	
}

Task* createTravoAttack(uint4 id,SimulationState* s,GroupVector& o,real8 timeout) {
	return new TravoAttack(id,s,o,createTravoAttack,"TravoAttack",timeout);
}



Task* createSmartJoin(uint4 id,SimulationState* s,GroupVector& o,real8 timeout) {
	return new SmartJoin(id,s,o,createSmartJoin,"SmartJoin",timeout);
}
Task* createSpreadAttack(uint4 id,SimulationState* s,GroupVector& o,real8 timeout) {
	return new SpreadAttack(id,s,o,createSpreadAttack,"SpreadAttack",timeout);
}
Task* createMassAttack(uint4 id,SimulationState* s,GroupVector& o,real8 timeout) {
	return NULL; //return new MassAttack(id,s,o,createMassAttack,"MassAttack",timeout);
}
Task* createHalfUnitsBaseDefence_MassAttack(uint4 id,SimulationState* s, GroupVector& o,real8 timeout) {
	return NULL; //return new HalfUnitsBaseDefence_MassAttack(id,s,o,createHalfUnitsBaseDefence_MassAttack,"HalfUnitsBaseDefence_MassAttack",timeout);
}
Task* createHunter(uint4 id,SimulationState* s, GroupVector& o,real8 timeout) {
	return new Hunter(id,s,o,createHunter,"Hunter",timeout);
}
Task* createHunter2(uint4 id,SimulationState* s, GroupVector& o,real8 timeout) {
	return new Hunter2(id,s,o,createHunter2,"Hunter2",timeout);
}

Task* createHunter3(uint4 id,SimulationState* s, GroupVector& o,real8 timeout) {
	return new Hunter3(id,s,o,createHunter2,"Hunter3",timeout);
}

Task* createJoinSpreadAttack(uint4 id,SimulationState* s, GroupVector& o,real8 timeout) {
	return new JoinSpreadAttack(id,s,o,createJoinSpreadAttack,"JoinSpreadAttack",timeout);
}

/*********************JAMES' STUFF **************************/  
NUS_TerrainBase::Loc find_a_spot(NUS_TerrainBase::Loc center_loc, GameStateModule& gsm, SimulationState *state){

	mapTerrain::TerrainMap terrMap(&gsm.get_game(), state);
	NUS_TerrainBase::Loc loc(0,0);
	sint4 standard = 9; sint4 range = 128;
	bool found = terrMap.find_free_loc(center_loc, loc, range, standard);
	range+= 32;
	//std::cout << "found = " << found << ", and meeting spot is (" <<  center_loc.x / 16 << "," << center_loc.y / 16 << ")" << std::endl;
	while(!found  && standard > 4){
		found = terrMap.find_free_loc(center_loc, loc, range, standard--);
	//std::cout << "found = " << found << ", and meeting spot is (" <<  center_loc.x / 16 << "," << center_loc.y / 16 << ")" << std::endl;
	}
	terrMap.print_map();
	return NUS_TerrainBase::Loc(loc.x, loc.y);
}
/*********************JAMES' STUFF **************************/ 

}



