
#include "NUS_SimulationAI.H"

#include "NUS_SimulationState.h"
#include "NUS_Player.h"

#include "GameStateModule.H"
#include "Game.H"
#include "GameChanges.H"
//#include "GfxModule.H"
#include "NUS_TerrainModule.H"
#include "NUS_SquadCombatAI.H"

#include "NUS_Player.h"
#include "Random.H"
#include "Options.H"

#include <math.h>

using namespace std;

static const bool debug = false;

namespace NUS_Simul {

	const sint4 SimulationAI::FROM = EventFactory::new_who();

	void SimulationAI::add_options()
	{
		::Options o("SimulationAI");
		o.put("-hunter4", "Hunter 4");
		o.put("-hunter2","Use Hunter, with target being closest to ALL our units in group");
		o.put("-spread", "every unit continuously attacks nearest base to it");
		o.put("-halfdef", "Use task that send half units to mass and attack, while the other half defends bases spread out evenly");
		o.put("-joinspread","Join into small groups, which then attack nearest bases");
		o.put("-groupsize",sint4(50),"Specify how large groups are(used by hunter currently(default is use all 50 units)");
		o.put("-smartjoinsize",sint4(5),"specify how large local groups are formed before doing a larger join");
		o.put("-smartjoin", "Specify whether to form local groups when joining, or whether to just go and join");
		o.put("-moveforattack","When told to attack a unit, instead of using an attack task, use a move to location task(means shooterAI is not called, and there is less replanning(used by hunter1/2 right now)");
		o.put("-findspot","finds a smarter spot to do joins than just the average location of the units");
		o.put("-findbeststrategy","finds best strategy");
		o.put("-searchStrategy", "search for best strategy");
		o.put("-waypoint", "stop and wait for group member at waypoint");
		o.put("-victoryType", sint4(0), "Victory type - 0 for all bases destroyed, 1 for all units destroyed");
		o.put("-influence", "use influence map");
		o.put("-hunter3", "Use Hunter, with target being closest enemy unit to any unit in group and has weaker influence");
		o.put("-timeoutJoin",real8(0.0), "timeout for joins");
		o.put("-game4", "hack to improve performance of game 4");
		o.put("-moveVicinty", sint4(-1), "distance for move vicinty");
		o.put("-attackMargin", real8(50.0), "attack margin when game4 option is set");
		o.put("-slowgrind", "slow grind movement");
		o.put("-kiasee", "afraid to die");
	}

	SimulationAI::SimulationAI(GameStateModule &gsm_, Terrain &timp_, sint4 seed) : gsm(gsm_), timp(timp_){
		simulationState = new SimulationState(this);
	// crashes here for some reason, need to put it later
	//simulationState->setOurID(gsm.get_game().get_client_player());
		bestStrategy = NULL;
		lastSearchTime = 0.0;
		
		simulationResult = NULL;
		
		rand.srand(seed);
		
		firstInit = false;
	}
	
	SimulationAI::~SimulationAI(){
		if (simulationResult) delete simulationResult;
		FORS (i, width) 
	      delete [] map[i];
	    delete [] map;
	}

	void SimulationAI::initObstacles() {
		Game &game = gsm.get_game();
		game_map = &(game.get_map()); 
		
	    width = game_map->get_width();
	    height = game_map->get_height();
	    tile_points = game.get_tile_points();

	    //Create a 2D array of the given sizes.
	    map = new Tile*[width];
	    FORS(i, width)
	      map[i] = new Tile[height];

	    FORS(i, width) {
	      FORS(j, height) {
		//Step through all the tiles and reproduce a simplified version here
		GameTile *temp = &game_map->get_tile(i,j);
		sint4 height = temp->get_max_h();
		map[i][j].x = i;
		map[i][j].y = j;
		map[i][j].freeValue = 0;

		//Tile type is cliff or water
		if(height != 0 || temp->get_typeW() == 1 || temp->get_typeW() == 3 ){
		  map[i][j].status = Tile::OBSTACLE;
		}
		//UNKNOWN Tile type -> behind FOW
		else if(temp->get_typeW() == 0) 
		  map[i][j].status = Tile::FOG;
		//Tile type is free
		else if(temp->get_typeW() == 2) 
		  map[i][j].status = Tile::FREE;
		else
		  ERR("Unknown tile type!!");
	      }
	    }
		/*
	    FORU (i, width) {
	      map[i][0].status = Tile::OBSTACLE;
	      map[i][height-1].status = Tile::OBSTACLE;
	    }

	    FORU (i, height) {
	      map[0][i].status = Tile::OBSTACLE;
	      map[width-1][i].status = Tile::OBSTACLE;
	    }
		*/
	}
	
	void SimulationAI::addBuilding(GameObj *obj) {
	      sint4 topLeftx, topLefty;
	      sint4 bottomRightx, bottomRighty;

	      if(!obj) return;

		if (debug) std::cout << "adding a building" << std::endl;
		
	      obj->get_p1(topLeftx, topLefty);
	      obj->get_p2(bottomRightx, bottomRighty);

	      sint4 tilei = topLeftx / tile_points;
	      sint4 tilej = topLefty / tile_points;

	      sint4 endi = bottomRightx / tile_points;
	      sint4 endj = bottomRighty / tile_points;

	      tilei = max(tilei, 0);
	      tilej = max(tilej, 0);
	      endi = min(endi, width-1);
	      endj = min(endj, height-1);

	      for(int j = tilei; j <= endi; j++){
			for(int k = tilej; k <= endj; k++){
		  		map[j][k].status = Tile::OBSTACLE;
			}
	      }		
	}
	
	void SimulationAI::removeBuilding(GameObj *obj) {
	      sint4 topLeftx, topLefty;
	      sint4 bottomRightx, bottomRighty;

	      if(!obj) return;

		if (debug) std::cout << "removing a building" << std::endl;

	      obj->get_p1(topLeftx, topLefty);
	      obj->get_p2(bottomRightx, bottomRighty);

	      sint4 tilei = topLeftx / tile_points;
	      sint4 tilej = topLefty / tile_points;

	      sint4 endi = bottomRightx / tile_points;
	      sint4 endj = bottomRighty / tile_points;

	      tilei = max(tilei, 0);
	      tilej = max(tilej, 0);
	      endi = min(endi, width-1);
	      endj = min(endj, height-1);

	      for(int j = tilei; j <= endi; j++){
			for(int k = tilej; k <= endj; k++){
		  		map[j][k].status = Tile::FREE;
			}
	      }		
	}
	
	// Rotate point angle radians around relPoint
	void SimulationAI::rotate(real8 &nx, real8 &ny, NUS_TerrainBase::Loc pos, real8 vecX, real8 vecY, real8 angle) {
		real8 sinAngle = sin(angle);
		real8 cosAngle = cos(angle);
		
		vecX += pos.x;
		vecY += pos.y;
		
		nx = vecX*cosAngle - vecY*sinAngle - pos.x*cosAngle + pos.y*sinAngle;
		ny =vecX*sinAngle + vecY*cosAngle - pos.x*sinAngle - pos.y*cosAngle;
	}

//===================================================================

	void SimulationAI::act(GameStateModule &/*gsm*/){
		fprintf(stderr,"u...");
		simulationState->updateStrategies();
		getMe()->doOrders(this);
	}

//===================================================================

/**
	* Returns true if event was handled.
*/
		bool SimulationAI::handle_event(const Event &e)
		{
			//if (debug) std::cout <<  "Handling of event " << e.get_what() << std::endl;
			bool needStrategyUpdate = false;

			if (e.get_what() == NUS_TerrainModule::PATH_SUCCESS) {
				//if (debug) std::cout <<  "Handling PATH_SUCCESS" << std::endl;

				const NUS_TerrainModule::PathStatusEvent *p =
				dynamic_cast<const NUS_TerrainModule::PathStatusEvent*>(&e);
				assert(p && "SUCCESS: not a PathStatusEvent");

				FORALL(simulationState->getPlayers(),it){
					
					std::map<GameObj*, Unit*>::iterator it2 = (it->second)->getObjectUnitMap().find(p->get_obj());
					if (it2 == (it->second)->getObjectUnitMap().end()) continue;

						// if we were just moving... then we are done!
						/*if(it->second.state == ExtraObjectData::MOVE){
					it2->second.isDone = true;
						//needStrategyUpdate = true;
					}*/

						//we want to do update path here
						//find the group that this unit belongs to
					Unit *unit = (it->second)->getObjectUnitMap()[p->get_obj()];
					(it->second)->getOrtsUnits()[unit].isDone = true;
					(it->second)->getOrtsUnits()[unit].isExecuting = false;

					Group* objInGroup = (it->second)->getUnitGroupMap()[unit];
					needStrategyUpdate = updatePath(objInGroup);
					break;
				}
				
				//if (debug) std::cout <<  "End Handling PATH_SUCCESS" << std::endl;
			}

				if (e.get_what() == NUS_TerrainModule::PATH_FAILURE) 
				{
					//if (debug) std::cout <<  "Handling PATH_FAILURE" << std::endl;

					const NUS_TerrainModule::PathStatusEvent *p =
						dynamic_cast<const NUS_TerrainModule::PathStatusEvent*>(&e);
					assert(p && "FAILURE: not a PathStatusEvent");

					//if (p->get_details() == NUS_TerrainModule::PATH_PLAN_FAILURE){
					if (debug) std::cout <<  "Path failure! This is bad, and shouldn't happen often" << endl;

					FORALL(simulationState->getPlayers(),it){

						std::map<GameObj*, Unit*>::iterator it2 = (it->second)->getObjectUnitMap().find(p->get_obj());
						if (it2 == (it->second)->getObjectUnitMap().end()) continue;

						/*	Unit *unit = (it->second)->getObjectUnitMap()[p->get_obj()];
							
							(it->second)->getStrategy()->removeUnit(unit);*/
						
						// TODO: right now, just say we are done if path fails.. not sure what to do here

						/*
						if(it2->second.hasReplanned){
							it2->second.isDone = true;
							needStrategyUpdate=true;
							continue;
						}

						if(it2->second.target == NULL){
							if(it2->second.location.x != -1){
								NUS_TerrainBase::Task task(NUS_TerrainBase::Goal::LOCATION, NUS_TerrainBase::Goal::TOUCH,
									NUS_TerrainBase::Task::INDIVIDUAL);



								task.goal.loc = it2->second.location;
								task.objs.insert(p->get_obj());
								doPath(task);
								it2->second.hasReplanned = true;

							}else{
								ERR("somethign is wrong, we should have either a target, or location");
							}
						}else{
							NUS_TerrainBase::Task task(NUS_TerrainBase::Goal::OBJ, NUS_TerrainBase::Goal::TOUCH,
								NUS_TerrainBase::Task::INDIVIDUAL);

							task.goal.obj = it2->second.target;
							task.objs.insert(p->get_obj());
							doPath(task);
							it2->second.hasReplanned = true;
							}*/
								
							Unit *unit = (it->second)->getObjectUnitMap()[p->get_obj()];
							//(it->second)->getOrtsUnits()[unit].isDone = true;

							Group *group = (it->second)->getUnitGroupMap()[unit];
							
							if (false)
							//if ((it->second)->getOrtsUnits()[unit].hasReplanned) 
							{
								UnitVector unitVector(group->getUnits());
								
								FORALL(unitVector,uit) (it->second)->getOrtsUnits()[*uit].isDone = true;
												
								Order *order = group->getOrder();
								
								if (order) order->setCompleted(true);

								group->hurry = true;
								
								//stopPath(getMe(), unitVector);
								//replan(unitVector);
								//(it->second)->getOrtsUnits()[unit].isExecuting = false;
							}
							else 
							{
								UnitVector unitVector;				
								//unitVector.clear();
							
								unitVector.push_back(unit);
							
								stopPath(getMe(), unitVector);

								replan(unitVector);
								
								(it->second)->getOrtsUnits()[unit].isDone = true;
							}

							bool allDone = true;

							UnitVector unitVector(group->getUnits());
							
							FORALL(unitVector,uit) {
								if (!(it->second)->getOrtsUnits()[unit].isDone) {
									allDone = false;
									break;
								}
							}
							
							if (allDone) {
								Order *order = group->getOrder();
								
								if (order) order->setCompleted(true);
							}
							
							needStrategyUpdate=true;

							(it->second)->getOrtsUnits()[unit].hasReplanned = true;
							
							//Group* objInGroup = (it->second)->getUnitGroupMap()[unit];
							//updatePath(objInGroup);							
						}
						
						//if (debug) std::cout <<  "End Handling PATH_FAILURE" << std::endl;
					}
				

					if (e.get_what() == NUS_SquadCombatAI::TARGET_DESTROYED) {
						if (debug) std::cout <<  "Handling NUS_SquadCombatAI::TARGET_DESTROYED" << std::endl;

						const NUS_SquadCombatEvent *s = dynamic_cast<const NUS_SquadCombatEvent*>(&e);
						assert(s && "FAILURE: not a NUS_SquadCombatEvent");

						FORALL (s->units, u) {
							Player *player = simulationState->getPlayer(*(*u)->sod.owner);

							std::map<GameObj*, Unit*>::iterator it = player->getObjectUnitMap().find(*u);
							if (it == player->getObjectUnitMap().end()) continue;

							Unit *unit = player->getObjectUnitMap()[*u];
							player->getOrtsUnits()[unit].isDone = true;
							needStrategyUpdate = true;
						}

						stopCombat(s->units);
						if (debug) std::cout <<  "End Handling NUS_SquadCombatAI::TARGET_DESTROYED" << std::endl;
					}						

					bool mode;
					if (e.get_who() == GameStateModule::FROM) {

						if (e.get_what() == GameStateModule::VIEW_MSG) {
							//if (debug) std::cout <<  "Handling VIEW_MSG" << std::endl;

							const sint4 me = gsm.get_game().get_client_player();

						// this is a dumb place to put it, but it crashes if it is put into the constuctor
							simulationState->setOurID(gsm.get_game().get_client_player());

							if (!firstInit) {
								initObstacles();
								firstInit = true;
							}
							
						// add new objects
						// NOTE: we assume this is only called once right now, since we see the whole map
						// this WILL NOT work for fog of war!

							bool addedUnit = false;

							FORALL (gsm.get_changes().new_objs, i) {
								GameObj *gob = (*i)->get_GameObj();
								if (gob == 0) continue;

								Player* player = simulationState->getPlayer(*gob->sod.owner);
								if(player == NULL){ // player does not exist yet.. create one
									
									bool useInfluence;
									::Options::get("-influence", useInfluence);
									
									if(useInfluence){
										sint4 tile_points = gsm.get_game().get_tile_points();

									  	sint4 width = gsm.get_game().get_map().get_width() * tile_points;
									  	sint4 height = gsm.get_game().get_map().get_height() * tile_points;
										player = new Player(*gob->sod.owner, width, height);
									}
									else{
										player = new Player(*gob->sod.owner);
									}
										
									simulationState->addPlayer(player);

									sint4 victoryType;
									::Options::get("-victoryType", victoryType);
								
									player->setVictoryType(victoryType);
								}

								if(isControllable(gob)){
									player->addOrtsUnit(gob);
									addedUnit = true;
								}else if(gob->bp_name() == "controlCenter"){
									player->addOrtsBase(gob);
									addBuilding(gob);
								}
								else if (!isControllable(gob)) sheep.push_back(gob);
							}
							
							Player* ourPlayer = simulationState->getPlayer(simulationState->getOurID());

							if(addedUnit){				
								if(ourPlayer == NULL) ERR("umm, weirdness, our player should exist always");

								if(ourPlayer->getStrategy() != NULL){
									ERR("GOT NEW OBJECTS IN MULTIPLE FRAMES! CAN'T HANDLE THIS RIGHT NOW!");
								}

								GroupVector ourControllableObjects = ourPlayer->getGroups();

							/*FORALL(ourPlayer->getOrtsUnits(),it){
								ourControllableObjects.push_back(it->first);
								}*/

								if(!ourControllableObjects.empty()){
									sint4 gSize,smartSize;
									bool useSmart = false;
									bool useMove = false;
									bool useFindSpot = false;
									real8 timeoutJoin = 0.0;
									::Options::get("-groupsize", gSize);
									::Options::get("-smartjoinsize", smartSize);
									::Options::get("-smartjoin", useSmart);
									::Options::get("-moveforattack", useMove);
									::Options::get("-findspot", useFindSpot);
									::Options::get("-timeoutJoin", timeoutJoin);
									
									simulationState->groupSize = (uint4)gSize;
									simulationState->smartJoinSize = (uint4)smartSize;
									simulationState->smartJoin = useSmart;
									simulationState->useMoveForAttack = useMove;
									simulationState->findspot = useFindSpot;
									simulationState->timeoutJoin = timeoutJoin;
									
									::Options::get("-searchStrategy", mode);
									if (mode)
									{
									// FIXME: Change strategy sometimes?
										if (bestStrategy) {
											ourPlayer->setStrategy(bestStrategy(me, simulationState, ourControllableObjects, 0.0));
										} 
										else
										{
											cout << "starting search for best strategy" << endl;

											SimulationResult simResult;

											simResult.isDone = false;
											simResult.ourStrategyIndex = 1;
											simResult.theirStrategyIndex = 1;
											simResult.bestValue = -2e30;
											simResult.currentValue = 0.0;
											simResult.bestStrategyIndex = 0;

											lastSearchTime = getCurrentTime();

											bestStrategy = simulationState->searchStrategy(me, getCurrentTime(), 10000.0, 1000.0, simResult);

											if (bestStrategy)
											{
												cout << "found a best strategy" << endl;
												ourPlayer->setStrategy(bestStrategy(me, simulationState, ourControllableObjects, 0.0));
											}
											else
											{
												// set a default?
												cout << "failed to find a best strategy" << endl;
											}
										}						
									}
									else
									{
										::Options::get("-findbeststrategy", mode);
										if (mode)
										{
											if (simulationResult == NULL) {
												simulationResult = new SimulationResult;
												
												simulationResult->isDone = false;
												simulationResult->ourStrategyIndex = 1;
												simulationResult->theirStrategyIndex = 1;
												simulationResult->bestValue = -2e30;
												simulationResult->currentValue = 0.0;
												simulationResult->bestStrategyIndex = 0;
											}
											
											cout << "starting search for best strategy" << endl;
											
											// lazy! :
											SimulationResult &simResult(*simulationResult);

											bestStrategy = simulationState->findBestStrategy(me, getCurrentTime(), 0.5, 10.0, simResult);

											if (simResult.isDone)
											{
												cout << "found a best strategy" << endl;
												ourPlayer->setStrategy(bestStrategy(me, simulationState, ourControllableObjects, 0.0));
												needStrategyUpdate = true;
												delete simulationResult;
												simulationResult = NULL;
											}
											else
											{
												//needStrategyUpdate = true;
												cout << "failed to find a best strategy" << endl;
											}
											
										}
										else
										{
											::Options::get("-hunter4", mode);
											if (mode)
												ourPlayer->setStrategy(createHunter4(me, simulationState, ourControllableObjects, 0.0));
											else{
												::Options::get("-hunter2", mode);
												if (mode)
													ourPlayer->setStrategy(createHunter2(me, simulationState, ourControllableObjects, 0.0));
												else{
													::Options::get("-spread", mode);
													if (mode)
														ourPlayer->setStrategy(createSpreadAttack(me, simulationState, ourControllableObjects, 0.0));
													else{
														::Options::get("-halfdef", mode);
														if (mode)
															ourPlayer->setStrategy(createHalfUnitsBaseDefence_MassAttack(me, simulationState, ourControllableObjects, 0.0));
														else{
															::Options::get("-joinspread", mode);
															if (mode)
																ourPlayer->setStrategy(createJoinSpreadAttack(me, simulationState, ourControllableObjects, 0.0));
															else{
																::Options::get("-hunter3", mode);
																if (mode)
																	ourPlayer->setStrategy(createHunter3(me, simulationState, ourControllableObjects, 0.0));
																else{
																	ourPlayer->setStrategy(createMassAttack(me, simulationState, ourControllableObjects, 0.0));
																}
															}
														}
													}
												}
											}
										}
									}
								}


								needStrategyUpdate = true;
							}

							// remove dead objects
							//if (debug) std::cout <<  "Number of dead objs:" << gsm.get_changes().dead_objs.size() << std::endl;
							
							FORALL (gsm.get_changes().dead_objs, i) {
								GameObj *gob = (*i)->get_GameObj();
								if (gob == 0) continue;
								
								if (debug) std::cout <<  "Something just died!" << std::endl;
								
								Player* player = simulationState->getPlayer(*gob->sod.owner);
								if(player == NULL) ERR("problem, we have an object that does not belong to a player we know about");

								// if it's our unit, we have to make sure our current task also loses it
								needStrategyUpdate = informObjectsOfDeath(gob);
								if (debug) std::cout <<  "informed objects of death" << std::endl;

								player->removeObject(gob);
								if (debug) std::cout <<  "player has handled death" << std::endl;								
								needStrategyUpdate=true;				
							}

							
							FORALL (gsm.get_changes().changed_objs, it) {
								GameObj *gob = (*it)->get_GameObj();
								if (gob == NULL) continue;
								
								simulationState->getPlayer(*gob->sod.owner)->updateObject(gob);
								/*
								if (simulationState->getPlayer(*gob->sod.owner) != ourPlayer) {
									FORALL(ourPlayer->getGroups(), git) {
										Order *order = (*git)->getOrder();
										
										if (order && order->getTargetObject() == *it) {
											// has it moved far??
											if (!order->getWayPoints().empty()) {
												Vector3D dest = order->getFinalDestination();
											
												if (sqrt( (dest.x-(*it)->get_int("x"))*(dest.x-(*it)->get_int("x")) \
												+ (dest.y-(*it)->get_int("y"))*(dest.y-(*it)->get_int("y"))) > 50.0 ) 
												{
													UnitVector &unitVector((*git)->getUnits());
													order->setCompleted(true);
													FORALL(unitVector,uit) ourPlayer->getOrtsUnits()[*uit].isDone = true;
													needStrategyUpdate=true;
													//replan(unitVector);
												}
											}
										}
									}
								}
								*/
								
								// REPLANS IF TARGET HAS MOVED -- DOES NOT WORK WELL!
								// check if it's a target of one of our groups
								if (simulationState->getPlayer(*gob->sod.owner) != ourPlayer) {
									FORALL(ourPlayer->getGroups(), git) {
										Order *order = (*git)->getOrder();
										
										if (order && order->getTargetObject() == *it) {
											// has it moved far??
											if (!order->getWayPoints().empty()) {
												Vector3D dest = order->getFinalDestination();
											
												if (sqrt( (dest.x-(*it)->get_int("x"))*(dest.x-(*it)->get_int("x")) \
												+ (dest.y-(*it)->get_int("y"))*(dest.y-(*it)->get_int("y"))) >= (*git)->getMaxAttackRange()*1.0 ) 
												{
													UnitVector &unitVector((*git)->getUnits());
														
													FORALL(unitVector,uit) ourPlayer->getOrtsUnits()[*uit].isDone = true;
													//FORALL(unitVector,uit) timp.update_obj((*uit)->getObject());	
													needStrategyUpdate=true;							
													order->setCompleted(true);
													//replan(unitVector);
												}
											}
										}
									}
								}
							}
							
								
							////recheck influence here ?
							::Options::get("-hunter3", mode);
							if(mode){
								
								FORALL(ourPlayer->getGroups(), git){
									Order* order = (*git)->getOrder();
									if(order && !order->isCompleted() && order->getActionType()=="ATTACK"){
										GameObj* target = order->getTargetObject();
										NUS_TerrainBase::Loc targetLoc(target->get_int("x"), target->get_int("y"));
										//std::cout <<"enemy influence " << simulationState->getEnemyInfluence(getMe()->getID(), targetLoc) << "our strength" <<  (*git)->getNumUnits() << std::endl;
										if(simulationState->getEnemyInfluence(getMe()->getID(), targetLoc) > (*git)->getNumUnits()){
											std::cout <<"we are chionging into danger!!! stop!!! " << std::endl;
											//stopPath(getMe(), (*git)->getUnits());

											//needStrategyUpdate=true;							
											//order->setCompleted(true);
										}
									}
								}
							}

							/* Do we need to replan? */
							::Options::get("-searchStrategy", mode);
							if ( mode && getCurrentTime() - lastSearchTime > 10.0 )
							{
								cout << "starting search for best strategy" << endl;

								const sint4 me = simulationState->getOurID();

								SimulationResult simResult;

								simResult.isDone = false;
								simResult.ourStrategyIndex = 1;
								simResult.theirStrategyIndex = 1;
								simResult.bestValue = -2e30;
								simResult.currentValue = 0.0;
								simResult.bestStrategyIndex = 0;

								lastSearchTime = getCurrentTime();

								bestStrategy = simulationState->searchStrategy(me, getCurrentTime(), 10000.0, 1000.0, simResult);

								if (bestStrategy)
								{
									cout << "found a best strategy" << endl;

									Player* ourPlayer = simulationState->getPlayer(me);
									if(ourPlayer == NULL) ERR("umm, weirdness, our player should exist always");

									GroupVector ourControllableObjects = ourPlayer->getGroups();

									ourPlayer->setStrategy(bestStrategy(me, simulationState, ourControllableObjects, 0.0));
								}
								else
								{
									needStrategyUpdate = true;
									cout << "failed to find a best strategy" << endl;
								}
							}	

							  Game &game = gsm.get_game();	
						      sint4 vf = game.get_view_frame();
						      sint4 af = game.get_action_frame();
						      //sint4 sa = game.get_skipped_actions();
						      //sint4 ma = game.get_merged_actions();

							::Options::get("-findbeststrategy", mode);
							if (mode)
							{								
								if (vf !=0 && simulationResult && !simulationResult->isDone) {
									cout << "continuing search for best strategy" << endl;

									const sint4 me = gsm.get_game().get_client_player();

									//Player* ourPlayer = simulationState->getPlayer(simulationState->getOurID());
									//GroupVector ourControllableObjects = ourPlayer->getGroups();

									// lazy! :
									SimulationResult &simResult(*simulationResult);

									bestStrategy = simulationState->findBestStrategy(me, getCurrentTime(), 0.5, 10.0, simResult);

									if (simResult.isDone)
									{
										cout << "found a best strategy" << endl;
										//ourPlayer->setStrategy(bestStrategy(me, simulationState, ourControllableObjects, 0.0));
										//delete simulationResult;
										//simulationResult = NULL;
										//needStrategyUpdate = true;
									}
									else
									{
										cout << "failed to find a best strategy" << endl;
									}
								}
							}							
							
							if (mode && simulationResult && simulationResult->isDone)
							{
								if (abs(vf-af) < 5) {
									GroupVector ourControllableObjects = ourPlayer->getGroups();
									bestStrategy = strategyList[simulationResult->bestStrategyIndex];
									ourPlayer->setStrategy(bestStrategy(me, simulationState, ourControllableObjects, 0.0));
									delete simulationResult;
									simulationResult = NULL;
									needStrategyUpdate = true;
								}
							}
													
							bool mode;
							::Options::get("-kiasee", mode);
							if (mode) {
								Player *me = getMe();

								// initialized to zero
								FORALL (getMe()->getOrtsUnits(), it) {
									Unit *unit = it->first;
									unit->steering.x = unit->steering.y = 0;
									unit->pushed = false;
									
									if (!getMe()->getOrtsUnits()[unit].isExecuting) 
									{
										UnitVector goUnits;
										goUnits.push_back(unit);
										replan(goUnits);											
									}
								}
																
								FORALL (getMe()->getOrtsUnits(), it) {
									Unit *unit = it->first;
									/*
									if (getMe()->getOrtsUnits()[unit].isExecuting) 
									{
											UnitVector stopUnits;
											stopUnits.push_back(unit);
											stopPath(getMe(), stopUnits);											
									}
										
									MoveTo(unit, Vector3D(width*tile_points/2, height*tile_points/2, 0), 1.0f);
									
									continue;
									*/
									Group *minGroup = NULL;
									real8 minDist = 2e8;

									FORALL(simulationState->getPlayers(),it3) {
										if((it3->second) == me) continue; 
										FORALL((it3->second)->getGroups(),it4){
											NUS_TerrainBase::Loc loc((*it4)->getPosition().x, (*it4)->getPosition().y);

											real8 dist = unit->distance_to(loc);
											if (dist<minDist) {
												minGroup = *it4;
												minDist = dist;
											}
										}
									}

									if (minGroup == NULL) break;
									
									NUS_TerrainBase::Loc nearestEnemyGroupLoc(minGroup->getPosition().x, minGroup->getPosition().y);

									real8 unitDistance = unit->distance_to(nearestEnemyGroupLoc);
									
									real8 distanceFactor = 0.99;
									/*
									bool useInfluence;
									::Options::get("-influence", useInfluence);
									
									if (useInfluence) {
										real4 enemyInfluence = simulationState->getEnemyInfluence(getMe()->getID(), nearestEnemyGroupLoc);
										real4 myInfluence = simulationState->getFriendlyInfluence(getMe()->getID(), unit->getPosition());
										
										real8 diff = enemyInfluence - myInfluence;
										
										if (diff > 5.0) 
											distanceFactor = 1.5;
										else if (diff > 2.0)
											distanceFactor = 1.2;
										else if (diff >= 0.0)
											distanceFactor = 0.99;
										else if (diff > -2 )
											distanceFactor = 0.75;
										else 
											distanceFactor = 0.5;
										
										// if we are quite big, then engage
										if (myInfluence > 10.0) 
											if (distanceFactor > 0.98) distanceFactor = 0.98;
											
										// if we are just waiting for reinforcements
										if (getMe()->getUnitGroupMap()[unit]->getNumUnits() - enemyInfluence >= -1.0)
											if (distanceFactor > 0.98) distanceFactor = 0.98;
									}*/

									if (unitDistance <= unit->getObject()->component("weapon")->get_int("max_ground_range")*distanceFactor) {
										if (getMe()->getOrtsUnits()[unit].isExecuting) 
										{
											UnitVector stopUnits;
											stopUnits.push_back(unit);
											stopPath(getMe(), stopUnits);											
										}
										
										GameObj *gob = unit->getObject();
										Vector<sint4> params;
										
										sint4 vecAwayX = unit->getPosition().x - nearestEnemyGroupLoc.x;
										sint4 vecAwayY = unit->getPosition().y - nearestEnemyGroupLoc.y;

										Group *myGroup = me->getUnitGroupMap()[unit];
										
										sint4 vecCenterX = myGroup->getPosition().x - unit->getPosition().x;
										sint4 vecCenterY = myGroup->getPosition().y - unit->getPosition().y;
										
										// check for obstacles and stuff
										//std::cout << "STEERING AWAY!" << std::endl;
										//steerAway(unit, (vecAwayX+vecCenterX)/2, (vecAwayY+vecCenterY)/2);			
										//steerAway(unit, vecAwayX, vecAwayY);			
										MoveTo(unit, Vector3D(unit->getPosition().x+vecAwayX, unit->getPosition().y+vecAwayY, 0), 1.0f);
									}
					
								}
								
								FORALL (getMe()->getOrtsUnits(), it) {
									Unit *unit = it->first;

									if (unit->steering.x == 0 && unit->steering.y ==0) continue;
									
									if (getMe()->getOrtsUnits()[unit].isExecuting) 
									{
										UnitVector stopUnits;
										stopUnits.push_back(unit);
										stopPath(getMe(), stopUnits);											
									}
									
									Vector<sint4> params;
									
									NUS_TerrainBase::Loc loc(unit->getPosition());
									
									if (debug && (loc.x != unit->getObject()->get_int("x") || loc.y != unit->getObject()->get_int("y"))) ERR("Unit location not accurate");
									
									params.push_back(loc.x+unit->steering.x);
									params.push_back(loc.y+unit->steering.y);

									params.push_back(*unit->getObject()->sod.max_speed);

									unit->getObject()->set_action("move", params);
								}
							}
							
							/// hack for game 4
							::Options::get("-slowgrind", mode);
							if(mode){
								FORALL(getMe()->getGroups(), it){
									if((*it)->getOrder() && (*it)->getOrder()->executed && (*it)->getOrder()->getActionType()=="ATTACK" && !(*it)->getOrder()->isCompleted()){
										
										if ((*it)->hurry) {
											UnitVector goUnits;
											FORALL((*it)->getUnits(), it2){
												if (!getMe()->getOrtsUnits()[*it2].isExecuting) 
													goUnits.push_back(*it2);
											}
											replan(goUnits);
											continue;
										}

										//check if closet unit and farthest unit to target unit is not too far..
										real8 maxDistance = 0;
										real8 minDistance = 2e8;
										Unit *minUnit, *maxUnit;
										GameObj* target =  (*it)->getOrder()->getTargetObject();
										FORALL((*it)->getUnits(), it2){
											if((*it2)->distance_to(target)>maxDistance){
												maxUnit = *it2;
												maxDistance = (*it2)->distance_to(target);
											}
											/*
											if((*it2)->distance_to(target)<minDistance){
												minUnit = *it2;
												minDistance = (*it2)->distance_to(target);
											}
											*/
										}
										
										Player *me = getMe();
										//std::cout << "distance is " << maxDistance - minDistance << std::endl;

										UnitVector stopUnits;
										UnitVector goUnits;
										UnitVector regroupUnits;
										
										real8 attackMargin;
										::Options::get("-attackMargin", attackMargin);
										
										NUS_TerrainBase::Loc groupPosition((*it)->getPosition().x, (*it)->getPosition().y);
										FORALL((*it)->getUnits(), it2){
											if ((*it2)->getObject()->is_dead()) continue;
											//this unit is too close, stop it
											
											if( ( maxDistance - (*it2)->distance_to(target) > attackMargin*2) ) {
												if (!getMe()->getOrtsUnits()[*it2].isExecuting)
													regroupUnits.push_back(*it2);
												else
													stopUnits.push_back(*it2);
											}
											else
											if( maxDistance - (*it2)->distance_to(target) > attackMargin && \
											(*it2)->distance_to(groupPosition) < attackMargin*2){
												if (getMe()->getOrtsUnits()[*it2].isExecuting) 
													stopUnits.push_back(*it2);
											}
											else
											if( maxDistance - (*it2)->distance_to(target) <= attackMargin){
												if (!getMe()->getOrtsUnits()[*it2].isExecuting) 
													goUnits.push_back(*it2);
											}
										}
										stopPath(getMe(), stopUnits);
										replan(goUnits);
										//stopPath(getMe(), regroupUnits);
										//if (!regroupUnits.empty())
										//	doPath(getMe(),regroupUnits,(*it)->getPosition(), (*it)->getExtents());
										//if (!stopUnits.empty()) 
										//	doPath(getMe(), stopUnits, (*it)->getPosition(), 0);	
										/*if (!regroupUnits.empty()) {
											FORALL(regroupUnits,uit) {
												GameObj *gob = (*uit)->getObject();
												Vector<sint4> params;
												params.push_back(groupPosition.x);
												params.push_back(groupPosition.y);
												params.push_back(*gob->sod.max_speed);

												gob->set_action("move", params);											
											}
										}*/
									}
								}
							}
							
							//if (debug) std::cout <<  "End Handling VIEW_MSG" << std::endl;
						}
					}


	// TODO for now, leave this check out.  it just results in slightly more work.  if i leave the check in
	// it seems to fail sometimes... not sure why.
	if(needStrategyUpdate){
		//if (debug) std::cout <<  "start needStrategyUpdate" << std::endl;
		act(gsm);
		//if (debug) std::cout <<  "Done needStrategyUpdate" << std::endl;
		//if (debug) std::cout <<  "Done handling event of type " << e.get_what() << std::endl;
		return true;
	}
	//if (debug) std::cout <<  "Done handling event of type " << e.get_what() << std::endl;
	return false;
	}

int SimulationAI::vectorCollisions(NUS_TerrainBase::Loc pos, sint4 radius, real8 vecX, real8 vecY) {
	bool friendlyCollision = false;
	
	const uint4 maxSteps = 8;
	
	Player *me = getMe();

	sint4 oi = pos.x / tile_points;
    sint4 oj = pos.y / tile_points;
		
	sint4 tilei = pos.x / tile_points;
    sint4 tilej = pos.y / tile_points;

	bool valid = ((tilei >= 0 && tilei < width && tilej >= 0 && tilej < height && map[tilei][tilej].status == Tile::FREE));
	
	sint4 px;
	sint4 py;
	
	for (uint4 i = 0; i < maxSteps; ++i) {
		px = (pos.x + round(vecX / real8(maxSteps) * (i+1)));
		py = (pos.y + round(vecY / real8(maxSteps) * (i+1)));
		NUS_TerrainBase::Loc loc(px,py);
		
		if ( sqrt((pos.x-px)*(pos.x-px) + (pos.y-py)*(pos.y-py)) < radius ) continue;

		if (px < 0 || px > width * tile_points || py < 0 || py > height * tile_points) return -1;
		
		tilei = px / tile_points;
	    tilej = py / tile_points;

		if (!valid && tilei == oi && tilej == oj) continue;
		
		// check if it collides with non-mobile stuff
		//if ((tilei >= 0 && tilei < width && tilej >= 0 && tilej < height && map[tilei][tilej].status == Tile::FREE))
		if (!besideObstacle(loc,radius))
		{
			// check sheep
			FORALL(sheep,it) {
				sint4 sx = (*it)->get_int("x");//*(*it)->sod.x;//(*it)->get_int("x");
				sint4 sy = (*it)->get_int("y");//*(*it)->sod.y;//(*it)->get_int("y");
				sint4 sr = *(*it)->sod.radius; //(*it)->get_int("radius");
				
				if ( sqrt((sx-px)*(sx-px) + (sy-py)*(sy-py)) < 2.1*(sr+radius)) {
					//if (debug) std::cout << "collides with sheep" << std::endl;
					return -2; // collides with a sheep
				}
			}
			
			// check our friends 		
			FORALL (me->getOrtsUnits(), it) {
				if (friendlyCollision) break;
				
				Unit *unit = it->first;
				
				if (unit->getObject()->is_dead()) continue;
				
				if ( unit->distance_to(loc) < radius * 2.1 ) friendlyCollision = true; // collides with a friendly
			}
		}
		else 
		{
			//if (debug) std::cout << "collides with wall" << std::endl;
			return -1; // hits stationary or out of map
		}
	}

	if ( sqrt((pos.x-px)*(pos.x-px) + (pos.y-py)*(pos.y-py)) < radius ) return -1;
	if (!valid && tilei == oi && tilej == oj) return -1;
	
	if (friendlyCollision)
		return -3;
	else
		return 1; // no collisions
}

bool SimulationAI::besideObstacle(NUS_TerrainBase::Loc pos, sint4 radius) {
	sint4 tilei = pos.x / tile_points;
    sint4 tilej = pos.y / tile_points;
	const real8 fudgeFactor = 1.2;
	
	// check if it collides with non-mobile stuff
	if (tilei < 0 || tilei >= width || tilej < 0 || tilej >= height || map[tilei][tilej].status != Tile::FREE)
		return true;
		
	tilei = (pos.x+fudgeFactor*radius) / tile_points;
    tilej = pos.y / tile_points;

	// check if it collides with non-mobile stuff
	if (tilei < 0 || tilei >= width || tilej < 0 || tilej >= height || map[tilei][tilej].status != Tile::FREE)
		return true;

	tilei = (pos.x-fudgeFactor*radius) / tile_points;
    tilej = pos.y / tile_points;

	// check if it collides with non-mobile stuff
	if (tilei < 0 || tilei >= width || tilej < 0 || tilej >= height || map[tilei][tilej].status != Tile::FREE)
		return true;

	tilei = pos.x / tile_points;
    tilej = (pos.y-fudgeFactor*radius) / tile_points;

	// check if it collides with non-mobile stuff
	if (tilei < 0 || tilei >= width || tilej < 0 || tilej >= height || map[tilei][tilej].status != Tile::FREE)
		return true;

	tilei = pos.x / tile_points;
    tilej = (pos.y+fudgeFactor*radius) / tile_points;

	// check if it collides with non-mobile stuff
	if (tilei < 0 || tilei >= width || tilej < 0 || tilej >= height || map[tilei][tilej].status != Tile::FREE)
		return true;

	tilei = (pos.x+fudgeFactor*radius) / tile_points;
    tilej = (pos.y+fudgeFactor*radius) / tile_points;

	// check if it collides with non-mobile stuff
	if (tilei < 0 || tilei >= width || tilej < 0 || tilej >= height || map[tilei][tilej].status != Tile::FREE)
		return true;

	tilei = (pos.x-fudgeFactor*radius) / tile_points;
    tilej = (pos.y+fudgeFactor*radius) / tile_points;

	// check if it collides with non-mobile stuff
	if (tilei < 0 || tilei >= width || tilej < 0 || tilej >= height || map[tilei][tilej].status != Tile::FREE)
		return true;

	tilei = (pos.x+fudgeFactor*radius) / tile_points;
    tilej = (pos.y-fudgeFactor*radius) / tile_points;

	// check if it collides with non-mobile stuff
	if (tilei < 0 || tilei >= width || tilej < 0 || tilej >= height || map[tilei][tilej].status != Tile::FREE)
		return true;

	tilei = (pos.x-fudgeFactor*radius) / tile_points;
    tilej = (pos.y-fudgeFactor*radius) / tile_points;

	// check if it collides with non-mobile stuff
	if (tilei < 0 || tilei >= width || tilej < 0 || tilej >= height || map[tilei][tilej].status != Tile::FREE)
		return true;
		
	return false;
}

int SimulationAI::unitCollisions(NUS_TerrainBase::Loc pos, sint4 speed, sint4 radius, real8 vecX, real8 vecY) {
	bool friendlyCollision = false;
	
	const real8 rotationStep = 3.1416 / 30.0;
	//const real8 rotationStep = radius / ( 3 * real8(speed) );

	real8 rayx = vecX;
	real8 rayy = vecY;
	int collisions = vectorCollisions(pos, radius, rayx, rayy);

	if (collisions == 1 || collisions == -3) {
		//if (debug) std::cout << "pass stage 1" << std::endl;

		if (collisions == -3) friendlyCollision = true;

		// if i am beside an obstacle or boundary, then don't whisker
		if (besideObstacle(pos, radius))
			if (friendlyCollision) 
				return -3;
			else
				return 1;
			
		rotate(rayx, rayy, pos, vecX, vecY, rotationStep);
		collisions = vectorCollisions(pos, radius, rayx, rayy);
		
		if (collisions == 1 || collisions == -3) {
			//if (debug) std::cout << "pass stage 2" << std::endl;

			if (collisions == -3) friendlyCollision = true;

			rotate(rayx, rayy, pos, vecX, vecY, -rotationStep);
			collisions = vectorCollisions(pos, radius, rayx, rayy);
			
			if (collisions == 1 || collisions == -3)
			{
				//if (debug) std::cout << "pass stage 3" << std::endl;
				if (collisions == -3) friendlyCollision = true;
				
				if (friendlyCollision) 
					return -3;
				else
					return 1;
			}
		}		
	}
	
	return -1;
}

void SimulationAI::pushFriendly(Unit *unit, real8 vecX, real8 vecY) {	
	sint4 radius = unit->getObject()->get_int("radius");

	const real8 rotationStep = 3.1416 / 30.0;
	//const real8 rotationStep = radius / ( 3 * real8(unit->getObject()->get_int("max_speed")) );

	real8 rayx = vecX;
	real8 rayy = vecY;
	NUS_TerrainBase::Loc pos(unit->getPosition());

	pushFriendly(unit, pos, radius, rayx, rayy);
	
	rotate(rayx, rayy, pos, vecX, vecY, rotationStep);
	pushFriendly(unit, pos, radius, rayx, rayy);

	rotate(rayx, rayy, pos, vecX, vecY, -rotationStep);
	pushFriendly(unit, pos, radius, rayx, rayy);
}

void SimulationAI::pushFriendly(Unit *myUnit, NUS_TerrainBase::Loc pos, sint4 radius, real8 vecX, real8 vecY) {
	const uint4 maxSteps = 4;
	
	Player *me = getMe();
	
	for (uint4 i = 0; i < maxSteps; ++i) {
		sint4 px = (pos.x + round(vecX / real8(maxSteps) * (i+1)));
		sint4 py = (pos.y + round(vecY / real8(maxSteps) * (i+1)));

		NUS_TerrainBase::Loc loc(px,py);

		// check our friends 		
		FORALL (me->getOrtsUnits(), it) {				
			Unit *unit = it->first;
				
			if (myUnit == unit) continue;
			if (unit->pushed) continue;
			if (unit->getObject()->is_dead()) continue;

			if ( unit->distance_to(loc) < radius * 2.1 ) {
				unit->pushed = true;
				steerAway(unit, vecX, vecY);
			}
		}
	}	
}

void SimulationAI::steerAway(Unit *unit, sint4 vecX, sint4 vecY) {
	GameObj *gob = unit->getObject();
	
	sint4 radius = gob->get_int("radius");
	
	NUS_TerrainBase::Loc loc(unit->getPosition());
	
	if (debug && loc.x != gob->get_int("x")) ERR("Unit loc not accurate!");
	if (debug && loc.y != gob->get_int("y")) ERR("Unit loc not accurate!");
	
	// normalize?
	real8 norm = sqrt( (vecX*vecX) + (vecY*vecY) );
	real8 rvecX = vecX * 3.0 / norm * gob->get_int("max_speed");
	real8 rvecY = vecY * 3.0 / norm * gob->get_int("max_speed");
	//real8 rvecX = vecX;
	//real8 rvecY = vecY;
	//vecX = ceil(vecX / norm);
	//vecY = ceil(vecY / norm); 
	//std::cout << "vector (" << vecX << "," << vecY << ")" << std::endl;
	
	const int maxSteps = 32;
	const real8 rotationStep = 3.14159 / (real8(maxSteps));
	
	int step = 0;
	
	real8 vx = rvecX;
	real8 vy = rvecY;
	
	real8 rayx, rayy;
	
	real8 wx, wy;
	int working = -1;
	
	while (true) {
		if (step == maxSteps) {
			vx = rvecX;
			vy = rvecY;
			std::cout << "cannot find good ray" << std::endl;
			//return;
			break;
		}
		
		rotate(rayx, rayy, loc, rvecX, rvecY, rotationStep * step);
		
		//std::cout << "loc:" << loc.x << "," << loc.y << std::endl;
		//std::cout << "vector (" << rvecX << "," << rvecY << ") - rotate " << rotationStep * step << "-->" << "(" << rayx << "," << rayy << ")" << std::endl;

		int collisions = unitCollisions(loc, gob->get_int("max_speed"), radius, rayx, rayy);
		
		if (collisions == 1 || collisions == -3) {
			// no collisions
			vx = rayx;
			vy = rayy;
			
			if (collisions == -3) {
				//ERR("only friendly disturbing us!");
				//std::cout << "only friendly disturbing us!" << std::endl;
				pushFriendly(unit, vx, vy);				
			}			
			break;
			
			if (step != 0) {
				wx = vx;
				wy = vy;
				if (collisions == 1) working = 1; else working = 2;
			}
			else {
				if (collisions == -3) {
					//ERR("only friendly disturbing us!");
					//std::cout << "only friendly disturbing us!" << std::endl;
					pushFriendly(unit, vx, vy);				
				}			
				break;
			}
		}
	
		if (step == 0 || step == maxSteps - 1) {
			step++; continue;
		}
		
		rotate(rayx, rayy, loc, rvecX, rvecY, -rotationStep * step);
		collisions = unitCollisions(loc, gob->get_int("max_speed"), radius, rayx, rayy);
		
		if (collisions == 1 || collisions == -3) {
			// no collisions
			vx = rayx;
			vy = rayy;
			
			if (working) {
				
				NUS_TerrainBase::Loc testloc(unit->getPosition());
				
				testloc.x += vecX;
				testloc.y += vecY;
				
				real8 dist1 = sqrt( (wx-testloc.x)*(wx-testloc.x) + (wy-testloc.y)*(wy-testloc.y) );
				real8 dist2 = sqrt( (vx-testloc.x)*(vx-testloc.x) + (vy-testloc.y)*(vy-testloc.y) );
				
				if (dist1 > dist2) {
					vx = wx;
					vy = wy;
				}
			}

			if (collisions == -3) {
					//ERR("only friendly disturbing us!");
					//std::cout << "only friendly disturbing us!" << std::endl;
					pushFriendly(unit, vx, vy);
			}
			break;
		}

		if (working) {
			vx = wx;
			vy = wy;
			
			if (working == 2) pushFriendly(unit, vx, vy);
			
			break;
		}
		
		step++;		
	}

	unit->steering.x += round(vx);
	unit->steering.y += round(vy);							
}

bool SimulationAI::informObjectsOfDeath(GameObj* obj){
	// FIXME: THIS FUNCTION SEEMS BUGGY!!
	
	bool needStrategyUpdate=false;
	//Player* player = simulationState->getPlayer(*obj->sod.owner);
	//if(player == NULL) ERR("bad stuff happening, our player should always exist!");

	Player* owner = simulationState->getPlayer(*obj->sod.owner);
	
	if (obj->bp_name() == "controlCenter") removeBuilding(obj);
	
	FORALL(simulationState->getPlayers(), it){
		
		Vector<GameObj*> stopAttacking;
		if (it->second == owner) 
		{
			//if (debug) std::cout <<  "SimulationAI::informObjectsOfDeath: handled own object death" << std::endl;
			
			if ((it->second)->getObjectUnitMap().find(obj) != (it->second)->getObjectUnitMap().end()) {
				Unit *unit = (it->second)->getObjectUnitMap()[obj];
				if ( (it->second)->getOrtsUnits()[unit].state == ExtraObjectData::ATTACK \
					|| (it->second)->getOrtsUnits()[unit].state == ExtraObjectData::MOVE )
				{
					//if (debug) std::cout <<  "stopping my own attack" << std::endl;
					(it->second)->getOrtsUnits()[unit].isDone = true;
					stopAttacking.push_back(obj);	
				}
			}
			needStrategyUpdate = true;		
		}
		else 
		{
			//if (debug) std::cout <<  "SimulationAI::informObjectsOfDeath: handled other player object death" << std::endl;
			
			Player *currPlayer = it->second;
			FORALL((it->second)->getOrtsUnits(),it2){
				Unit *unit = it2->first;
				if (unit == NULL) ERR("SimulationAI::informObjectsOfDeath: NULL object encountered");
				
				ExtraObjectData data( it2->second );
				// if the objects we had a target to was killed... this unit is done its task
				if(data.state == ExtraObjectData::ATTACK || data.state == ExtraObjectData::DEFEND){
				if((data.targetUnit && data.targetUnit->getObject() == obj) \
				|| (data.targetBase && data.targetBase->getObject() == obj) ){
					// stop this attack!
					//if (debug) std::cout <<  "stopping other attack!" << std::endl;
					stopAttacking.push_back(it2->first->getObject());
					data.isDone = true;
					needStrategyUpdate = true;
					
					Order *order = currPlayer->getUnitGroupMap()[unit]->getOrder();
					//if (order == NULL) ERR("SimulationAI::informObjectsOfDeath: order is NULL"); 
					if (order) order->setCompleted(true);
				}
				}
			}
			
		}

		if (!stopAttacking.empty())
			call_handlers(NUS_TerrainModule::PathStopEvent(FROM, NUS_TerrainModule::FIND_PATH_STOP, stopAttacking));

	}
	return needStrategyUpdate;
// TODO: maybe add updates for other objects later.  not needed right now, since we only control friendly units
}

bool SimulationAI::isControllable(GameObj* obj){
	// first, check if we even have a weapon...and can move...
	if ((obj->get_max_speed() > 0) && (obj->component("weapon") != NULL)){
		// if we can, make sure it can be used....this will likely always pass
		if (obj->component("weapon")->has_attr("max_damage")){
			// make sure we can actually do dmg
			if(obj->component("weapon")->get_int("max_damage") > 0){ // if we can attack
				return true;
		}
	}
}
return false;
}

//might want to combine this with the following function, but keep it first...
sint4 SimulationAI::doPath(NUS_TerrainBase::Task& task){
	//if (debug) std::cout <<  "Send Pathfinding Message from GAME 2\n";
	Player *player = getMe();
	
	FORALL (task.objs, it)
	{
		GameObj *gameObj = (GameObj*)*it;
		Unit *unit = player->getObjectUnitMap()[gameObj];
		
		if (unit->getObject()->is_dead()) continue;
		//if (player->getOrtsUnits()[unit].isExecuting) stopPath(task);
		
		player->getOrtsUnits()[unit].isExecuting = true;
	}
	
	return call_handlers(NUS_TerrainModule::PathEvent(FROM, NUS_TerrainModule::FIND_PATH_MSG, task));
}

//send move command for an unit
//precondition: currObj is GameObj of unit to move
//							radius is the radius of circular area
// 							dest is centre in Vector3D
//postcondition: command sent, success or failure got to check from events
sint4 SimulationAI::doPath(Player* currPlayer, Vector<Unit*> &unitVector, Vector3D dest, sint4 radius, bool ignore_ALL){

	FORALL (unitVector, it) {
		//if (currPlayer->getOrtsUnits()[*it].isExecuting) stopPath(currPlayer, unitVector);
		currPlayer->getOrtsUnits()[*it].isExecuting = true;
	}
	
	NUS_TerrainBase::Loc destinationLoc(sint4(dest.x),sint4(dest.y));

	NUS_TerrainBase::Task task(NUS_TerrainBase::Goal::LOCATION, NUS_TerrainBase::Goal::VICINITY,
		NUS_TerrainBase::Task::INDIVIDUAL);

	task.consider = NUS_TerrainBase::Task::IGNORE_ALL;
	if (ignore_ALL) task.consider = NUS_TerrainBase::Task::IGNORE_ALL;
	if (!ignore_ALL) task.consider = NUS_TerrainBase::Task::IGNORE_FOES;
	
	ExtraObjectData newData;
	newData.location = destinationLoc;
	newData.state = ExtraObjectData::MOVE;
	newData.isDone = false;

	//note: we move unit to be within a circular target area by using NUS_TerrainBase::Goal::VICINITY option
	task.goal.loc = destinationLoc;
	
	sint4 groupsize;
	::Options::get("-groupsize", groupsize);

	task.goal.distance = radius; //groupsize * unitVector[0]->getObject()->get_int("radius") / 2;

	FORALL (unitVector,it) {
		if ((*it)->getObject()->is_dead()) continue;
		task.objs.clear();
		currPlayer->getOrtsUnits()[*it] = newData;
		task.objs.insert((*it)->getObject());
		doPath(task);
	}
	
	return 1;//doPath(task);
}

sint4 SimulationAI::stopPath(Player* currPlayer, Vector<Unit*> &unitVector){
	Vector<GameObj*> gameObjects;
	
	FORALL (unitVector,it) {
		gameObjects.push_back((*it)->getObject());
		
		currPlayer->getOrtsUnits()[*it].isExecuting = false;
	}
	
	//return 1;
	return call_handlers(NUS_TerrainModule::PathStopEvent(FROM, NUS_TerrainModule::FIND_PATH_STOP, gameObjects));
}

//plan path for a group
//preconditon: start and end point in Vector3D, empty WayPointVector
//post condition: WayPointVector contain the path in vector
real8 SimulationAI::planPath(const NUS_TerrainBase::Loc startPt, const NUS_TerrainBase::Loc endPt, sint4 radius, WayPointVector& wayPoints){

	Path path;
	// options are NUS_TerrainBase::IGNORE_ALL_MOBILE_OBJS and 
	/*
	enum ConsiderObjects {
		CONSIDER_ALL                = 0,
		/// Ignore all friendly (owned) mobile objects.
			IGNORE_MOBILE_FRIENDS   = 0x001, 
		/// Ignore all enemy mobile objects.
			IGNORE_MOBILE_FOES      = 0x010,
		/// Ignore all neutral ("sheep") mobile objects.
			IGNORE_MOBILE_NEUTRALS  = 0x100,
		/// Ignore all mobile objects, considering only boundaries and non-mobile objects.
			IGNORE_ALL_MOBILE_OBJS      = IGNORE_MOBILE_FRIENDS | IGNORE_MOBILE_FOES | IGNORE_MOBILE_NEUTRALS,
		};
	*/	
	real8 length=0;

	//if (debug) std::cout <<  "planPath: start" << std::endl;
    	
	//length = timp.find_path(startPt, endPt, radius, &path, NUS_TerrainBase::IGNORE_ALL_MOBILE_OBJS );
	//length = timp.find_path(startPt, endPt, radius, &path, NUS_TerrainBase::IGNORE_MOBILE_FOES );
	//length = timp.find_path(startPt, endPt, radius, &path, NUS_TerrainBase::CONSIDER_ALL);
	
	//FORALL(path.locs, it){
		//wayPoints.push_back(Vector3D(it->x,it->y,0.0));
		//if (debug) std::cout <<  Vector3D(it->x,it->y,0.0).toString() << "\t";
	//}
	
	wayPoints.clear();
	wayPoints.push_back(Vector3D(startPt.x,startPt.y,0.0));
	wayPoints.push_back(Vector3D(endPt.x, endPt.y ,0.0));
	
	return length;
}


//plan path for a group
//preconditon: start and end point in Vector3D, empty WayPointVector
//post condition: WayPointVector contain the path in vector
real8 SimulationAI::planPath(const NUS_TerrainBase::Loc startPt, const GameObj *target, sint4 radius, WayPointVector& wayPoints){

	Path path;
	// options are NUS_TerrainBase::IGNORE_ALL_MOBILE_OBJS and 
	/*
	enum ConsiderObjects {
		CONSIDER_ALL                = 0,
		/// Ignore all friendly (owned) mobile objects.
			IGNORE_MOBILE_FRIENDS   = 0x001, 
		/// Ignore all enemy mobile objects.
			IGNORE_MOBILE_FOES      = 0x010,
		/// Ignore all neutral ("sheep") mobile objects.
			IGNORE_MOBILE_NEUTRALS  = 0x100,
		/// Ignore all mobile objects, considering only boundaries and non-mobile objects.
			IGNORE_ALL_MOBILE_OBJS      = IGNORE_MOBILE_FRIENDS | IGNORE_MOBILE_FOES | IGNORE_MOBILE_NEUTRALS,
		};
	*/	
	real8 length = 0;

	//if (debug) std::cout <<  "planPath: start" << std::endl;
	NUS_TerrainBase::Goal goal(NUS_TerrainBase::Goal::OBJ, NUS_TerrainBase::Goal::VICINITY);
	goal.obj = target;
	goal.distance = target->get_int("radius"); // + radius;
	
	// NOTE: This is based on modified version of pathfinding routines in NUS_ST_ForceField. UNTESTED!
	//length = timp.find_path(startPt, goal, radius, &path, NUS_TerrainBase::IGNORE_ALL_MOBILE_OBJS );
    	
	//length = timp.find_path(startPt, endPt, radius, &path, NUS_TerrainBase::IGNORE_ALL_MOBILE_OBJS );
	//length = timp.find_path(startPt, endPt, radius, &path, NUS_TerrainBase::IGNORE_MOBILE_FOES );
	//length = timp.find_path(startPt, endPt, radius, &path, NUS_TerrainBase::CONSIDER_ALL);
	
	//FORALL(path.locs, it){
	//	wayPoints.push_back(Vector3D(it->x,it->y,0.0));
		//if (debug) std::cout <<  Vector3D(it->x,it->y,0.0).toString() << "\t";
	//}
	wayPoints.push_back(Vector3D(target->get_int("x"), target->get_int("y"), 0.0));
	
	return length;
}

//execute the path of order of a group, called once when we want to execute an order for plan
//precondition: currGroup is group to move
//postcondition: move command sent for each unit

void SimulationAI::executePath(Group* currGroup){
	if (debug) std::cout <<  "executing Path" << std::endl;
	if(!currGroup->getOrder()->isCompleted()){
		bool mode;
		::Options::get("-waypoint", mode);
		Vector3D dest;
		
		if(mode){	//go waypoint by waypoint
			dest = currGroup->getOrder()->getDestination();
		}else{		//go straight to the final destination
			dest = currGroup->getOrder()->getFinalDestination();
		}
		
		FORALL(currGroup->getUnits(), it) {
			MoveTo(*it, dest, 1.0f);
		}
		
		Player* ourPlayer = getMe();

		if(ourPlayer == NULL) ERR("umm, weirdness, our player should exist always");

		/*FORALL(currGroup->getUnits(), it) {
			doPath(ourPlayer, *it, dest, sint4(currGroup->getExtents()));
		}*/
		
		sint4 groupsize;
		::Options::get("-groupsize", groupsize);
		sint4 radius = ((currGroup->getNumUnits() > groupsize ) ? currGroup->getNumUnits() : groupsize ) * currGroup->getUnits()[0]->getObject()->get_int("radius") / 3; 
		//sint4 radius = currGroup->getNumUnits() * currGroup->getUnits()[0]->getObject()->get_int("radius") / 3; 
		
		if (radius < currGroup->getUnits()[0]->getObject()->get_int("radius")) 
			radius = currGroup->getUnits()[0]->getObject()->get_int("radius");

		sint4 moveVicinty;
		::Options::get("-moveVicinty", moveVicinty);
		
		if (moveVicinty != -1) radius = moveVicinty;

		doPath(ourPlayer, currGroup->getUnits(), dest, radius); 
		//sint4(currGroup->getExtents()));
	}
}

	// WARNING! 
	// Assumes units are from the same group
	void SimulationAI::replan(UnitVector &unitVector) {
		
		// REPLAN MOVES WITH ONLY IGNORE_MOBILE_FOES (IF IT ACTUALLY WORKS)	
		if (unitVector.empty()) return;
		
		Player* ourPlayer = getMe();
		
		Group* group = ourPlayer->getUnitGroupMap()[unitVector[0]];
		
		Order *currentOrder = group->getOrder();
		
		sint4 OLD_REPLAN_TRIGGER = timp.REPLAN_TRIGGER;
		sint4 OLD_OCCILATION_TRIGGER = timp.OCCILATION_TRIGGER;
		sint4 OLD_FIELD_RANGE_MULTIPLIER = timp.FIELD_RANGE_MULTIPLIER;
		sint4 OLD_REPULSOR_BASELINE = timp.REPULSOR_BASELINE;
		sint4 OLD_PATH_LOOK_AHEAD = timp.PATH_LOOK_AHEAD;
		
		timp.REPLAN_TRIGGER = 3; ///< Number of occilation steps before a replan will be triggered
		timp.OCCILATION_TRIGGER = 4;  ///< Number of occilation steps before a path movement failure will be passed
	 	timp.FIELD_RANGE_MULTIPLIER = 2; //2;
		timp.REPULSOR_BASELINE = 400; //200;
		timp.PATH_LOOK_AHEAD = 5;
		
		if (currentOrder && !currentOrder->isCompleted()) {
			if (currentOrder->getActionType() == "MOVE") {
				bool mode;
				::Options::get("-waypoint", mode);
				Vector3D dest;

				if(mode){	//go waypoint by waypoint
					dest = group->getOrder()->getDestination();
				}else{		//go straight to the final destination
					dest = group->getOrder()->getFinalDestination();
				}
				
				sint4 groupsize;
				::Options::get("-groupsize", groupsize);
				sint4 radius = group->getUnits()[0]->getObject()->get_int("radius")*4; //((group->getNumUnits() > groupsize ) ? group->getNumUnits() : groupsize ) * group->getUnits()[0]->getObject()->get_int("radius") / 3; 
				
				//move to the next way point
				FORALL (unitVector,it) {
					UnitVector uV;
					uV.push_back(*it);
					
					MoveTo(*it, dest, 1.0f);
					
					doPath(ourPlayer, uV, dest, radius);//, false);
					//group->getExtents());							
				}
			}
			else if (currentOrder->getActionType() == "ATTACK") {
				GameObj* currObj = unitVector[0]->getObject();
				GameObj* target = currentOrder->getTargetObject();
				
				if (target->is_dead()) return;

				uint4 distance = 0;

				uint4 weaponRange = currObj->component("weapon")->get_int("max_ground_range")/2;
				uint4 objRadius = currObj->get_int("radius")*4;
				distance = objRadius;

				if (distance > weaponRange) distance = weaponRange;

				NUS_TerrainBase::Task task(NUS_TerrainBase::Goal::OBJ, NUS_TerrainBase::Goal::VICINITY,
					NUS_TerrainBase::Task::INDIVIDUAL);

				task.goal.obj = target;
				//task.goal.loc.x = target->get_int("x");
				//task.goal.loc.y = target->get_int("y");

				task.goal.distance = distance;

				ExtraObjectData newData;
				newData.state = ExtraObjectData::ATTACK;

				if (currentOrder->targetUnit) {
					newData.targetUnit = currentOrder->targetUnit;
				}
				else {
					if (currentOrder->targetBase) {
						newData.targetBase = currentOrder->targetBase;
					}
					else ERR("Attack: No target set");
				}

				//task.consider = NUS_TerrainBase::Task::IGNORE_FOES;
				task.consider = NUS_TerrainBase::Task::IGNORE_ALL;
				
				Vector3D locVec = ToVector3D(NUS_TerrainBase::Loc(target->get_int("x"), target->get_int("y")));
				
				FORALL(unitVector,it) {
					if ((*it)->getObject()->is_dead()) continue;
					ourPlayer->getOrtsUnits()[*it] = newData;
				
					task.objs.clear();
					task.objs.insert((*it)->getObject());
					//if (debug) std::cout <<  "sending group to attack by moving nearer" << std::endl;
					
					MoveTo(*it, locVec, 1.0f);
					
					doPath(task);		
				}
			}
		}

		timp.REPLAN_TRIGGER = OLD_REPLAN_TRIGGER; ///< Number of occilation steps before a replan will be triggered
		timp.OCCILATION_TRIGGER = OLD_OCCILATION_TRIGGER;  ///< Number of occilation steps before a path movement failure will be passed		
	 	timp.FIELD_RANGE_MULTIPLIER = OLD_FIELD_RANGE_MULTIPLIER; //2;
		timp.REPULSOR_BASELINE = OLD_REPULSOR_BASELINE; //200;
		timp.PATH_LOOK_AHEAD = OLD_PATH_LOOK_AHEAD;
	}

//update the path, move to next waypoint, if all units has reach the current waypoint
//called whenever we got a path success msg
	bool SimulationAI::updatePath(Group* group){

		bool needStrategyUpdate = false;
		
		//std::cout<< "updating path" << std::endl;
		
		bool mode;
		::Options::get("-waypoint", mode);

		if (group->getOrder() && group->getOrder()->getActionType() == "MOVE") {
			
			if(group->getOrder()->isCompleted()) return true; //ERR("Order is already completed");
			
			if(mode){	//we are going from waypoint to waypoint
				Player* ourPlayer = getMe();
				if(ourPlayer == NULL) ERR("umm, weirdness, our player should exist always");

				bool allDone = true;

				FORALL(group->getUnits(), it) {
					if ((*it)->getObject()->is_dead()) continue;
					if(!ourPlayer->getOrtsUnits()[*it].isDone) {
						allDone = false;
						break;
					}
				}

				if(allDone){
					group->getOrder()->goToNextWayPoint();
	
					if(group->getOrder()->isCompleted()){		//if this path is done
						needStrategyUpdate = true;
					}
					else{
						sint4 groupsize;
						::Options::get("-groupsize", groupsize);
						sint4 radius = ((group->getNumUnits() > groupsize ) ? group->getNumUnits() : groupsize ) * group->getUnits()[0]->getObject()->get_int("radius") / 3; 
						//sint4 radius = currGroup->getNumUnits() * currGroup->getUnits()[0]->getObject()->get_int("radius") / 3; 

						if (radius < group->getUnits()[0]->getObject()->get_int("radius")) 
							radius = group->getUnits()[0]->getObject()->get_int("radius");
						
						Vector3D dest = group->getOrder()->getDestination();
						//move to the next way point
						FORALL(group->getUnits(), it){
							UnitVector unitVector;
							unitVector.push_back(*it);
							doPath(ourPlayer,unitVector, dest, radius);
						}
					}
				} //else we wait for all unit to reach its current way point first
			}
			else{	//we are going to the final destination straight
				group->getOrder()->setCompleted(true);	//we are completed already
				needStrategyUpdate = true;
			}
		} 
		else {
			if (group->getOrder() && group->getOrder()->getActionType() == "ATTACK") {		
				GameObj *target = group->getOrder()->getTargetObject();
				uint4 weaponRange = group->getMaxAttackRange();
				real8 groupDistance = group->distance_to(target);
				Order* currentOrder = group->getOrder();
				
				Player* ourPlayer = getMe();
				if(ourPlayer == NULL) ERR("umm, weirdness, our player should exist always");
				
				bool allWithinRange = true;
				
				Vector<GameObj::WPtr> foo;
				FORALL(group->getUnits(), it) {
					if ((*it)->getObject()->is_dead()) continue;
					ourPlayer->getOrtsUnits()[*it].isDone = true;
					
					if ((*it)->distance_to(target) < weaponRange) {
						foo.clear();
						foo.push_back((*it)->getObject());
						GameObj *target = group->getOrder()->getTargetObject();
						doCombat(foo, target);
					} 
					else 
						allWithinRange = false;
				}
				// attack as a squad (does not seem to work well)
				/*
				if (!foo.empty()) {
					GameObj *target = group->getOrder()->getTargetObject();
					doCombat(foo, target);
				}
				*/
				if (allWithinRange) {
					group->getOrder()->setCompleted(true);		
				}
				needStrategyUpdate = true;
			}
		}
		
		return needStrategyUpdate;
	}

void SimulationAI::executeAttack(Group* currGroup, int phase){

	Player* ourPlayer = getMe();

	if(ourPlayer == NULL) ERR("umm, weirdness, our player should exist always");

	//if (debug) std::cout <<  "SimulationAI::executeAttack: starting" << std::endl;

	if (currGroup->getOrder() == NULL) ERR("SimulationAI::executeAttack: order is null");
	if (currGroup->getOrder()->getTargetObject() == NULL) ERR("SimulationAI::executeAttack: target is NULL");

	if (currGroup->getOrder()->getTargetObject()->is_dead()) {
		//if (debug) std::cout <<  "SimulationAI::executeAttack: trying to attack dead object" << std::endl;
		currGroup->getOrder()->setCompleted(true); //ERR("Trying to attack a dead object");
		return;
	}

	Order* currentOrder = currGroup->getOrder();
	GameObj* currObj = currGroup->getUnits()[0]->getObject();
	GameObj* target = currGroup->getOrder()->getTargetObject();
		
	uint4 distance = 0;
	
	if (phase == 0) {
		uint4 weaponRange = currObj->component("weapon")->get_int("max_ground_range")/2;
		uint4 objRadius = currObj->get_int("radius")*4;
		distance = objRadius;
		
		if (distance > weaponRange) distance = weaponRange;
	}
	
	NUS_TerrainBase::Task task(NUS_TerrainBase::Goal::OBJ, NUS_TerrainBase::Goal::VICINITY,
		NUS_TerrainBase::Task::INDIVIDUAL);
		
	task.goal.obj = target;
	//task.goal.loc.x = target->get_int("x");
	//task.goal.loc.y = target->get_int("y");	
	task.goal.distance = distance;

	ExtraObjectData newData;
	newData.state = ExtraObjectData::ATTACK;

	if (currentOrder->targetUnit) {
		newData.targetUnit = currentOrder->targetUnit;
	}
	else {
		if (currentOrder->targetBase) {
			newData.targetBase = currentOrder->targetBase;
		}
		else ERR("Attack: No target set");
	}
	
	Vector3D dest = ToVector3D(NUS_TerrainBase::Loc(target->get_int("x"), target->get_int("y")));
	
	FORALL(currGroup->getUnits(), it) {			
		currObj = (*it)->getObject();
		ourPlayer->getOrtsUnits()[(*it)] = newData;
		
		MoveTo(*it, dest, 1.0f);

		task.objs.clear();
		task.objs.insert(currObj);
		//if (debug) std::cout <<  "sending group to attack by moving nearer" << std::endl;
		doPath(task);		
	}

}

sint4 SimulationAI::doCombat(Vector<GameObj::WPtr> units, GameObj::WPtr target) {
	//if (debug) std::cout <<  "attack target" << endl;
	if (*((GameObj*)target)->sod.owner == *((GameObj *)units[0])->sod.owner) return -1;
	return call_handlers(NUS_SquadCombatEvent(FROM, NUS_SquadCombatAI::ATTACK_TARGET, units, target));
}

sint4 SimulationAI::stopCombat(Vector<GameObj::WPtr> units) {
	//if (debug) std::cout <<  "stop attack" << endl;
	return call_handlers(NUS_SquadCombatEvent(FROM, NUS_SquadCombatAI::STOP_ATTACK, units));
}

	//get my player
Player* SimulationAI::getMe(){
	const sint4 me = gsm.get_game().get_client_player();

	return simulationState->getPlayer(me);
}





//1024x768
//112
float SimulationAI::dist2dCompare(NUS_TerrainBase::Loc a, NUS_TerrainBase::Loc b)
{
	return (a.x-b.x)*(a.x-b.x) + (a.y-b.y)*(a.y-b.y);
}

float SimulationAI::dist2dCompare(float a,float b,float x,float y)
{
	return (a-x)*(a-x) + (b-y)*(b-y);
}

float SimulationAI::dist2d(NUS_TerrainBase::Loc a, NUS_TerrainBase::Loc b)
{
	return sqrt((a.x-b.x)*(a.x-b.x) + (a.y-b.y)*(a.y-b.y));
}  

NUS_Simul::Unit* SimulationAI::GetNearestEnemy(NUS_TerrainBase::Loc pos)
{
	float nearest =  10000000;
	Unit* enemyunit = 0;
	
	FORALL(simulationState->getPlayers(),it) 
	{
		if(it->second == getMe()) continue; // don't check our own units
		FORALL((it->second)->getOrtsUnits(),it2)
		{
			Unit* unit = it2->first;
			if (unit->getObject()->component("weapon") == NULL) continue;
			float dist = dist2dCompare(pos, unit->getPosition());
			if (dist<nearest)
			{
				nearest=dist;
				enemyunit = unit;
			}
			
			
		}			
	}
	
	return enemyunit;	
}

Vector3D SimulationAI::ToVector3D(NUS_TerrainBase::Loc pos)
{
	Vector3D position;
	position.x=pos.x; position.y=pos.y;
	return 	position;
}

NUS_TerrainBase::Loc SimulationAI::ToTerrainBase(Vector3D vec3)
{
	NUS_TerrainBase::Loc position;
	position.x=vec3.x; position.y=vec3.y;
	return 	position;
} 

bool SimulationAI::OffMapLimits(Vector3D destination)
{
	if (destination.x>(width*tile_points)) {return true;}
	if (destination.x<0) {return true;}
	if (destination.y>(height*tile_points)) {return true;}
	if (destination.y<0) {return true;}
	return false;
}

bool SimulationAI::TestPtClear(float x, float y, Unit* ourunit)
{
			float dismissalradius = 15.5f*15.5f;
			//float dismissalradius = 12.5f*12.5f;
			FORALL(simulationState->getPlayers(),it) 
			{
				FORALL((it->second)->getOrtsUnits(),it2)
				{
					Unit* otherunit = it2->first;
					//if (otherunit->getObject()->component("weapon") == NULL) dismissalradius=10.0f*10.0f;
					if (ourunit->getObject()!=otherunit->getObject())
					{
						NUS_TerrainBase::Loc posTB = otherunit->getPosition(); // TerrainBase::Loc
						float a= posTB.x;
						float b= posTB.y;
						if (dist2dCompare(x,y,a,b)<dismissalradius)
						{
							return false;
						}
					}
					//if (otherunit->getObject()->component("weapon") == NULL) dismissalradius=8.0f*8.0f;
				}
			}
			
			dismissalradius = 12.5f*12.5f;
			FORALL(sheep, it) {
				GameObj *gob = *it;
				float a= *(gob->sod.x);
				float b= *(gob->sod.y);
				if (dist2dCompare(x,y,a,b)<dismissalradius)
				{
					return false;
				}
			}
			return true;
}

//bool SimulationAI::Prod(float x, float y, float heading, float speed, Unit* ourunit)
//{
	//for (int i=0; i<
//}



int SimulationAI::turnangle (float curDirection, float desDirection, float turnspeed)
{
	
		int dir, turndirection;
		float turnwidth;  //dir : 1= anticlockwise, dir: 0= clockwise
		
		turndirection=3;

		turnwidth = desDirection - curDirection;
		dir = 1;
		if ( turnwidth < 0) 
		{
			turnwidth = curDirection - desDirection; 
			dir = 0;
		}
		
	  if (turnwidth < (turnspeed * 2.0f))  
		{
			turndirection = 3;
		}
	  else
		{
				

				
		  if (dir==1) 
		  {
			if ((6.28f-turnwidth)< turnwidth)  
				{
					turndirection = -1 ;
				}
			else
				{
					turndirection = 1 ;
				}
		  }


		  if (dir==0) 
		  {
			if ((6.28f-turnwidth)< turnwidth)  
				{
					turndirection = 1;
				}
			else
				{
					turndirection = -1;
				}
		  }

		
		}
	  
	  
	  return turndirection;

}

bool SimulationAI::TestTerrainClear(float x, float y)
{
	//return !(besideObstacle(ToTerrainBase(Vector3D(x,y,0)), 1.0f));
	//return true;
	
	sint4 tilei = x / tile_points;
    sint4 tilej = y / tile_points;
	
	// check if it collides with non-mobile stuff
	if (tilei < 0 || tilei >= width || tilej < 0 || tilej >= height || map[tilei][tilej].status != Tile::FREE)
	{
		return false;
	}
	else
	{
		return true;
	}
}

bool SimulationAI::MTTA (float x, float y, float heading)
{
	float dx = sinf(heading);
	float dy = cosf(heading);
	for (float i=16.0f; i<38.0f; i+=2.0f)
	{
		if (TestTerrainClear(x+dx*i, y+dy*i) == false)
		{ 
			return false;
		}
	}
	
	dx = sinf(heading-0.7f);
	dy = cosf(heading-0.7f);
	for (float i=7.0f; i<9.0f; i+=1.0f)
	{
		if (TestTerrainClear(x+dx*i, y+dy*i) == false)
		{ 
			return false;
		}
	}
	dx = sinf(heading+0.7f);
	dy = cosf(heading+0.7f);
	for (float i=7.0f; i<9.0f; i+=1.0f)
	{
		if (TestTerrainClear(x+dx*i, y+dy*i) == false)
		{ 
			return false;
		}
	}
	dx = sinf(heading-1.5f);
	dy = cosf(heading-1.5f);
	for (float i=7.0f; i<8.0f; i+=2.0f)
	{
		if (TestTerrainClear(x+dx*i, y+dy*i) == false)
		{ 
			return false;
		}
	}
	dx = sinf(heading+1.5f);
	dy = cosf(heading+1.5f);
	for (float i=7.0f; i<8.0f; i+=2.0f)
	{
		if (TestTerrainClear(x+dx*i, y+dy*i) == false)
		{ 
			return false;
		}
	}
	return true;
}

Vector3D SimulationAI::MTFA(float x, float y, float heading, float speed, Unit* ourunit)
{
	/*
	float dir=1;
	int leftcount, rightcount;
	leftcount=rightcount=0;
			FORALL(simulationState->getPlayers(),it) 
			{
				FORALL((it->second)->getOrtsUnits(),it2)
				{
					Unit* otherunit = it2->first;
					//if (otherunit->getObject()->component("weapon") == NULL) dismissalradius=10.0f*10.0f;
					if (ourunit->getObject()!=otherunit->getObject())
					{
						float angle = angfrom(x,y, otherunit->getPosition().x, otherunit->getPosition().y);
						if (turnangle(heading,angle,0.0001f)==-1) leftcount+=1; else rightcount+=1; 		 
					}
				}
			}
			if (leftcount>rightcount) dir=-1;
	*/
	for (float i=0.03f; i<6.28f; i+=0.03f)
	{
		Vector3D testpt;
		if (MTTA(x,y, heading+i)==true)
		{
		
			testpt.x = x+sinf(heading+i)*speed;
			testpt.y = y+cosf(heading+i)*speed;
			testpt.z =0;
			
			if (TestPtClear(testpt.x, testpt.y, ourunit)==true) return testpt;
		}
		
		
		if (MTTA(x,y, heading-i)==true)
		{
		
			testpt.x = x+sinf(heading-i)*speed;
			testpt.y = y+cosf(heading-i)*speed;
			testpt.z =0;
			
			if (TestPtClear(testpt.x, testpt.y, ourunit)==true) return testpt;
		}
		
	} 
	return Vector3D(x,y,0);
}



void SimulationAI::MoveTo(Unit* unit, Vector3D destination, float recommendedspeed)
{
	Vector3D newpos = destination;
	Vector3D direction = (newpos - ToVector3D(unit->getPosition())).normalized();
	Vector3D hundredforwardpos = ToVector3D(unit->getPosition()) + direction * 100.0f; 
	bool slide=OffMapLimits(hundredforwardpos);
	
	
	if (besideObstacle(unit->getPosition(), 7.1f))
	{
		float angle = angfrom(ToVector3D(unit->getPosition()).x, ToVector3D(unit->getPosition()).y, destination.x, destination.y);
		for (float dist = 0; dist< 20.0f; dist+=1.0f)
		{
		for (float i=0.03f; i<3.14f; i+=0.1f)
		{
			Vector3D newpos = destination;
			newpos.x = unit->getPosition().x+sinf(angle+i)*(10.0f+dist);
			newpos.y = unit->getPosition().y+cosf(angle+i)*(10.0f+dist);
			newpos.z = 0;
			
			if (besideObstacle(ToTerrainBase(newpos), 7.1f)==false)
			{
					
	
					    Vector<sint4> params;
				        params.push_back(newpos.x);
				        params.push_back(newpos.y);
				        GameObj *gob = unit->getObject();
				        params.push_back(*gob->sod.max_speed*recommendedspeed);
				
				        gob->set_action("move", params);
				        return;
				        
			}
		}
		}
		return;
	} 
	
	
	
	
	if (slide==true)
	{
		Vector3D tenforwardpos = ToVector3D(unit->getPosition()) + direction * 100.0f;
		if (tenforwardpos.x<0) newpos.x =  8;
		if (tenforwardpos.x>width*tile_points) tenforwardpos.x=width*tile_points-8;
		
		if (tenforwardpos.y<0) newpos.y =  8;
		if (tenforwardpos.y>height*tile_points) tenforwardpos.y=height*tile_points-8;
		
		Vector3D movevector = newpos-ToVector3D(unit->getPosition());
		
        GameObj *gob = unit->getObject();
		newpos = ToVector3D(unit->getPosition())+movevector.normalized() * (*gob->sod.max_speed*recommendedspeed);
		
	}
	
	
	if (newpos.x>(width*tile_points-1)) {newpos.x=(width*tile_points-1);}
	if (newpos.x<1) {newpos.x=1;}
	if (newpos.y>(height*tile_points-1)) {newpos.y=(height*tile_points-1);}
	if (newpos.y<1) {newpos.y=1;}
	
	
	
	
	Unit* ourunit = unit;
	Vector3D ourunitpos = ToVector3D(ourunit->getPosition());
	
	float MaxSpeed = (float) ourunit->getObject()->get_int("max_speed");
	Vector3D NextPos = ourunitpos+ (newpos-ourunitpos).normalized()*MaxSpeed;
	
	/*
			FORALL(simulationState->getPlayers(),it) 
			{
				FORALL((it->second)->getOrtsUnits(),it2)
				{
					Unit* otherunit = it2->first;
					if (ourunit->getObject()!=otherunit->getObject())
					{
						NUS_TerrainBase::Loc posTB = otherunit->getPosition(); // TerrainBase::Loc
						Vector3D otherunitpos = ToVector3D(posTB);
						if ((otherunitpos-NextPos).length()<8.1f)
						{
							//std::cout<<"collllide "<<(otherunitpos-NextPos).length();
							Vector3D tootherunit = otherunitpos-ourunitpos;
							tootherunit.z =0;
							tootherunit = tootherunit.normalized();
							Vector3D direction = newpos-ourunitpos;
							direction.z =0;
							direction = direction.normalized();
							
							if (direction.dot(tootherunit)>0)
							{
								newpos.x = ourunitpos.x - tootherunit.y*10.0f;
								newpos.y = ourunitpos.y + tootherunit.x*10.0f;
							}
							else
							{
								newpos.x = ourunitpos.x + tootherunit.y*10.0f;
								newpos.y = ourunitpos.y - tootherunit.x*10.0f;
							}
						}
					}
				}			
			}
			*/
	if (MTTA(ourunitpos.x, ourunitpos.y, angfrom(ourunitpos.x, ourunitpos.y, NextPos.x, NextPos.y))==false)
	{
		newpos = MTFA(ourunitpos.x, ourunitpos.y, angfrom(ourunitpos.x, ourunitpos.y, NextPos.x, NextPos.y), MaxSpeed, ourunit);
	}
	if (TestPtClear(NextPos.x, NextPos.y,ourunit)==false)
	{
		newpos = MTFA(ourunitpos.x, ourunitpos.y, angfrom(ourunitpos.x, ourunitpos.y, NextPos.x, NextPos.y), MaxSpeed, ourunit);
	}
			
	//new MoveToLocation(this,myGroupVector,function,name,0.0,new MoveStruct(ToTerrainBase(newpos)));
	// Assume group of 1 unit
	/*NUS_TerrainBase::Loc startPt(unit->getPosition().x, unit->getPosition().y);
	WayPointVector wayPoints;
	
	real8 length = state->simAI->planPath(startPt, NUS_TerrainBase::Loc(newpos.x,newpos.y), unit->getExtents(), wayPoints);
	Order* order = new Order(wayPoints, unit->getMaxSpeed());
	order->setActionType("MOVE");
	unit->setOrder(order);
	std::cout << "trying to move" << std::endl;
	*/
	    Vector<sint4> params;
        params.push_back(newpos.x);
        params.push_back(newpos.y);
        GameObj *gob = unit->getObject();
        params.push_back(*gob->sod.max_speed*recommendedspeed);

        gob->set_action("move", params);
	
} 

float SimulationAI::angfrom (float x1, float y1, float x2, float y2)   
{     
	float hdist, vdist, angle;
	hdist = (x1 - x2);
	vdist = (y1 - y2);

	


    if (vdist == 0.0f )
	{
        if (hdist < 0.0f)  {angle = 4.71238898f;} else {angle = 1.570796327f;}
	}   
    else
	{
		angle = (float) atan(hdist / vdist) + 3.141592654f;		
		if (vdist > 0.0f) {angle = angle + 3.14592654f;}
	}
	

	angle = angle + 3.14592654f;
	if (angle > 6.283185307f) {angle = angle - 6.283185307f;}
	if (angle < 0.0f) {angle = angle + 6.283185307f;}
	return angle;

}

struct TravoUnit
{
	Vector3D vec;
};

float SimulationAI::angdist (float curDirection, float desDirection)
{
	
		int dir;
		float turnwidth;  //dir : 1= anticlockwise, dir: 0= clockwise
		
		turnwidth = desDirection - curDirection;
		
		dir = 1;
		if ( turnwidth < 0) 
		{
			turnwidth = curDirection - desDirection; 
			dir = 0;
		}
		
	  if (turnwidth ==0)  
		{
			return 0;
		}
	  else
		{
				

				
		  if ((dir==1) && ((6.28f-turnwidth)< turnwidth))  
				{
					turnwidth = (6.28f-turnwidth);
				}
		  


		  if ((dir==0) && ((6.28f-turnwidth)< turnwidth))
				{
					turnwidth = (6.28f-turnwidth);
				}
		  

		
		}
	  
	  
	  return turnwidth;

}



}
