#include"GLee.h"
#include<GL/gl.h>
#include<GL/glu.h>
#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include"glutil.h"

int printerror(char *where)
{
  int err=glGetError();
  switch(err)
  { 
    case 0: break; //not an error
    case GL_INVALID_ENUM: printf("%s: invalid enum\n",where);
			  break;
    case GL_INVALID_OPERATION: printf("%s: invalid operation\n",where);
			  break;
    case GL_INVALID_VALUE: printf("%s: invalid value\n",where);
			  break;
    default:
    printf("%s: opengl error %d\n",err);
  }
  return err;
}

unsigned int loadtga(const char *filename)
{
  int id;
  FILE *f=fopen(filename,"rb");
  if(!f)return 0;
  id=loadtga(f);
  fclose(f);
  return id;
}

unsigned int loadtga(FILE *f)
{
  char *data;
  short w,h;
  unsigned char bpp;
  GLuint id;
  
  char tga_header_rgb[12] = { 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0};
  char tga_header_gray[12] = {0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0}; 
  char file_header[12];

  fread(file_header,sizeof(file_header),1,f);

  if (memcmp(file_header, tga_header_rgb, 12) &&
      memcmp(file_header, tga_header_gray, 12)) {
    fclose(f);
    return 0;
  }

  fread(&w,sizeof(w),1,f);
  fread(&h,sizeof(h),1,f);
  fread(&bpp,sizeof(bpp),1,f);
  bpp=bpp/8;
  fgetc(f);

  int size = w*h*bpp;

  data = new char[size];

  fread(data,size,1,f);

  if (bpp != 1) // flip red and blue
    for (int i = 0; i < size; i += bpp)
    {
      char tmp=data[i];
      data[i]=data[i+2];
      data[i+2]=tmp;
    }

  glGenTextures(1, &id);
  glBindTexture(GL_TEXTURE_2D,id);

  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,
                                          GL_LINEAR_MIPMAP_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,
                                          GL_LINEAR_MIPMAP_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
  glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);

  GLuint type;

  switch(bpp) {
    case 1:  type=GL_LUMINANCE; break;
    case 3:  type=GL_RGB;   break;
    case 4:  type=GL_RGBA;  break;
  } 

  gluBuild2DMipmaps(GL_TEXTURE_2D, type, w, h, type, GL_UNSIGNED_BYTE, data);

  delete [] data;
  return id;
}

unsigned int loadtga1D(const char *filename)
{
  int id;
  FILE *f=fopen(filename,"rb");
  if(!f)return 0;
  id=loadtga1D(f);
  fclose(f);
  return id;
}

unsigned int loadtga1D(FILE *f)
{
  char *data;
  short w,h;
  unsigned char bpp;
  GLuint id;
  
  char tga_header_rgb[12] = { 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0};
  char tga_header_gray[12] = {0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0}; 
  char file_header[12];

  fread(file_header,sizeof(file_header),1,f);

  if (memcmp(file_header, tga_header_rgb, 12) &&
      memcmp(file_header, tga_header_gray, 12)) {
    fclose(f);
    return 0;
  }

  fread(&w,sizeof(w),1,f);
  fread(&h,sizeof(h),1,f);
  fread(&bpp,sizeof(bpp),1,f);
  bpp=bpp/8;
  fgetc(f);

  int size = w*h*bpp;

  data = new char[size];

  fread(data,size,1,f);

  if (bpp != 1) // flip red and blue
    for (int i = 0; i < size; i += bpp)
    {
      char tmp=data[i];
      data[i]=data[i+2];
      data[i+2]=tmp;
    }

  glGenTextures(1, &id);
  glBindTexture(GL_TEXTURE_2D,id);

  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,
                                          GL_LINEAR_MIPMAP_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,
                                          GL_LINEAR_MIPMAP_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
  glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);

  GLuint type;

  switch(bpp) {
    case 1:  type=GL_LUMINANCE; break;
    case 3:  type=GL_RGB;   break;
    case 4:  type=GL_RGBA;  break;
  } 

  gluBuild2DMipmaps(GL_TEXTURE_2D, type, w, h, type, GL_UNSIGNED_BYTE, data);

  delete [] data;
  return id;
}


unsigned int loadppm(const char *filename)
{
  int id;
  FILE *f=fopen(filename,"rb");
  if(!f)return 0;
  id=loadppm(f);
  fclose(f);
  return id;
}

unsigned int loadppm(FILE *f)
{
  // no comments support
  unsigned char *data;
  int w,h,m;
  unsigned int id;

  if(fgetc(f)!='P' && fgetc(f)!='6') return 0;

  while(fgetc(f)!='\n');
  fscanf(f,"%d %d %d",&w,&h,&m);
  fgetc(f);

  data=(unsigned char *)malloc(w*h*3);

  fread(data,1,w*h*3,f);

  glGenTextures(1,&id);

  glBindTexture(GL_TEXTURE_2D,id);

  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,
                                          GL_LINEAR_MIPMAP_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,
                                          GL_LINEAR_MIPMAP_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
  glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);

  gluBuild2DMipmaps(GL_TEXTURE_2D,3,w,h,
                    GL_RGB,GL_UNSIGNED_BYTE,data);

  free(data);
  return id;
}

int screenshot(char *filename)
{
  int l;
  FILE *f=fopen(filename,"wb");
  if(!f)return 0;
  l=screenshot(f);
  fclose(f);
  return l;
}

int screenshot(FILE *f)
{
  char *data;
  int vp[4];
  
  glGetIntegerv(GL_VIEWPORT,vp);

  char tga_header_rgb[12] = { 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0};

  fwrite(tga_header_rgb,sizeof(tga_header_rgb),1,f);
  fwrite(vp+2,2,1,f);
  fwrite(vp+3,2,1,f);
  fputc(32,f);
  fputc(8,f);

  int size = vp[2]*vp[3]*4;

  data = new char[size];

  glReadPixels(vp[0],vp[1],vp[2],vp[3],GL_BGRA,GL_UNSIGNED_BYTE,data);
  
  fwrite(data,size,1,f);

  delete data;
  return 18+size;
}

void begin2d()
{
  int vp[4];
  glGetIntegerv(GL_VIEWPORT,vp);
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  glOrtho(0,vp[2],0,vp[3],-0.1,1000);
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();
  printerror("begin2d");
}

void end2d()
{
  glMatrixMode(GL_PROJECTION);
  glPopMatrix();
  glMatrixMode(GL_MODELVIEW);
  glPopMatrix();
  printerror("end2d");
}


void printProgramStats()
{
    int len,mlen;

    if(printerror("pixel shader 2 :")){
        int errPos;
        glGetIntegerv( GL_PROGRAM_ERROR_POSITION_ARB, &errPos );
        char  * str = (char *)glGetString(GL_PROGRAM_ERROR_STRING_ARB);
        printf("error at %d %s",errPos,str);
        exit(0);
    }

    glGetProgramivARB(GL_FRAGMENT_PROGRAM_ARB,
                      GL_MAX_PROGRAM_INSTRUCTIONS_ARB,
                      &mlen);
    glGetProgramivARB(GL_FRAGMENT_PROGRAM_ARB,
                      GL_PROGRAM_INSTRUCTIONS_ARB,
                      &len);
    printf("Instructions %d/%d\n",len,mlen);
    glGetProgramivARB(GL_FRAGMENT_PROGRAM_ARB,
                      GL_MAX_PROGRAM_NATIVE_INSTRUCTIONS_ARB,
                      &mlen);
    glGetProgramivARB(GL_FRAGMENT_PROGRAM_ARB,
                      GL_PROGRAM_NATIVE_INSTRUCTIONS_ARB,
                      &len);
    printf("Native Instructions %d/%d\n",len,mlen);
    glGetProgramivARB(GL_FRAGMENT_PROGRAM_ARB,
                      GL_MAX_PROGRAM_TEMPORARIES_ARB,
                      &mlen);
    glGetProgramivARB(GL_FRAGMENT_PROGRAM_ARB,
                      GL_PROGRAM_TEMPORARIES_ARB,
                      &len);
    printf("Program Temporaries %d/%d\n",len,mlen);
    glGetProgramivARB(GL_FRAGMENT_PROGRAM_ARB,
                      GL_MAX_PROGRAM_TEX_INSTRUCTIONS_ARB,
                      &mlen);
    glGetProgramivARB(GL_FRAGMENT_PROGRAM_ARB,
                      GL_PROGRAM_TEX_INSTRUCTIONS_ARB,
                      &len);
    printf("Tex instructions %d/%d\n",len,mlen);

    glGetProgramivARB(GL_FRAGMENT_PROGRAM_ARB,
                      GL_MAX_PROGRAM_ALU_INSTRUCTIONS_ARB,
                      &mlen);
    glGetProgramivARB(GL_FRAGMENT_PROGRAM_ARB,
                      GL_PROGRAM_ALU_INSTRUCTIONS_ARB,
                      &len);
    printf("ALU instructions %d/%d\n",len,mlen);

    glGetProgramivARB(GL_FRAGMENT_PROGRAM_ARB,
                      GL_MAX_PROGRAM_TEX_INDIRECTIONS_ARB,
                      &mlen);
    glGetProgramivARB(GL_FRAGMENT_PROGRAM_ARB,
                      GL_PROGRAM_TEX_INDIRECTIONS_ARB, &len);

    printf("Tex indirections %d/%d\n",len,mlen);

    glGetProgramivARB(GL_FRAGMENT_PROGRAM_ARB,
                      GL_MAX_PROGRAM_NATIVE_TEX_INSTRUCTIONS_ARB,
                      &mlen);
    glGetProgramivARB(GL_FRAGMENT_PROGRAM_ARB,
                      GL_PROGRAM_NATIVE_TEX_INSTRUCTIONS_ARB,
                      &len);
    printf("Native Tex instructions %d/%d\n",len,mlen);

    glGetProgramivARB(GL_FRAGMENT_PROGRAM_ARB,
                      GL_MAX_PROGRAM_NATIVE_ALU_INSTRUCTIONS_ARB,
                      &mlen);
    glGetProgramivARB(GL_FRAGMENT_PROGRAM_ARB,
                      GL_PROGRAM_NATIVE_ALU_INSTRUCTIONS_ARB,
                      &len);
    printf("Native ALU instructions %d/%d\n",len,mlen);

    glGetProgramivARB(GL_FRAGMENT_PROGRAM_ARB,
                      GL_MAX_PROGRAM_NATIVE_TEX_INDIRECTIONS_ARB,
                      &mlen);
    glGetProgramivARB(GL_FRAGMENT_PROGRAM_ARB,
                      GL_PROGRAM_NATIVE_TEX_INDIRECTIONS_ARB,
                      &len);
    printf("Native Tex indirections %d/%d\n",len,mlen);
}


