// $Id: PWidget.C,v 1.19 2005/07/18 01:22:11 orts_furtak Exp $

// This is an ORTS file (c) Michael Buro, licensed under the GPL

#include "PWidget.H"
#include "GfxGlobal.H"

using namespace std;

stype PWidget::event_key_down(stype key, stype /*mod*/, stype /*x*/, stype /*y*/)
{
  cout << "KEY " << (sint4)key << " down" << endl;
  
  sint4 index = key;
  
  switch(index) {
  case 's':
    display_segments = !display_segments;
    break;
  case 'o':
    display_obstacles = !display_obstacles;
    break;
  case 'g':
    display_graph = !display_graph;
    break;
  case 'p':
    //display_path = !display_path;
    break;
  case 'c':
    display_corridors = !display_corridors;
    break;
  case 'd':
    display_decomposition = !display_decomposition;
    break;
  default:
    break;
  }
  return 0L;
}

stype PWidget::draw()
{
  //  sint4 x1 = x;
  //sint4 y1 = y;
  //sint4 x2 = x+h;
  //sint4 y2 = y+w;

  glEnable(GL_SCISSOR_TEST);
  //glScissor(x-200,y-200,w+400,h+400);
  glScissor(x-10, y-5, w+20, h+5);
  
  glColor4f(1, 0, 0, 1);

  glDisable(GL_TEXTURE_2D);

  //draw the bounding box
#if 0  
  glBegin(GL_LINE_LOOP);
  glVertex2f(x2, y1);
  glVertex2f(x2, y2);
  glVertex2f(x1, y2);
  glVertex2f(x1, y1);
  glEnd();
#endif

  glBegin(GL_LINES);
  Point2i p1,p2,p3,p4;
  p1.x = 4;
  p1.y = 4;
  p2.x = 4+world_width;
  p2.y = 4;
  p3.x = 4+world_width;
  p3.y = 4+world_height;
  p4.x = 4;
  p4.y = 4+world_height;
  
  draw_line(p1,p2);
  draw_line(p2,p3);
  draw_line(p3,p4);
  draw_line(p4,p1);
  
  glEnd();
	
  /*
    glBegin(GL_LINES);
    glVertex2i(x1,y1);
    glVertex2i(x2,y2);
    glEnd();
  */

  //cout<<world_width<<" "<<world_height<<endl;
  
  real4 scale_x = (real4)w/(real4)world_width;
  real4 scale_y = (real4)h/(real4)world_height;

  /*
    cout<<x<<" "<<y<<endl;
 
    cout<<w<<" "<<h<<endl;
    
    cout<<world_width<<" "<<world_height<<endl;
    cout<<scale_x<<" "<<scale_y<<endl;
  */

  //draw the boundaries
  
#if 1

  real4 max_y = 0;
  
  if (display_segments) {

    glBegin(GL_LINES);

    FORALL (lines, it) {
	    
      /*
	Point2i p1,p2;
	p1.x = it->x1;
	p1.y = it->y1;
	p2.x = it->x2;
	p2.y = it->y2;

	draw_line(p1,p2);

	continue;
      */	  
      real4 u1 = it->x1;
      real4 v1 = it->y1;
      real4 u2 = it->x2;
      real4 v2 = it->y2;

      if (v1 > max_y) max_y = v1;
      if (v2 > max_y) max_y = v2;
	    
      sint4 h = (sint4)this->h;
      sint4 x = (sint4)this->x;
      sint4 y = (sint4)this->y;

      u1 = u1*scale_x + x;
      v1 = v1*scale_y + y;
	    
      u2 = u2*scale_x + x;
      v2 = v2*scale_y + y;
	    
      v1 += h - 2*(v1-y);
      v2 += h - 2*(v2-y);
	    
      glVertex2d(u1, v1);
      glVertex2d(u2, v2);
    }

    glEnd();
  }
#endif


  sint4 new_max_y = 0;
  
  //draw the decomposition
  if (display_decomposition || display_obstacles) {

    glColor3f(0, 1, 0);

    //sint4 count = 0;
  
    FORALL (curDecomp->sectors,i) {
      FORALL (i->nodes,j) {
        Decomp::Node *n = *j;
        
        if (n->type != Decomp::Node::R_NODE) continue;
        
        if (!display_decomposition) {
          if (n->data == 0) continue;
        }

        if (!display_obstacles) {
          if (n->data != 0) continue;
        }
        
	real4 u1 = n->r.x0;
	real4 v1 = n->r.y0;
	real4 u2 = n->r.x1;
	real4 v2 = n->r.y1;

	if (n->r.y1 > new_max_y) {
	  new_max_y = n->r.y1;
	}

	//cout<<u1<<" "<<v1<<" "<<u2<<" "<<v2<<endl;
        
        sint4 h = (sint4)this->h;
        sint4 x = (sint4)this->x;
        sint4 y = (sint4)this->y;
        
	u1 = u1*scale_x + x;
	v1 = v1*scale_y + y;
	u2 = u2*scale_x + x;
	v2 = v2*scale_y + y;
		  
	v1 += h - 2*(v1-y);
	v2 += h - 2*(v2-y);
		  
	glBegin(GL_LINE_LOOP);
	glVertex2d(u1,v1);
	glVertex2d(u2,v1);
	glVertex2d(u2,v2);
	glVertex2d(u1,v2);
	glEnd();
      }
    }
  }

  //draw the graph
  if (display_graph) {
    if (graph!=NULL) {
      glColor3f(0,0,1);
      
      sint4 count = 0;
      
      Point2i p1,p2;
      p1.x = 0;
      p1.y = 0;
      p2.x = world_width;
      p2.y = world_height;
      
      draw_line(p1,p2);
      
      glBegin(GL_LINES);
      
      FORALL (graph->all_nodes,j) {
        //draw all lines connected to the current point
        FORALL ((*j)->neighbours, j0) {

          Graph_node *neighbour = *j0;
          new_max_y = max((*j)->p.y , new_max_y);
          if (neighbour->p.y > new_max_y) new_max_y = neighbour->p.y;
          draw_line((*j)->p, neighbour->p);
          count++;
        }
      }

      glEnd();
    }
  }

  //cout<<"new max y"<<new_max_y<<endl;

  //draw the tunnel
  
  //draw the path
  if (display_path) {
    if (pp!=NULL) {
      FORALL(pp->paths,i) {
        
        glColor3f(1,1,0);
        
        glPointSize(3);
        glBegin(GL_POINTS);
        
        real4 scale_x = (real4)w/(real4)world_width;
        real4 scale_y = (real4)h/(real4)world_height;
        
        Point2i p;
        p.x = (i->unit)->get_int("x");
        p.y = (i->unit)->get_int("y");
        
        //shift it by displacement
        p.x += displacement;
        p.y += displacement;
        
        real4 u1 = p.x;
        real4 v1 = p.y;
        
        sint4 h = (sint4)this->h;
        sint4 x = (sint4)this->x;
        sint4 y = (sint4)this->y;

        u1 = u1*scale_x + x;
        v1 = v1*scale_y + y;
        
        v1 += h - 2*(v1-y);
        
        glVertex2d(u1,v1);
        
        glEnd();
        
        draw_path(i->path);
        
        glColor3f(1,1,0);
#if 0
        FORALL(i->tunnel,j) {
          real4 u1 = j->x0;
          real4 v1 = j->y0;
          real4 u2 = j->x1;
          real4 v2 = j->y1;
          
          u1 = u1*scale_x + x;
          v1 = v1*scale_y + y;
          u2 = u2*scale_x + x;
          v2 = v2*scale_y + y;
          
          v1 += h - 2*(v1-y);
          v2 += h - 2*(v2-y);
          
          glBegin(GL_LINE_LOOP);
          glVertex2d(u1,v1);
          glVertex2d(u2,v1);
          glVertex2d(u2,v2);
          glVertex2d(u1,v2);
          glEnd();
          
        }
#endif
        //draw corridors
        glColor3f(0,1,1);
        
        //cout<<"number of points"<<i->path.size()<<endl;
        //cout<<"number of cells"<<i->tunnel.size()<<endl;
        //cout<<"number of corridors"<<i->doors.size()<<endl;
        glBegin(GL_LINES);
        
        //        cout << i->doors.front().p1.x << " " << i->doors.back().p2.x << endl;
        //        cout<<i->doors.front().p1.y<<" "<<i->doors.back().p2.y<<endl;
        
        if (i->doors.empty()) continue;
        
        if (!display_corridors) continue;
        
        draw_line(i->doors.front().p1,i->doors.front().p2);
        draw_line(i->doors.back().p1,i->doors.back().p2);
        
        //draw_line(i->doors.front().p1,i->doors.back().p1);
        //draw_line(i->doors.front().p2,i->doors.back().p2);
        
        FORALL(i->doors,j) {
          
          //draw_line((*j).p1,(*j).p2);
          
          if (j!=i->doors.begin()) {
            draw_line((*(j-1)).p1,(*(j)).p1);
            draw_line((*(j-1)).p2,(*(j)).p2);
          }
        }
        glEnd();
      }
    }
  }
  glEnable(GL_TEXTURE_2D);
  glDisable(GL_SCISSOR_TEST);
  return 0L;
}


void PWidget::init_world(sint4 new_width, sint4 new_height)
{
    world_width = new_width;
    world_height = new_height;
    //std::cout<<"what "<<world_width<<" "<<world_height<<endl;
}


void PWidget::empty_lines()
{
  lines.erase(lines.begin(), lines.end());
}


void PWidget::insert_line(Segment2i s)
{
  lines.push_back(s);
}


void PWidget::set_decomposition(SectorDecomp* newDecomp)
{
  curDecomp = newDecomp;
}


void PWidget::set_graph(PolyGraph *new_graph)
{
  graph = new_graph;
}


void PWidget::draw_line(Point2i a, Point2i b)
{
  real4 scale_x = (real4)w/(real4)world_width;
  real4 scale_y = (real4)h/(real4)world_height;

  real4 u1 = a.x;
  real4 v1 = a.y;
  real4 u2 = b.x;
  real4 v2 = b.y;

  sint4 h = (sint4)this->h;
  sint4 x = (sint4)this->x;
  sint4 y = (sint4)this->y;

  u1 = u1*scale_x + x;
  v1 = v1*scale_y + y;
    
  u2 = u2*scale_x + x;
  v2 = v2*scale_y + y;
    
  v1 += h - 2*(v1-y);
  v2 += h - 2*(v2-y);
    
  glVertex2d(u1,v1);
  glVertex2d(u2,v2);
}

void PWidget::set_paths(const PathProcessor *new_p)
{
  pp = new_p;
}

void PWidget::draw_path(const Vector<Point2i> &path)
{
  glColor3f(1, 1, 1);
  
  glBegin(GL_LINES);
  
  FORALL (path,i) {
    if (i != (path.end()-1)) {
      Point2i p1 = *i;
      Point2i p2 = *(i+1);
      
      p1.x += displacement;
      p1.y += displacement;
      p2.x += displacement;
      p2.y += displacement;
      
      draw_line(p1,p2);
    }
  }
  glEnd();
}

