// $Id: BaseDefense.C,v 1.1 2006/08/10 17:13:39 nicholas Exp $

// This is an ORTS file (c) Tim Furtak, licensed under the GPL

#include "BaseDefense.H"
#include "Options.H"
#include "GameStateModule.H"
#include "Game.H"
#include "GameChanges.H"
#include "TerrainModule.H"
#include "TerrainBase.H"


using namespace std;

const sint4 BaseDefense::FROM = EventFactory::new_who();

//const std::string SuicideAttackAI::bp_name = "suicide_attack";

//==================================================================

void BaseDefense::add_options(){
  ::Options o("BaseDefense");
  o.put("-BRadius", sint4(128), "'clean zone' radius around bases, that will trigger retaliation");
}

//===================================================================

BaseDefense::BaseDefense(GameStateModule &gsm_) : gsm(gsm_)
{
  Options::get("-BRadius", clean_zone);
}

//===================================================================

BaseDefense::~BaseDefense()
{
}

//===================================================================

// void BaseDefense::stop(GameObj *obj)
// {

//   cout << "stopping object\n";
//   FIND (obj2dat, it, obj);
//   if (it != obj2dat.end()) obj2dat.erase(it);
// }

//===================================================================

void BaseDefense::act(GameStateModule &/*gsm*/)
{
   
  FORALL (need_act, i) {
    FIND (defender2data, it, *i);
    if (it == defender2data.end()) continue;

    DefenderData &d_data = it->second;
   
    switch(d_data.state) {

    case DefenderData::MOVE_TO:
      {
	if (!d_data.base){
	  d_data.base = find_base(*i);
	}
	if (d_data.base){
	  //add the unit to the list of defenders for this base
	  FIND(base2data, it2, d_data.base){
	    if (it2 != base2data.end()){
	      it2->second.defenders.insert(*i);
	    }
	    else{
	      assert(0 && "cannot find base");
	    }
	  }
	  //fixme: replace TerrainBase::Task::ATTACK with
	  //TerrainBase::Task::VICINITY, once it is implemented
	  TerrainBase::Task task(TerrainBase::Goal::OBJ, TerrainBase::Goal::VICINITY,
				 TerrainBase::Task::INDIVIDUAL);
	  task.goal.distance = 16;
	  task.goal.obj = d_data.base;
	  task.objs.insert(*i);
	  call_handlers(TerrainModule::PathEvent(FROM, TerrainModule::FIND_PATH_MSG, task));
	}
	else{
	  std::cout <<"No more bases\n";
	}
      }
      continue;

    case DefenderData::DEFEND:
      {
	std::cout << "Defend Base\n";

	FIND(base2data, base_it, d_data.base);
	if (base_it != base2data.end()){
	  
	  assert(base_it->second.threat && "threat must be non-zero");
	  
	  TerrainBase::Task task(TerrainBase::Goal::OBJ, TerrainBase::Goal::ATTACK,
				 TerrainBase::Task::INDIVIDUAL);
	  
	  task.goal.obj = base_it->second.threat;
	  task.objs.insert(*i);
	  call_handlers(TerrainModule::PathEvent(FROM, TerrainModule::FIND_PATH_MSG, task));
	}
      
	//Shooter Ai should handle target selection:

      }
      continue;
    case DefenderData::WAIT:
      {
	assert (0 && "no action required");
      }
      continue;
    default:
      ERR("??");
    }
  }

  need_act.clear();
}

//===================================================================

/**
 * Perform appropriate action based on what even was passed
 * Returns true if event was handled.
 */
bool BaseDefense::handle_event(const Event &e)
{
  if (e.get_what() == TerrainModule::PATH_SUCCESS) {

    const TerrainModule::PathStatusEvent *p =
      dynamic_cast<const TerrainModule::PathStatusEvent*>(&e);
    assert(p && "SUCCESS: not a PathStatusEvent");
    
    cout << "DefendBase AI got path success\n";
    FIND (defender2data, it, p->get_obj());
    if (it == defender2data.end()) return false;
    
    DefenderData::State &s = it->second.state;
    if (s == DefenderData::MOVE_TO) s = DefenderData::WAIT;    
    return true;
  }

  if (e.get_what() == TerrainModule::PATH_FAILURE) {
    //cout << "DefendBase AI got path failure" << endl;

    const TerrainModule::PathStatusEvent *p =
      dynamic_cast<const TerrainModule::PathStatusEvent*>(&e);
    assert(p && "FAILURE: not a PathStatusEvent");

    FIND (defender2data, it, p->get_obj());
    if (it == defender2data.end()) return false;
    
    DefenderData::State &s = it->second.state;
    
    if (s == DefenderData::MOVE_TO){
      //find a new base that is different from the old one
  
      //fixme: just sending it back on the path it failed... probably not a good idea
      it->second.base = 0;
      need_act.push_back(p->get_obj());
      //act(gsm);
      
    }
    else if (s == DefenderData::DEFEND){
      s = DefenderData::MOVE_TO; //move back to base
      need_act.push_back(p->get_obj());
      //act(gsm);
    }
    return true;
  }


  if (e.get_who() == GameStateModule::FROM) {

    if (e.get_what() == GameStateModule::VIEW_MSG) {

      const sint4 me = gsm.get_game().get_client_player();
      const sint4 npc = gsm.get_game().get_player_num();

      // add new objects
      FORALL (gsm.get_changes().new_objs, i) {
        GameObj *gob = (*i)->get_GameObj();
        if (gob == 0) continue;
        if (*gob->sod.owner == me){
	  //Add units with weapons who can move
	  ScriptObj *weapon = gob->component("weapon");
	  if (weapon && gob->get_max_speed() >0){
	    DefenderData dd;
	    dd.base = 0; //determined when needed
	    dd.state = DefenderData::MOVE_TO;
	    defender2data[gob] = dd;
	    need_act.push_back(gob);
	  }
	  else if(gob->bp_name() == "controlCenter") add_base(gob);
	}
	else if (*gob->sod.owner != npc && gob->bp_name() != "controlCenter"){
	  enemies.insert(gob);
	}
      }
      
      // remove dead objects
      FORALL (gsm.get_changes().dead_objs, i) {
        GameObj *gob = (*i)->get_GameObj();
        if (gob == 0) continue;
        if (*gob->sod.owner == me){
	  if (gob->bp_name() == "controlCenter")
	    del_base(gob);
	  else{
	    del_unit(gob);
	  }
	}
	else{
	  del_enemy_unit(gob);
	}
      }
      
      // compute actions
      update_targets();
      act(gsm);

      return true;
    }
  }

  return false;
}

//===================================================================

void BaseDefense::add_base(GameObj *gob) {
  BaseData bd;
  bd.threat = 0;
  base2data[gob] = bd;
  
//   FORALL(obj2dat, it){
//     SuicideData &s_data = it->second;
//     GameObj *target = find_target(it->first);
//     if (s_data.base != target){
//       s_data.base = target;
//       s_data.state = SuicideData::MOVE_TO;
//       need_act.push_back(it->first);
//     }
//   }

}

//===================================================================
void BaseDefense::del_base(GameObj *gob) {
  FIND(base2data, it, gob);
  if (it != base2data.end()){
    //change the state of all the defenders of the deleted base
    FORALL(it->second.defenders, i){
      FIND(defender2data, it2, *i);
      if (it2 != defender2data.end()){
	it2->second.base = 0;
	it2->second.state = DefenderData::MOVE_TO;
	need_act.push_back(it2->first);
	
      }
    }
    base2data.erase(it);
  }
}

//===================================================================

void BaseDefense::del_unit(GameObj *gob){
  FIND(defender2data, it, gob);
  if (it != defender2data.end()){
    FIND(base2data, it2, it->second.base);
    if (it2 != base2data.end()) 
      it2->second.defenders.erase(gob);
      
    defender2data.erase(it);
    
  }
}

void BaseDefense::del_enemy_unit(GameObj *gob){
  FORALL(base2data, i){
    if (i->second.threat == gob){
      i->second.threat = 0;
    }
  }
  enemies.erase(gob);
}


//===================================================================

void BaseDefense::update_targets(){
  Game &game = gsm.get_game();
  FORALL(base2data, i){
    //If the old threat is gone - nothing needs to be done if the old threat is still present
    if (!(i->second.threat) || game.distance(*(i->first), *(i->second.threat)) > clean_zone){
      
      //find a new threat
      i->second.threat = find_threat(i->first);

      //Alter the states of all defenders appropriately
      FORALL(i->second.defenders, j){
	FIND(defender2data, it, *j);
	if (it != defender2data.end()){
	  
	  //if no new threat was found
	  if (!(i->second.threat)){
	    if (it->second.state == DefenderData::DEFEND){
	      it->second.state = DefenderData::MOVE_TO;
	      need_act.push_back(it->first);
	    }
	  }
	  //if a new threat was found
	  else{
	    std::cout << "threat found!\n";
	    it->second.state = DefenderData::DEFEND;
	    need_act.push_back(it->first);
	  }
	}
      }
    }
  }
}

GameObj* BaseDefense::find_threat(GameObj *base){
  Game &game = gsm.get_game();
  GameObj *threat = 0;
  real8 min_distance = 99999999; 
  real8 temp;
  FORALL(enemies, i){
    temp = game.distance(*base, **i);
    if (temp <= clean_zone){
      if (!threat || (min_distance > temp)){
	threat = *i;
	min_distance = temp;
      }
    }
  }
  return threat; //returns 0 if no threat found
}

/**
 * Find and set a target for the targeter. If there are no more 
 * enemy command centers left, remove the targeter from the 
 */
GameObj* BaseDefense::find_base(GameObj *targeter, GameObj *baseToAvoid){
  Game &game = gsm.get_game();
  GameObj *target = 0, *baseObj;
  real8 min_distance = 999999999; //set to something to avoid warning... still safe, I think
  real8 temp;
  FORALL(base2data, i){
    baseObj = i->first;
    if (baseObj != baseToAvoid){
      temp = game.distance(*targeter, *baseObj);
      if (!target || (temp < min_distance && temp >= 0)){
	min_distance = temp;
	target = baseObj;
      }
    }
  }
  
  if (target){
    cout << "new target found.\n";
    return target;
  }
  cout << "no more targets found.\n";
  return 0;

}

