// $Id: GatherAI.C,v 1.16 2007/05/01 15:51:27 orts_furtak Exp $

// This is an ORTS file (c) Tim Furtak, licensed under the GPL

#include "GatherAI.H"
#include "GameStateModule.H"
#include "Game.H"
#include "GameChanges.H"
#include "GameObj.H"
#include "GfxModule.H"
#include "TerrainModule.H"

using namespace std;

const sint4 GatherAI::FROM = EventFactory::new_who();
const sint4 GatherAI::GATHER_START = EventFactory::new_what();
const sint4 GatherAI::GATHER_STOP = EventFactory::new_what();

const std::string GatherAI::bp_name = "gather";

//===================================================================

GatherAI::GatherAI(GameStateModule &gsm_) : gsm(gsm_)
{

}

//===================================================================

GatherAI::~GatherAI()
{
}

//===================================================================

void GatherAI::stop(GameObj *obj)
{
  FIND (obj2dat, it, obj);
  if (it != obj2dat.end()) obj2dat.erase(it);
  FIND (mining, it2, obj);
  if (it2 != mining.end()) mining.erase(it2);

  FIND (returning, it3, obj);
  if (it3 != returning.end()) returning.erase(it3);
}

//===================================================================

void GatherAI::act(GameStateModule &/*gsm*/)
{
  //------------------------------

  for (set<GameObj*>::iterator i = mining.begin(); i != mining.end();) {
    GameObj *g = *i;
    ScriptObj *px = g->component("pickaxe");
    ScriptObj *gx = g->component("gastank");
    if ((!px || px->get_int("active") == 0) && (!gx || gx->get_int("active") == 0)) {
      need_act.push_back(g);
      GatherData &gd = obj2dat[g];

      gd.state = GatherData::MOVE_BACK;
      
      FIND(resources, r_it, gd.res);
      if (r_it != resources.end()){
 
	if (r_it->second.size() > 0){
	  GameObj* gob = r_it->second.front();
	  r_it->second.pop();
	  map<GameObj*, GatherData>::iterator map_it = obj2dat.find(gob);
	  while (map_it == obj2dat.end() && r_it->second.size() > 0){
	    gob = r_it->second.front();
	    r_it->second.pop();
	    map_it = obj2dat.find(gob);
	  }
	  if (map_it != obj2dat.end()){
	    map_it->second.state = GatherData::GATHER;
	    need_act.push_back(gob);
	  }
	}
      }
      set<GameObj*>::iterator j = i;
      ++i;
      mining.erase(j);
      continue;
    }
    ++i;
  }

  //------------------------------

  for (set<GameObj*>::iterator i = returning.begin(); i != returning.end();) {
    GameObj *g = *i;
    const sintptr *mx = g->get_int_ptr("minerals");
    const sintptr *gx = g->get_int_ptr("gas");
    if ((!mx || *mx == 0) && (!gx || *gx == 0)) {
      need_act.push_back(g);
      obj2dat[g].state = GatherData::MOVE_TO;
      set<GameObj*>::iterator j = i;
      ++i;
      returning.erase(j);
      continue;
    }
    ++i;
  }
  //------------------------------


  FORALL (need_act, i) {
    FIND (obj2dat, it, *i);
    if (it == obj2dat.end()) continue;

    GatherData &gd = it->second;

    switch(gd.state) {

    case GatherData::MOVE_TO:
      {
// 	TerrainBase::Task task;
// 	task.goal.target = TerrainBase::Goal::OBJ;
// 	task.goal.obj = gd.res;
// 	task.goal.mode = TerrainBase::Goal::VICINITY;
// 	task.goal.distance = (*i)->component("pickaxe")->get_int("max_range");
// 	task.objs.insert(*i);
// 	call_handlers(TerrainModule::PathEvent(FROM, TerrainModule::FIND_PATH_MSG, task));

	if (resources.find(gd.res) == resources.end()) {
          stop(*i);
          cout << "no resource?" << endl;
          continue;
        }

	call_handlers(TerrainModule::PathEvent(FROM, TerrainModule::FIND_PATH_MSG, 
					       TerrainBase::Obj2ObjTask(*i,gd.res)));

	
      }
      continue;

    case GatherData::GATHER:
      {
        //cout << "gather ai -> pickaxe" << endl;
        string bp = gd.res->bp_name();
        //        sint4 code = gd.res->code();
        if (resources.find(gd.res) == resources.end()) {
          stop(*i);
          cout << "no resource?" << endl;
          continue;
        }

	if (is_extractor(gd.res)
	    && gd.res->get_int("num_miners") >= gd.res->get_int("max_miners")){
	  gd.state = GatherData::QUEUE;
	  resources[gd.res].push(it->first);
	}
	else{
	 
	  Vector<sint4> args;
	  args.push_back(gsm.get_game().get_cplayer_info().get_id(gd.res));
          //	  it->first->set_action("gather", args);
          it->first->component("pickaxe")->set_action("mine", args);
	  mining.insert(*i);
	}
      }
      continue;

    case GatherData::MOVE_BACK:
      {
        cout << "gather ai -> move back" << endl;
        string bp = gd.res->bp_name();
        //        sint4 code = gd.res->code();
        //ResBase &rb = res_map[bp];
        if (!gd.base || gd.last_updated < bases.last_updated){
	  gd.base = choose_base(it->first);
          gd.last_updated = gsm.get_game().get_view_frame();
	}
        if (!gd.base) {
	  cout << "no base found\n" ;
	  continue;
	}
// 	TerrainBase::Task task;
// 	task.goal.target = TerrainBase::Goal::OBJ;
// 	task.goal.obj = gd.res;
// 	task.goal.mode = TerrainBase::Goal::VICINITY;
// 	task.goal.distance = (*i)->get_int("return_range");
// 	task.objs.insert(*i);
// 	call_handlers(TerrainModule::PathEvent(FROM, TerrainModule::FIND_PATH_MSG, task));
	call_handlers(TerrainModule::PathEvent(FROM, TerrainModule::FIND_PATH_MSG, 
	                                      TerrainBase::Obj2ObjTask(*i, gd.base)));
      }
      continue;

    case GatherData::RETURN:
      {
        cout << "gather ai -> return" << endl;
        string bp = gd.res->bp_name();
        //        sint4 code = gd.res->code();
        //ResBase &rb = res_map[bp];
        if (bases.bases.find(gd.base) == bases.bases.end()) {
          stop(*i);
          continue;
        }
        Vector<sint4> args;
        args.push_back(gsm.get_game().get_cplayer_info().get_id(gd.base));
        it->first->set_action("return_resources", args);
        returning.insert(*i);
      }
      continue;
    case GatherData::QUEUE:
      {
	cout << "Queued Unit. Why are we here?\n";
	continue;
      }
    default:
      ERR("??");
    }
  }

  need_act.clear();
}

//===================================================================


bool GatherAI::handle_event(const Event &e)
{
  if (e.get_what() == TerrainModule::PATH_SUCCESS) {

    const TerrainModule::PathStatusEvent *p =
      dynamic_cast<const TerrainModule::PathStatusEvent*>(&e);
    assert(p && "SUCCESS: not a PathStatusEvent");

    FIND (obj2dat, it, p->get_obj());
    if (it == obj2dat.end()) return false;

    GatherData::State &s = it->second.state;
    if (s == GatherData::MOVE_TO) s = GatherData::GATHER;
    if (s == GatherData::MOVE_BACK) s = GatherData::RETURN;
    need_act.push_back(p->get_obj());

    return true;
  }

  if (e.get_what() == TerrainModule::PATH_FAILURE) {
  
    const TerrainModule::PathStatusEvent *p =
      dynamic_cast<const TerrainModule::PathStatusEvent*>(&e);
    
    if (p->get_details() == TerrainModule::PATH_MOVE_FAILURE){
      if (obj2dat.find(p->get_obj()) != obj2dat.end()){
	cout << "gather ai got path move failure - retrying path\n";
	need_act.push_back(p->get_obj());
      }
    }
    else if (p->get_details() == TerrainModule::PATH_PLAN_FAILURE){
      assert(p && "FAILURE: not a PathStatusEvent");
      if (obj2dat.find(p->get_obj()) != obj2dat.end()){
	cout << "gather ai got path plan failure\n";
      }
      stop(p->get_obj());
      return true;
    }
  }

  if (e.get_who() == GfxModule::FROM) {
    /*
    if (e.get_what() == GfxModule::ADD_CWIDGETS_MSG) {
      init_widget();
      return true;
    }
    */

    if (e.get_what() == GatherAI::GATHER_START) {
      REM("GATHER EVENT");
      const GatherEvent &ge = static_cast<const GatherEvent&>(e);
      if (ge.obj && ge.targ) {
	string targ_name = ge.targ->bp_name();
	assert (targ_name == "mineral" || targ_name == "extractor"
		|| targ_name == "pantheon_well" || targ_name == "reservoir");
        GatherData gd;
        gd.res = ge.targ;
        gd.base = 0; // compute later
        gd.state = GatherData::MOVE_TO;
        obj2dat[ge.obj] = gd;
        need_act.push_back(ge.obj);
      }
      return true;
    }

    if (e.get_what() == GatherAI::GATHER_STOP) {
      REM("STOP GATHER EVENT");
      const GatherEvent &ge = static_cast<const GatherEvent&>(e);
      stop(ge.obj);
      return true;
    }
  }

  if (e.get_who() == GameStateModule::FROM) {

    if (e.get_what() == GameStateModule::VIEW_MSG) {

      const sint4 me = gsm.get_game().get_client_player();
      
      // add new objects
      FORALL (gsm.get_changes().new_objs, i) {
        GameObj *gob = (*i)->get_GameObj();
        if (gob == 0) continue;
        if (gob->bp_name() == "mineral")       add_res(gob);
	if (is_extractor(gob)){
	  cout << "added extractor\n";
	  add_res(gob);
	} 
        if (*gob->sod.owner != me) continue;
	//	if (is_extractor(gob)){
	//  cout << "added extractor\n";
	//  add_res(gob);
	//} 
        if (gob->bp_name() == "controlCenter"
	    || gob->bp_name() == "pantheon_acropolis"
	    || gob->bp_name() == "mound"
            || gob->bp_name() == "tor"
	    || gob->bp_name() == "citadel") add_base(gob);
      }
      
      // remove dead objects
      FORALL (gsm.get_changes().dead_objs, i) {
        GameObj *gob = (*i)->get_GameObj();
        if (gob == 0) continue;
        if (gob->bp_name() == "mineral")       del_res(gob);
        if (*gob->sod.owner != me) continue;
        stop(gob);
	if (is_extractor(gob)){
	  del_res(gob);
	}
	if (gob->bp_name() == "controlCenter"
	    || gob->bp_name() == "pantheon_acropolis"
	    || gob->bp_name() == "mound"
            || gob->bp_name() == "tor"
	    || gob->bp_name() == "citadel") del_base(gob);
      }
      
      // remove all vanished objects
      //fixme: do something different for vanished objects?
      FORALL (gsm.get_changes().vanished_objs, i) {
	GameObj *gob = (*i)->get_GameObj();
        if (gob == 0) continue;
        if (gob->bp_name() == "mineral")       del_res(gob);
        if (*gob->sod.owner != me) continue;
        stop(gob);
	if (is_extractor(gob)){
	  del_res(gob);
	}
	if (gob->bp_name() == "controlCenter"
	    || gob->bp_name() == "pantheon_acropolis"
	    || gob->bp_name() == "mound"
            || gob->bp_name() == "tor"
	    || gob->bp_name() == "citadel") del_base(gob);
      }
      // compute actions
      act(gsm);

      return true;
    }
  }

  return false;
}

//===================================================================

void GatherAI::add_res(GameObj *gob) {

  resources[gob] = std::queue<GameObj*>();
  // res_map[gob->bp_name()].resources.insert(gob);
}

void GatherAI::del_res(GameObj *gob) {

  resources.erase(gob);
  // res_map[gob->bp_name()].resources.erase(gob);
  
}
void GatherAI::add_base(GameObj *gob){
  bases.bases.insert(gob);
  bases.last_updated = gsm.get_game().get_view_frame();
}
//void GatherAI::add_base(const string &res, GameObj *gob) {
//  res_map[res].bases.insert(gob);
//  res_map[res].last_updated = gsm.get_game().get_view_frame();
//}

void GatherAI::del_base(GameObj *gob){
  bases.bases.erase(gob);
  bases.last_updated = gsm.get_game().get_view_frame();
}
//void GatherAI::del_base(const string &res, GameObj *gob) {
  //fixme: this is untested... I couldn't find what erase actually returns (it returns a size_type, but I don't know what it signifies) so I'm assuming it returns 0 if it doesn't erase anything.
// if (res_map[res].bases.erase(gob))
//    res_map[res].last_updated = gsm.get_game().get_view_frame();
//}

//===================================================================

GameObj* GatherAI::choose_base(GameObj* gatherer){
  if (bases.bases.empty()) return NULL;
  
  Game &game = gsm.get_game();
  GameObj *target = 0, *baseObj;
  real8 min_distance = -1; //set to -1 to avoid warning... still safe, I think
  real8 temp;
  FORALL(bases.bases, i){
    baseObj = *i;
    temp = game.distance(*gatherer, *baseObj);
    if (!target || (temp < min_distance && temp >= 0)){
      min_distance = temp;
      target = baseObj;
    }
    
  }
  return target; //returns 0 if no base was found

}

//===================================================================

bool GatherAI::is_extractor(GameObj* gob){
  string bp = gob->bp_name();
  return (bp == "extractor" || bp == "pantheon_well" || bp == "reservoir");
}

//===================================================================
