
#include "Vanished.H"
#include "GameStateModule.H"
#include "Game.H"
#include "GameChanges.H"

using namespace std;

//===================================================================

/** Look at all of the static objects and see if they should be visible.
    Does not care if the object is still there or not.  If it still exists
    it will have been sent again.
*/

VanishedObjects::VanishedObjects(GameStateModule &gsm_) : gsm(gsm_)
{
}

//-------------------------------------------------------------------

void VanishedObjects::add(GameObj *gob)
{
  static_objects.insert(ScriptObj::Ptr(gob));
}

//-------------------------------------------------------------------

void VanishedObjects::del(GameObj *gob)
{
  ScriptObj::Ptr ptr(gob);
  static_objects.erase(ptr);
}

//-------------------------------------------------------------------

void VanishedObjects::check_static_objects()
{
  Game &game = gsm.get_game();

  sint4 tile_points = game.get_tile_points();

  const Map<GameTile> &pf = game.get_map();
  PlayerInfo &pi = game.get_cplayer_info();
  
  Vector<ScriptObj::Ptr> to_remove;

  FORALL (static_objects, s) {

    bool visible = false;

    GameObj *gob = (*s)->get_GameObj();
    if (!gob) { to_remove.push_back(*s); continue; }

    if (*gob->sod.x < 0 ||
        *gob->sod.y < 0 ||
        *gob->sod.x / tile_points >= pf.get_width() ||
        *gob->sod.y / tile_points >= pf.get_height()) continue;

    //    if (*gob->sod.cloaked2 >= 1) continue;
    
    MapView *mapW;
    MapView *mapE;
    
    if (*gob->sod.cloaked >= 1) {
      mapW = &pi.detected_W; mapE = &pi.detected_E;
    } else {
      mapW = &pi.visible_W;  mapE = &pi.visible_E;
    }
    
    Vector<sint4> obj_tiles_W;
    Vector<sint4> obj_tiles_E;
    
    game.compute_footprint(*gob, obj_tiles_W, obj_tiles_E);
    
    // check if any part of the footprint is visible
    // => the object would be visible if it didn't move

    FORALL (obj_tiles_W, i) {
      sint4 x = pf.ind2x(*i);
      sint4 y = pf.ind2y(*i);
      
      if (mapW->is_set(x, y)) {
        visible = true;
        break;
      }
    }

    if (!visible) {
      FORALL (obj_tiles_E, i) {
	sint4 x = pf.ind2x(*i);
	sint4 y = pf.ind2y(*i);
        
	if (mapE->is_set(x, y)) {
	  visible = true;
	  break;
	}
      }
    }

    if (visible) to_remove.push_back(*s);
  }
  
  FORALL (to_remove, it) {
    static_objects.erase(*it);
  }
}

//===================================================================
