#ifndef BTASK_H
#define BTASK_H


#include "Global.H"
#include "LowAI.H"
#include "EventHandler.H"
#include "ClientModule.H"

#include "TerrainBase.H"

namespace BSimul {


  class Task;
  class BuildTasksState;

//typedef Vector<uint4> GroupVector;
  typedef Vector<Task*> TaskVector;

  typedef Vector<GameObj*> ObjectVector;

//typedef Task* (*TaskCreateFunctionPtr)(uint4 id,GameState* s,GroupVector& g);
  typedef Task* (*TaskCreateFunctionPtr)(uint4 id,BuildTasksState* s,ObjectVector& o,real8 timeout);

  // timeout refers to time it would take to complete moving a certain distance
  class MoveStruct
  {
  public:

    MoveStruct(const TerrainBase::Loc& p) {
      pos = p;
      obj = NULL;
    }
    MoveStruct(GameObj* object) {
      obj = object;
    }
    MoveStruct* clone() { return new MoveStruct(*this); }
    TerrainBase::Loc pos;
    GameObj* obj;

  private:
    MoveStruct();
    MoveStruct &operator=(const MoveStruct &);
  };



  class Task
  {
  public:

    Task();
    /*
    Task(Task* parent,GroupVector& origGroups,TaskCreateFunctionPtr function,std::string name,
	 bool baseAttackPreference = false,MoveStruct* extraData = NULL);
    Task(uint4 playerID,GameState* state,GroupVector& origGroups,TaskCreateFunctionPtr function,std::string name,
	 bool baseAttackPreference = false,MoveStruct* extraData = NULL);
*/
    Task(Task* parent,ObjectVector& origObjects,TaskCreateFunctionPtr function,std::string name,real8 timeout,
         MoveStruct* extraData = NULL);
    Task(uint4 playerID,BuildTasksState* state,ObjectVector& origObjects,TaskCreateFunctionPtr function,
         std::string name,real8 timeout,MoveStruct* extraData = NULL);

    virtual ~Task();

    virtual Task* clone(BuildTasksState* state,Task* parent)=0;

    BuildTasksState* getSimulationState(){ return state;}
    Task* getParent(){ return parent;}

    //GroupVector& getGroups(){ return groups;}
    ObjectVector& getObjects() { return objects;}

    uint4 getPlayerID() const {return playerID;}

    void setLineNum(uint4 lineN) { lineNum = lineN;}
    uint4 getLineNum() const { return lineNum;}

    TaskCreateFunctionPtr getFunction(){return function;}
    std::string getName() const {return name;}

    void removeObject(GameObj* obj);
    void childCompleted(Task* child);
    bool update();

    void setTimeout(real8 time){timeout = time;}
    real8 getTimeout() const{return timeout;}

    void setStartTime(real8 time){startTime = time;}
    real8 getStartTime() const {return startTime;}

    virtual bool exec()=0;

  protected:

    Task* parent;
    TaskVector children;

    BuildTasksState* state;

    ObjectVector objects;

    uint4 playerID;
    uint4 lineNum;
    MoveStruct* extraData;

    TaskCreateFunctionPtr function; // right now, ONLY used for identification!

    Task(const Task&);
    std::string name;

    real8 startTime;
    real8 timeout;

  private:
    Task &operator=(const Task &);
  };

}

#endif

