
#include "SType.H"
#include "PlayerInfo.H"
#include "ByteData.H"

using namespace std;

//===================================================================

void stype::gen_diff(const stype *orig, bool allied, string &u, PlayerInfo *pi) const
{
  if (!allied || type == NONE) {
    u += char(0x40); // "-0" (aka not a number)
    return;
  }
  
  if (orig) {
    write_b_int4(u, type - orig->type);
  } else {
    write_b_int4(u, type);
  }
  
  switch (type) {
    
  case OBJ:
    {
      ObjId id = pi->get_id(val.o);
      write_b_int4(u, id);
    }
    break;
    
  case CSTR:
  case ADDR:
    SCRIPT_ASSERT(0 && "Type cannot be sent");
    break;
    
  case INT:
  case DIR:
  default:
    write_b_int4(u, val.i);
    break;
  }
}

//-------------------------------------------------------------------

bool stype::apply_diff(bool first, ByteStream &s, PlayerInfo *pi)
{
  bool changed = false;
  
  if (s.peek_b_char() == 0x40) {
    changed = type != NONE;
    val.i = UNDEF_INT;
    type = NONE;
    s.advance();
    return changed;
  }
  
  sint4 dtype = s.read_b_int4();
  if (first) {
    type = (Type)dtype;
  } else {
    type = (Type)((sint4)(type) + dtype);
  }
  changed |= dtype != 0;
  
  sint4 dval = s.read_b_int4();
  
  switch (type) {
    
  case OBJ:
    {
      ScriptObj *obj = pi->get_obj(dval);
      changed |= obj != val.o;
      val.o = obj;
    }
    break;
    
  case CSTR:
  case ADDR:
    SCRIPT_ASSERT(0 && "Type cannot be received");
    break;
    
  case INT:
  case DIR:
  default:
    changed |= val.i != dval;
    val.i = dval;
    break;
  }

  return changed;
}

//===================================================================

string stype::print_va_list(const char *format, const Vector<stype> &v, sint4 n)
{
  boost::format f(format);
  FORALL (v, it)
    if (it - v.begin() >= n)
      f % it->to_int();
  return f.str();
}

//-------------------------------------------------------------------

string stype::my_printf(const char *format, const Vector<stype> &v, sint4 n)
{
  string f(format);
  stringstream g;

  uint4 m = (uint4)n;

  size_t p;
  while ((p = f.find('%', 0)) != string::npos) {
    if (p != 0) {
      string sub = f.substr(0, p);
      char *temp = new char[p+1];
      snprintf(temp, p+1, sub.c_str());
      g << string(temp);
      delete [] temp;
    }
    if (p < f.size()-1) {
      char t = f[p+1];
      switch (t) {
      case '%': g << '%'; break;
      case 'd': SCRIPT_ASSERT(m < v.size()); g << v[m++].to_int(); break;
      case 'c': SCRIPT_ASSERT(m < v.size()); g << (char)v[m++].to_int(); break;
      case 's': SCRIPT_ASSERT(m < v.size()); g << v[m++].to_cstr(); break;
      default: SCRIPT_ASSERT(0 && "Unsupported type specifier (sorry)");
      }
      f = f.substr(p+2, f.size()-(p+2));
    } else {
      SCRIPT_ASSERT(0 && "Ending a formatting string with '%'?");
    }
  }
  return g.str();
}

//===================================================================

