#ifndef OBJMNGER_H
#define OBJMNGER_H

#include "ScriptVector.H"  //For object movement.

#include "OgreMD3Model.H"  //includes OgreGfxAnimationSet
#include <OgreStringConverter.h>
#include <list>
#include <map>
#include <iostream>

/**======================================================================================
Class ObjectManager;
This is a SINGLETON class that should be used to hold all information of all the objects.
Right now it holds a list of BasicObjects(objLst) which should be changed to OgreMD3Models
once that is working with animation.  As everything in orts is a md3 except the terrain.
=======================================================================================*/
class ObjectManager{
public:
	/**The Ogre SceneManager where nodes/objects are added to become visible.*/
	Ogre::SceneManager* gdSceneMgr;

	/**TEMP: placeholder jsut to test the OgreMD3Model. This loads and stores the mech. should become an entire list.*/
	OgreMD3Model* mech;   

	/**Every object in Ogre has string name &_<number> this is the last name number added.*/
	Ogre::Real currentHighestNameNumber; 

	/**This will generate a random name for an entity in ogre as every obj must be unique name.*/
	Ogre::String generateUniqueName();   

	/**bp(script) to the mesh name map.*/
        std::map<std::string, std::string> bpName2MeshNameMap;   
	
	/**function which adds objected*/
	BasicObject* addObject(Ogre::String objName, GameObj *gameObj);  
	
	/**this is called every frame and used to update obj's.*/
        bool frameUpdate(const Ogre::FrameEvent& evt);  

	/**used during object selection(shift + click etc)*/                
	bool clearSelectedObjects();	

	/**used during object selection(shift + click etc)*/
	bool addSelectedObject();	
	
        BasicObject* getObjectByName(Ogre::String name);  
	BasicObject* getObjectByGameObj( GameObj *gob );
	bool addSelectedObject(BasicObject *tempObj);
        /**used during object selection(shift + click etc)*/
	bool removeSelectedObject(BasicObject *tempObj);

        /**used during object selection(shift + click etc)*/
	bool doesSelectedObjectExist(BasicObject *tempObj);

	/**Tells an object to move to required location*/
	bool processMovementRequest( Ogre::Vector3 location );

	/**if there is a change in the GameObj(through process changes, then
	this function is called and appropriate action is take */
        bool updateChangedObject( GameObj *gob );    

	/**singleton: use ObjectManager::getSingleton().<function>*/
	static ObjectManager& getSingleton(){
		static ObjectManager _myObjM;
		return _myObjM;
	}

private:
	/**List of all objects currently in the object manager*/
	std::list< BasicObject* > objList;
	/**List of all object currently selected(accepting commands from input)*/
	std::list< BasicObject* > selectedObjList;
	ObjectManager();
};

#endif
