#include "OgreGfxModuleTerrainTileGrid.H"

//Create a completely empty tile grid
OgreGfxModuleTerrainTileGrid::OgreGfxModuleTerrainTileGrid(){
	rows = 0;
	columns=0;
	startNode = NULL;
	tileWidth = 0;
	tileHeight= 0;
	revealed = NULL;
	visible = NULL;
	mSceneMgr = NULL;
	megaTiles = NULL;
}


//Retrieve all our allocated memory
OgreGfxModuleTerrainTileGrid::~OgreGfxModuleTerrainTileGrid(){
	if(revealed!= NULL)
		free(revealed);
	if(visible != NULL)
		free(visible);
	//if(mapVals != NULL)
	//	free(mapVals);
	if(megaTiles != NULL){
		for(int i = 0; i<nRows; i++){
			free(megaTiles[i]);
		}
		free(megaTiles);
	}
}

//Initialize a tile grid with numr rows and numc columns of megatiles
OgreGfxModuleTerrainTileGrid::OgreGfxModuleTerrainTileGrid(Ogre::SceneManager* sceneManager, Ogre::SceneNode* startingNode,int mr,int mc,int numr,int numc,float w,float h)
{
	//---Initializations
	rows = mr*numr;//total number of rows
	columns = mc*numc;
	nRows = numr;//number of megatile rows
	nCols = numc;
	mRows = mr;//how many rows a megatile has
	mCols = mc;
	startNode = startingNode;//node representing what everything is relative to
	tileWidth = w;//info regarding how much to scale a tile
	tileHeight= h;
	mSceneMgr = sceneManager;//always gotta have a reference to the scene manager
	bRebuildNeeded= true;//our terrain is probably going to need rebuilding
	bFogOfWarRefreshNeeded= true;//our fog of war is going to need refresing
	entityName = "grassyTile";//a default name for all our terrain tiles
	tileCount = 0;
	
	//---Create and attach a border for the terrain map
	Ogre::Entity *borderEnt = mSceneMgr->createEntity( "Battlefield Border", "outline.mesh" );
	startNode->attachObject( borderEnt );
	startNode->setScale(rows*w/10.0f,1.0,columns*h/10.0f);

	//Ogre::SceneNode* miniNode = mSceneMgr->getRootScennVerticeseNode()->createChildSceneNode("miniNode",Ogre::Vector3(600,50,600));
	//Ogre::Entity* minifog = mSceneMgr->createEntity("Mini-fog", "flatSheet.mesh");
	//miniNode->scale(200,50,200);
	//miniNode->attachObject(minifog);

	//---Create the two fog of war arrays...
	revealed = (bool**)malloc(rows*sizeof(bool*));
	visible =  (bool**)malloc(rows*sizeof(bool*));
	for(int i = 0; i<rows; i++){
		revealed[i] = (bool*)malloc(columns*sizeof(bool));
		visible[i]  = (bool*)malloc(columns*sizeof(bool));
	}
	//...and fill them with "false"Wno-missing-declarations
	for(int i = 0; i<rows; i++){
		for(int j=0; j<columns; j++){
			revealed[i][j]= false;
			visible[i][j] = false;
		}
	}

	alphaVals = (int**)malloc(rows*sizeof(int*));
	
	for(int i = 0; i<rows; i++)
		alphaVals[i] = (int*)malloc(columns*sizeof(int));
	
	for(int i = 0; i<rows; i++)
		for(int j=0; j<columns; j++)
			alphaVals[i][j] = 245;

	//---Build an array of OgreGfxModuleMegaTiles as specified
	megaTiles = (OgreGfxModuleMegaTile**)malloc(nRows*sizeof(OgreGfxModuleMegaTile*));
	for(int i = 0; i<nRows; i++)
		megaTiles[i] = (OgreGfxModuleMegaTile*)malloc(nCols*sizeof(OgreGfxModuleMegaTile));
	for(int i = 0; i<nRows; i++)
		for(int j=0; j<nCols; j++)
			megaTiles[i][j] = OgreGfxModuleMegaTile(i*mRows,j*mCols,mRows,mCols,tileWidth,tileHeight,revealed,visible,mSceneMgr);
	
	prepareTerrainMaterial("Fog of War Texture", "Terrain Material");

	//---Lastly, instruct that the terrain should be rebuilt, and the fog of war refreshed
	refreshFogOfWar();
	refreshGeometry();
	//minifog->setMaterialName(terrainMaterialName);
}

//---Create a material for the terrrain
void OgreGfxModuleTerrainTileGrid::prepareTerrainMaterial(Ogre::String fogOfWarName, Ogre::String landscapeName){

	terrainMaterial = Ogre::MaterialManager::getSingleton().create(
		landscapeName, // name
		Ogre::ResourceGroupManager::DEFAULT_RESOURCE_GROUP_NAME);

	terrainMaterialName = landscapeName;

	terrainMaterial->getTechnique(0)->getPass(0)->createTextureUnitState("terrain.tga");
	//terrainMaterial->getTechnique(0)->getPass(0)->setSceneBlending(Ogre::SBT_ADD);
	
	prepareFogOfWar(rows, columns, fogOfWarName);
	
	return;
}

//---Creates a completely black, unrevealed fog of war
void OgreGfxModuleTerrainTileGrid::prepareFogOfWar(unsigned int widthInPixels,unsigned int heightInPixels,Ogre::String texturesName){
	
	fogOfWarTexturePixelsWidth = widthInPixels;
	fogOfWarTexturePixelsHeight=heightInPixels;
	
	//---Create the texture
	fogOfWartexture = Ogre::TextureManager::getSingleton().createManual(
		texturesName, // name
		Ogre::ResourceGroupManager::DEFAULT_RESOURCE_GROUP_NAME,
		Ogre::TEX_TYPE_2D,      // type
		widthInPixels, heightInPixels,         // width & height
		0,                // number of mipmaps
		(Ogre::PF_BYTE_BGRA),     // pixel format
		(Ogre::TU_DYNAMIC_WRITE_ONLY_DISCARDABLE));      //should be TU_DYNAMIC_WRITE_ONLY_DISCARDABLE for
						// textures updated very often (e.g. each frame)

	// Get the pixel buffer
	Ogre::HardwarePixelBufferSharedPtr pixelBuffer = fogOfWartexture->getBuffer();

	// Lock the pixel buffer and get a pixel box
	pixelBuffer->lock(Ogre::HardwareBuffer::HBL_DISCARD); // for best performance use HBL_DISCARD not HBL_NORMAL!
	const Ogre::PixelBox& pixelBox = pixelBuffer->getCurrentLock();

	Ogre::uint8* pDest = static_cast<Ogre::uint8*>(pixelBox.data);

	//Fill in the pixel data. This will give the texture a solid black colouration
	for (size_t j = 0; j < heightInPixels; j++)
		for(size_t i = 0; i < widthInPixels; i++)
		{
			*pDest++ =   0; // B
			*pDest++ =   0; // G
			*pDest++ =   0; // R
			*pDest++ = 255; // A
		}

	// Unlock the pixel buffer
	pixelBuffer->unlock();

	//---Add the fog of war texture to the terrain material
	terrainMaterial->getTechnique(0)->getPass(0)->createTextureUnitState(texturesName, 1);
	//terrainMaterial->getTechnique(0)->getPass(0)->createTextureUnitState("fish.png", 1);
	terrainMaterial->getTechnique(0)->getPass(0)->getTextureUnitState(1)->setColourOperation(Ogre::LBO_ALPHA_BLEND);
	
	return;
}


void OgreGfxModuleTerrainTileGrid::setVisibility(int r, int c, bool bVisible, bool bRevealed){
	bFogOfWarRefreshNeeded = true;
	visible[r][c] = bVisible;
	revealed[r][c]= bRevealed;
}


/*
N = North, NE = North-east, SW = South-west, etc...
USE_1 = U-corridinate of the South-east Corner of first UVW map.
*/
Ogre::Entity* OgreGfxModuleTerrainTileGrid::createNewTileWithTheseVerticies(float vertices[], int nVertices){
  
  //Make sure the meshes have a unique identifying name
  tileCount++;
  Ogre::String uniqueMeshName = "terrainMesh" + Ogre::StringConverter::toString(tileCount);
  Ogre::String uniqueEntityName="terrainEntity"+Ogre::StringConverter::toString(tileCount);
  
  /// Create the mesh via this convienent MeshManager
  Ogre::MeshPtr msh = Ogre::MeshManager::getSingleton().createManual(uniqueMeshName, "General");
  
  /// Create one submesh
  Ogre::SubMesh* sub = msh->createSubMesh();

  /// Define the vertices (nVertices vertices, each consisting of 4 groups of 3 floats
  size_t vbufCount = (3+3+3+3)*nVertices;
  /*
    1 -- 8 -- 7
    | \  |  / |
    |  \ | /  |
    2 -- 0 -- 6
    |  / | \  |
    | /  |  \ |
    3 -- 4 -- 5
  */  
  /// Define 8 triangles for a non-cliff tile
  /// The values in this table refer to vertices in the above table
  
  //Assume, by default, that the tile is a cliff-tile
  size_t ibufCount = 54; //18*3
  unsigned short faces[54] = {
			0,1,2,
			3,2,1,
			
			4,5,6,
			7,6,5,
			
			8, 9, 10,
			11,10,9,
			
			12,13,14,
			15,14,13,
			
			16,17,18,
			19,18,17,
			
			20,21,22,
			23,22,21,
			
			24,25,26,
			27,26,25,
			
			28,29,30,
			31,30,29,
			
			32,33,34,
			35,34,33
  };

  if(nVertices == 9){ //i.e. if we actually have a non-cliff tile...
    ibufCount = 24;//8*3
    
    faces[ 0]= 0; faces[ 1]= 1; faces[ 2]= 2;
    faces[ 3]= 0; faces[ 4]= 2; faces[ 5]= 3;
    faces[ 6]= 0; faces[ 7]= 3; faces[ 8]= 4;
    faces[ 9]= 0; faces[10]= 4; faces[11]= 5;
    faces[12]= 0; faces[13]= 5; faces[14]= 6;
    faces[15]= 0; faces[16]= 6; faces[17]= 7;
    faces[18]= 0; faces[19]= 7; faces[20]= 8;
    faces[21]= 0; faces[22]= 8; faces[23]= 1;
    
  }
  /* const size_t ibufCount2 = 54; //18*3
  unsigned short faces[ibufCount2] = {
			0,1,2,
			0,2,3,
			0,3,4,
			0,4,5,
			0,5,6,
			0,6,7,
			0,7,8,
			0,8,1,
			};*/

  prepareFogOfWarTextureCoordinates(vertices, nVertices);

  
  /// Create vertex data structure for the vertices shared between submeshes
  msh->sharedVertexData = new Ogre::VertexData();
  msh->sharedVertexData->vertexCount = nVertices;
  
  /// Create declaration (memory format) of vertex data
  Ogre::VertexDeclaration* decl = msh->sharedVertexData->vertexDeclaration;
  size_t offset = 0;

  // 1st buffer
  decl->addElement(0, offset, Ogre::VET_FLOAT3, Ogre::VES_POSITION);
  offset += Ogre::VertexElement::getTypeSize(Ogre::VET_FLOAT3);
  decl->addElement(0, offset, Ogre::VET_FLOAT3, Ogre::VES_NORMAL);
  offset += Ogre::VertexElement::getTypeSize(Ogre::VET_FLOAT3);
  decl->addElement(0, offset, Ogre::VET_FLOAT3, Ogre::VES_TEXTURE_COORDINATES,0);
  offset += Ogre::VertexElement::getTypeSize(Ogre::VET_FLOAT3);
  decl->addElement(0, offset, Ogre::VET_FLOAT3, Ogre::VES_TEXTURE_COORDINATES,1);
  offset += Ogre::VertexElement::getTypeSize(Ogre::VET_FLOAT3);
	
  /// Allocate vertex buffer of the requested number of vertices (vertexCount) 
  /// and bytes per vertex (offset)
  Ogre::HardwareVertexBufferSharedPtr vbuf = 
    Ogre::HardwareBufferManager::getSingleton().createVertexBuffer(offset, msh->sharedVertexData->vertexCount, 
								   Ogre::HardwareBuffer::HBU_STATIC_WRITE_ONLY);
  /// Upload the vertex data to the card
  vbuf->writeData(0, vbuf->getSizeInBytes(), vertices, true);
  
  /// Set vertex buffer binding so buffer 0 is bound to our vertex buffer
  Ogre::VertexBufferBinding* bind = msh->sharedVertexData->vertexBufferBinding; 
  bind->setBinding(0, vbuf);
  
  /// Allocate index buffer of the requested number of vertices (ibufCount) 
  Ogre::HardwareIndexBufferSharedPtr ibuf = Ogre::HardwareBufferManager::getSingleton().
    createIndexBuffer(Ogre::HardwareIndexBuffer::IT_16BIT, ibufCount, 
		      Ogre::HardwareBuffer::HBU_STATIC_WRITE_ONLY);

  /// Upload the index data to the card
  ibuf->writeData(0, ibuf->getSizeInBytes(), faces, true);
  
  /// Set parameters of the submesh
  sub->useSharedVertices = true;
  sub->indexData->indexBuffer = ibuf;
  sub->indexData->indexCount = ibufCount;
  sub->indexData->indexStart = 0;
	
  /// Set bounding information (for culling)
  msh->_setBounds(Ogre::AxisAlignedBox(0,-1,0,1,1,1));
  msh->_setBoundingSphereRadius(Ogre::Math::Sqrt(3*1*1));
	
  /// Notify Mesh object that it has been loaded
  msh->load();

  Ogre::Entity* newlyMadeTile = mSceneMgr->createEntity(uniqueEntityName, uniqueMeshName );
  newlyMadeTile->setMaterialName(terrainMaterialName);
  //Set all tiles to have the correct normals no matter what (note: disable this to boost performance).
  newlyMadeTile->setNormaliseNormals(true);
  
  return newlyMadeTile;
}

void OgreGfxModuleTerrainTileGrid::prepareFogOfWarTextureCoordinates(float vertices[], int nVertices){
  
  for(int i = 0; i < nVertices * 12; i+=12){
    vertices[i+9] = vertices[i+0]/( ((float)(nCols)) * (tileWidth /2.0f) );
    vertices[i+10]= vertices[i+2]/( ((float)(nRows)) * (tileHeight/2.0f) );
    vertices[i+11]= 0;
  }
  
  return;
}

Ogre::Entity* OgreGfxModuleTerrainTileGrid::createNewBasicTile(
	float MID,
	float N, float NE, float E, float SE,
	float S, float SW, float W, float NW,
	float UNW_1, float VNW_1, float UNE_1, float VNE_1, 
	float USW_1, float VSW_1, float USE_1, float VSE_1, 
	float UNW_2, float VNW_2, float UNE_2, float VNE_2, 
	float USW_2, float VSW_2, float USE_2, float VSE_2)
{
	float UN_1 =(UNE_1 + UNW_1)/2.0f;
	float UE_1 = UNE_1;
	float US_1 =(USE_1 + USW_1)/2.0f;
	float UW_1 = UNW_1;
	
	float VN_1 = VNW_1;
	float VE_1 =(VSE_1 + VNE_1)/2.0f;
	float VS_1 = VSE_1;
	float VW_1 =(VSW_1 + VNW_1)/2.0f;
	
	float UN_2 =(UNE_2 + UNW_2)/2.0f;
	float UE_2 = UNE_2;
	float US_2 =(USE_2 + USW_2)/2.0f;
	float UW_2 = UNW_2;
	
	float VN_2 = VNW_2;
	float VE_2 =(VSE_2 + VNE_2)/2.0f;
	float VS_2 = VSE_2;
	float VW_2 =(VSW_2 + VNW_2)/2.0f;

	float UMD_1 = (UE_1 + UW_1)/2.0f;
	float VMD_1 = (VS_1 + VN_1)/2.0f;
	float UMD_2 = (UE_2 + UW_2)/2.0f;
	float VMD_2 = (VS_2 + VN_2)/2.0f;
	
	//Make sure the meshes have a unique identifying name
	tileCount++;
	Ogre::String uniqueMeshName = "terrainMesh" + Ogre::StringConverter::toString(tileCount);
	Ogre::String uniqueEntityName="terrainEntity"+Ogre::StringConverter::toString(tileCount);
		
	/// Create the mesh via this convienent MeshManager
	Ogre::MeshPtr msh = Ogre::MeshManager::getSingleton().createManual(uniqueMeshName, "General");

	/// Create one submesh
	Ogre::SubMesh* sub = msh->createSubMesh();

	//const float sqrt13 = 0.577350269f; /* sqrt(1/3) */

	/// Define the vertices (9 vertices, each consisting of 4 groups of 3 floats
	const size_t nVertices = 9;
	const size_t vbufCount = (3+3+3+3)*nVertices;
	/*
	0 -- 1 -- 2
	| \  |  / |
	|  \ | /  |
	3 -- 4 -- 5
	|  / | \  |
	| /  |  \ |
	6 -- 7 -- 8
	*/
	float vertices[vbufCount];
	{
		vertices[  0]=0.0;   vertices[  1]= NW;   vertices[  2]=0.0;  //0 position
		vertices[  3]=0.0;   vertices[  4]=1.0;   vertices[  5]=0.0;  //0 normal
		vertices[  6]=UNW_1; vertices[  7]=VNW_1; vertices[  8]=0.0;  //0 first uvw  map
		vertices[  9]=UNW_2; vertices[ 10]=VNW_2; vertices[ 11]=0.0;  //0 second uvw map
		
		vertices[ 12]= .5;   vertices[ 13]= N ;   vertices[ 14]=0.0;  //1 position
		vertices[ 15]=0.0;   vertices[ 16]=1.0;   vertices[ 17]=0.0;  //1 normal
		vertices[ 18]=UN_1;  vertices[ 19]=VN_1;  vertices[ 20]=0.0;  //1 first uvw  map
		vertices[ 21]=UN_2;  vertices[ 22]=VN_2;  vertices[ 23]=0.0;  //1 second uvw map
		
		vertices[ 24]=1.0;   vertices[ 25]= NE;   vertices[ 26]=0.0;  //2 position
		vertices[ 27]=0.0;   vertices[ 28]=1.0;   vertices[ 29]=0.0;  //2 normal
		vertices[ 30]=UNE_1; vertices[ 31]=VNE_1; vertices[ 32]=0.0;  //2 first uvw  map
		vertices[ 33]=UNE_2; vertices[ 34]=VNE_2; vertices[ 35]=0.0;  //2 second uvw map
		
		vertices[ 36]=0.0;   vertices[ 37]= W ;   vertices[ 38]= .5;  //3 position
		vertices[ 39]=0.0;   vertices[ 40]=1.0;   vertices[ 41]=0.0;  //3 normal
		vertices[ 42]=UW_1;  vertices[ 43]=VW_1;  vertices[ 44]=0.0;  //3 first uvw  map
		vertices[ 45]=UW_2;  vertices[ 46]=VW_2;  vertices[ 47]=0.0;  //3 second uvw map
		
		vertices[ 48]= .5;   vertices[ 49]=MID;   vertices[ 50]= .5;  //4 position
		vertices[ 51]=0.0;   vertices[ 52]=1.0;   vertices[ 53]=0.0;  //4 normal
		vertices[ 54]=UMD_1; vertices[ 55]=VMD_1; vertices[ 56]=0.0;  //4 first uvw  map
		vertices[ 57]=UMD_2; vertices[ 58]=VMD_2; vertices[ 59]=0.0;  //4 second uvw map
		
		vertices[ 60]=1.0;   vertices[ 61]= E ;   vertices[ 62]= .5;  //5 position
		vertices[ 63]=0.0;   vertices[ 64]=1.0;   vertices[ 65]=0.0;  //5 normal
		vertices[ 66]=UE_1;  vertices[ 67]=VE_1;  vertices[ 68]=0.0;  //5 first uvw  map
		vertices[ 69]=UE_2;  vertices[ 70]=VE_2;  vertices[ 71]=0.0;  //5 second uvw map
		
		vertices[ 72]=0.0;   vertices[ 73]= SW;   vertices[ 74]=1.0;  //6 position
		vertices[ 75]=0.0;   vertices[ 76]=1.0;   vertices[ 77]=0.0;  //6 normal
		vertices[ 78]=USW_1; vertices[ 79]=VSW_1; vertices[ 80]=0.0;  //6 first uvw  map
		vertices[ 81]=USW_2; vertices[ 82]=VSW_2; vertices[ 83]=0.0;  //6 second uvw map
		
		vertices[ 84]= .5;   vertices[ 85]= S ;   vertices[ 86]=1.0;  //7 position
		vertices[ 87]=0.0;   vertices[ 88]=1.0;   vertices[ 89]=0.0;  //7 normal
		vertices[ 90]=US_1;  vertices[ 91]=VS_1;  vertices[ 92]=0.0;  //7 first uvw  map
		vertices[ 93]=US_2;  vertices[ 94]=VS_2;  vertices[ 95]=0.0;  //7 second uvw map
		
		vertices[ 96]=1.0;   vertices[ 97]= SE;   vertices[ 98]=1.0;  //8 position
		vertices[ 99]=0.0;   vertices[100]=1.0;   vertices[101]=0.0;  //8 normal
		vertices[102]=USE_1; vertices[103]=VSE_1; vertices[104]=0.0;  //8 first uvw  map
		vertices[105]=USE_2; vertices[106]=VSE_2; vertices[107]=0.0;  //8 second uvw map
	}

	/// Define 8 triangles
	/// The values in this table refer to vertices in the above table
	const size_t ibufCount = 24;
	unsigned short faces[ibufCount] = {
			4,0,3,
			4,1,0,
			4,2,1,
			4,5,2,
			4,3,6,
			4,6,7,
			4,7,8,
			4,8,5,
	};

	/// Create vertex data structure for 9 vertices shared between submeshes
	msh->sharedVertexData = new Ogre::VertexData();
	msh->sharedVertexData->vertexCount = nVertices;

	/// Create declaration (memory format) of vertex data
	Ogre::VertexDeclaration* decl = msh->sharedVertexData->vertexDeclaration;
	size_t offset = 0;

	// 1st buffer
	decl->addElement(0, offset, Ogre::VET_FLOAT3, Ogre::VES_POSITION);
	offset += Ogre::VertexElement::getTypeSize(Ogre::VET_FLOAT3);
	decl->addElement(0, offset, Ogre::VET_FLOAT3, Ogre::VES_NORMAL);
	offset += Ogre::VertexElement::getTypeSize(Ogre::VET_FLOAT3);
	decl->addElement(0, offset, Ogre::VET_FLOAT3, Ogre::VES_TEXTURE_COORDINATES);
	offset += Ogre::VertexElement::getTypeSize(Ogre::VET_FLOAT3);
	decl->addElement(0, offset, Ogre::VET_FLOAT3, Ogre::VES_TEXTURE_COORDINATES,1);
	offset += Ogre::VertexElement::getTypeSize(Ogre::VET_FLOAT3);
	
	/// Allocate vertex buffer of the requested number of vertices (vertexCount) 
	/// and bytes per vertex (offset)
	Ogre::HardwareVertexBufferSharedPtr vbuf = 
		Ogre::HardwareBufferManager::getSingleton().createVertexBuffer(
		offset, msh->sharedVertexData->vertexCount, Ogre::HardwareBuffer::HBU_STATIC_WRITE_ONLY);
	/// Upload the vertex data to the card
	vbuf->writeData(0, vbuf->getSizeInBytes(), vertices, true);

	/// Set vertex buffer binding so buffer 0 is bound to our vertex buffer
	Ogre::VertexBufferBinding* bind = msh->sharedVertexData->vertexBufferBinding; 
	bind->setBinding(0, vbuf);

	/// Allocate index buffer of the requested number of vertices (ibufCount) 
	Ogre::HardwareIndexBufferSharedPtr ibuf = Ogre::HardwareBufferManager::getSingleton().
		createIndexBuffer(
		Ogre::HardwareIndexBuffer::IT_16BIT, 
		ibufCount, 
		Ogre::HardwareBuffer::HBU_STATIC_WRITE_ONLY);

	/// Upload the index data to the card
	ibuf->writeData(0, ibuf->getSizeInBytes(), faces, true);

	/// Set parameters of the submesh
	sub->useSharedVertices = true;
	sub->indexData->indexBuffer = ibuf;
	sub->indexData->indexCount = ibufCount;
	sub->indexData->indexStart = 0;

	/// Set bounding information (for culling)
	msh->_setBounds(Ogre::AxisAlignedBox(0,-1,0,1,1,1));
	msh->_setBoundingSphereRadius(Ogre::Math::Sqrt(3*1*1));

	/// Notify Mesh object that it has been loaded
	msh->load();

	Ogre::Entity* newlyMadeTile = mSceneMgr->createEntity(uniqueEntityName, uniqueMeshName );
	newlyMadeTile->setMaterialName(terrainMaterialName);
	
	return newlyMadeTile;
}

void OgreGfxModuleTerrainTileGrid::addTileToGeometry(int tiler, int tilec, Ogre::Entity* newOgreTile){
	(megaTiles[tiler/mRows][tilec/mCols]).addGeometry(newOgreTile,tileWidth*(float)tilec,0.0f,tileHeight*(float)tiler, tileWidth, tileWidth, tileHeight);
	bRebuildNeeded = true;
}

//---have the fog of war's alphaValues move towards what they should
void OgreGfxModuleTerrainTileGrid::updateTemporalFade(int fadeAmount){
	
  
  for(int i = 0; i< rows; i++){
    for(int j = 0; j< columns; j++){
      if(!revealed[i][j] && alphaVals[i][j] != 255){
	alphaVals[i][j] += fadeAmount;
	if(alphaVals[i][j] > 255)
	  alphaVals[i][j] = 255;
      }
      else if(revealed[i][j] && !visible[i][j] && alphaVals[i][j] != 128){
	alphaVals[i][j] += fadeAmount;
	if(alphaVals[i][j] > 128)
	  alphaVals[i][j] = 128;
      }

      else if(visible[i][j] && alphaVals[i][j] != 0){
	alphaVals[i][j] -= fadeAmount;
	if(alphaVals[i][j] < 0)
	  alphaVals[i][j] = 0;
      } 
    }
  }

  refreshFogOfWar();
}

void OgreGfxModuleTerrainTileGrid::refreshFogOfWar(){
	if(bFogOfWarRefreshNeeded){
		Ogre::HardwarePixelBufferSharedPtr pixelBuffer = fogOfWartexture->getBuffer();
		
		//---Lock the pixel buffer and start messing with it
		pixelBuffer->lock(Ogre::HardwareBuffer::HBL_DISCARD); //for best performance don't use HBL_NORMAL!
		const Ogre::PixelBox& pixelBox = pixelBuffer->getCurrentLock();

		Ogre::uint8* pDest = static_cast<Ogre::uint8*>(pixelBox.data);

		//---Update the fog of war's texture
		float rScaleFactor = (float)fogOfWarTexturePixelsWidth/(float)rows;
		float cScaleFactor = (float)fogOfWarTexturePixelsHeight/(float)columns;

		for (size_t i = 0; i < (size_t)fogOfWarTexturePixelsHeight; i++)
			for(size_t j = 0; j < (size_t)fogOfWarTexturePixelsWidth; j++){
				pDest++; // B
				pDest++; // G
				pDest++; // R
				*pDest++ = alphaVals[(int)((float)i/rScaleFactor)][(int)((float)j/cScaleFactor)]; // A
			}

		// Unlock the pixel buffer
		pixelBuffer->unlock();
	}
	return;
}

//---Has all megaTiles rebuild themselves (i.e. refresh)
void OgreGfxModuleTerrainTileGrid::refreshGeometry(){
	if(bRebuildNeeded){
		for(int i = 0; i< nRows; i++)
			for(int j=0; j< nCols; j++)
				(megaTiles[i][j]).rebuild();
		bRebuildNeeded= false;
		synchronizeVisibilityGrid();
	}
}

//---Non-vital function who's sole-purpose is to make the fog of war look 'nicer'
void OgreGfxModuleTerrainTileGrid::synchronizeVisibilityGrid(){
  
  for(int i = 0; i < rows; i++)
    for(int j = 0; j < columns; j++){
      if(revealed[i][j] && i-1>0 && i+1<rows && j-1>0 && j+1<columns)
	if(!revealed[i-1][j] || !revealed[i+1][j] || !revealed[i][j-1] || !revealed[i][j+1])
	  visible[i][j] = false;
    }
  
  return;
}
