#ifndef PRAENGINE_H
#define PRAENGINE_H

// $Id: PRAEngine.H,v 1.1 2006/12/31 04:09:40 orts_furtak Exp $

/** \brief Pathfinding Engine for PRA* pathfinding
 *
 * Uses HOG classes to find the paths...
 *
 */

#include "TerrainBase.H"
#include "Global.H"
#include "Options.H"
#include "mapCliqueAbstraction.h"
//#include "sharedAMapGroup.h"
#include "map.h"
#include "graph.h"
#include "pr_reservationProvider.H"
#include "praStar.h"
#include "PRATerrain.H"

//using namespace HOG;
//namespace SimpleTerrain{

class PR_Terrain::PRAEngine {

public:

  PRAEngine(sint4 tiles_x_, sint4 tiles_y_, sint4 tile_points_, int gran);
  ~PRAEngine();
  //---------------------------------------------------------------------
  /// \name Changes to the world.
  //---------------------------------------------------------------------
  //@{
  void insert_object(const Object *obj);
  void remove_object(const Object *obj);
  void update_object(const Object *obj);
  void insert_boundary(const TerrainBase::Segment& line);
  //@}
  //---------------------------------------------------------------------
  void repair_abstraction();
  
  /// Perform pathfinding, return true iff path found, and the path itself in 'output'
  bool find_path(const Object *obj, const TerrainBase::Loc &goal, TerrainBase::Path& output);
  bool find_path(const Object *obj, const Object *goal, TerrainBase::Path& output);

protected:
  HOG::Map map;
  HOG::mapCliqueAbstraction aMap;
  HOG::PR_ReservationProvider res_provider;
  HOG::praStar prastar;

  sint4 subtile_points;
  sint4 disp;

  bool find_path(HOG::node *start, HOG::node *end, TerrainBase::Path& output);
  
  /** find the node in the abstract map corresponding to the
      central point of the object */
  HOG::node* get_center_node(const Object *obj);
  
  void remove_edges(sint4 x, sint4 y);
  //void ir_bucket(const Demo_SimpleTerrain::Bucket &bucket, Object::ZCat zcat, bool insert);
  void ir_rectangle(TerrainBase::Loc &topleft, TerrainBase::Loc &bottomright, Object::ZCat zcat, bool insert);
  void ir_obj(const Object *obj, bool insert, bool use_prev_attributes);
  void remove_cross_edge(sint4 x, sint4 y, HOG::tCorner corner);
 
  void smooth_path(Vector<TerrainBase::Loc> &path);
  
  sint4 world2x(sint4 worldx) const { return ((worldx+disp) / subtile_points); };
  sint4 world2y(sint4 worldy) const { return ((worldy+disp) / subtile_points); };
  sint4 x2world(sint4 x) const { return (x)*subtile_points-disp + disp; };
  sint4 y2world(sint4 y) const { return (y)*subtile_points-disp + disp; };
  bool on_line(const TerrainBase::Loc &a, const TerrainBase::Loc &b, const TerrainBase::Loc &c);
  
  
  
};
//}

#endif
 
