#ifndef DCDT_H
#define DCDT_H

#include "Global.H"

namespace DCDT {
  
  struct Point
  {
    sint4 x, y;

    Point() { x = y = 0; }
    
    Point(sint4 x_, sint4 y_) : x(x_), y(y_) { } 

    bool operator==(const Point &other) {
      return x == other.x && y == other.y;
    }

    double distance(const Point &other) {
      real8 x_ = x - other.x;
      real8 y_ = y - other.y;

      return sqrt(square(x_)+square(y_));
    }
  };


  struct Segment
  {
    sint4 x1, y1, x2, y2;

    Segment()
    {
      x1 = y1 = x2 = y2 = 0;
    }
    
    Segment(sint4 x1_, sint4 y1_, sint4 x2_, sint4 y2_)
      : x1(x1_), y1(y1_), x2(x2_), y2(y2_)
    { }

    // left-bottom point first
    void normalize() {
      if (x2 < x1 || (x1 == x2 && y2 < y1)) { std::swap(x1, x2); std::swap(y1, y2); }
    }
    
    void write(std::ostream &os = std::cout) const {
      os << "[" << x1 << "," << y1 << "]-"
	 << "[" << x2 << "," << y2 << "]";
    }
  };

}

#endif
