/*
    Copyright (C) 2001 by W.C.A. Wijngaards

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

// modified by Michael Buro 2003

#ifndef Maze_H
#define Maze_H

#include "Global.H"

///
/// node for Maze
///
class MazeNode {
public:
  /// visited
  bool visited;
  /// openings to n,e,s,w?
  bool opening[4];
};


///
/// Class to generate gridlike mazes
///

class Maze {

  /// maze size
  sint4 width, height;

  /// maze contents
  MazeNode *maze;

  /// straightness
  float straightness;

  /// cyclicalness
  float cyclicalness;

public:

  /// give width and height of maze, initialises maze.
  Maze(sint4 w, sint4 h);

  ~Maze();
  /// init maze to fully walled in.
  void init_maze();

  /// set the straightness of the corridors in the maze
  void set_straightness(float n) {straightness = n;}

  /// get the straightness of the corridors in the maze
  float get_straightness() const {return straightness;}

  /// set the cyclicalness of the corridors in the maze
  void set_cyclicalness(float n) {cyclicalness = n;}

  /// get the cyclicalness of the corridors in the maze
  float get_cyclicalness() const {return cyclicalness;}

  /// Set a position at an edge to be an 'access point' to the maze.
  /// there will be an opening to the outside there.
  /// (corner points will make double accesses)
  void make_access(sint4 x, sint4 y);

  /// generate the maze from a given starting point
  void generate(sint4 x, sint4 y);

  /// Prints ascii Maze - amazing :-)
  void print();

  /// Return actual height - allows for thickness of walls
  sint4 actual_height();

  /// Return actual width - allows for thickness of walls
  sint4 actual_width();

  /// Returns true if x,y is solid - note this function adjusts
  /// for tickness of walls
  bool actual_solid(sint4 x, sint4 y);

  /// get a maze node
  MazeNode& get_node(sint4 x, sint4 y)
  {
    assert(x >= 0 && x < width);
    assert(y >= 0 && y < height);
    return maze[y*width+x];
  }

  /// visit a node, used to gen maze.
  void visit_node(sint4 x, sint4 y, sint4 direction);

  /// in gen maze: generate an order with preference
  void gen_order(sint4 order[4], sint4 direction);

  /// get neightboring position in direction
  void get_neighbor(sint4 dir, sint4 x, sint4 y, sint4& nx, sint4& ny);

  /// make an opening between two nodes
  void make_opening(sint4 x1, sint4 y1, sint4 x2, sint4 y2);

  /// is there an opening between two nodes?
  bool opening(sint4 x1, sint4 y1, sint4 x2, sint4 y2);

};


#endif
